/*
 PPU.h

 * MacFCEU 0.9a2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* ---- constants */

/* number of PPU cycles/rendered pixels per scanline - 256 rendered, 85 invisible */
#define LCYCS			341

/* PPU type bits */
#define PPU2C02			(1<<0)
#define RP2C03B			(1<<1)		/* bog standard RGB palette */
#define RP2C03G			(1<<2)		/* similar palette, may have 1 color changed */
#define RP2C04_0001		(1<<3)		/* scrambled palette + new colors */
#define RP2C04_0002		(1<<4)		/* different scrambling, different new colors */
#define RP2C04_0003		(1<<5)		/* (similar to above) */
#define RP2C04_0004		(1<<6)		/* (similar to above) */
#define kRC2C03B		7
#define RC2C03B			(1<<kRC2C03B)
#define RC2C03C			(1<<8)		/* 1 changed color or so */
#define RC2C05_01		(1<<9)		/* all five of these have the normal palette...*/
#define RC2C05_02		(1<<10)		/*...but with different bits returned on 2002 */
#define RC2C05_03		(1<<11)
#define RC2C05_04		(1<<12)
#define RC2C05_05		(1<<13)
/* RC2C05_99 - Sharp Titler */
#define PPU2C07			(1<<14)
#define PPU6538			(1<<15)		/* Dendy */

#define RCandP2C03B		(RP2C03B|RC2C03B)

/* ---- globals */

extern void FP_FASTAPASS(1) (*PPU_hook)(uint32 A);
extern void (*PPU_A12hook)(void), (*GameHBIRQHook)(void), (*GameHBIRQHook2)(void);

/* For cart.c and banksw.h, mostly */
extern uint8 NTARAM[0x800], *vnapage[4];
extern uint8 PPUNTARAM, PPUCHRRAM;

extern uint8 PALRAM[0x20], PPUvoidcolor;

extern int scanline;
extern uint32 PPUVFrameCounter;
extern uint8 PPUVFirstField, PPUBurstCount;
extern uint16 PPUChipType;
extern uint8 PPUReg[4];

/* ---- prototypes */

void FCEUPPU_Init(void);
void FCEUPPU_Reset(int NESboot);
void FCEUPPU_Power(void);
int FCEUPPU_Loop(int skip);

void FCEUPPU_LineUpdate();
void FCEUPPU_SetVideoSystem(int which); /* GIV_* */

void FCEUPPU_SaveState(void);
void FCEUPPU_LoadState(int version);

/* call this after modifying PALRAM; 2010.04.15 */
void FCEUPPU_UpdatePalette(void);

int FCEUPPU_GetPixel(void);
uint8 FCEUPPU_MemPeek(uint16 a);
void FCEUPPU_MemPoke(uint16 a,
	uint8 v,
	int hl);
int FCEUPPU_MemGetBank(uint16 A);

uint8 FCEUPPU_OAMPeek(uint16 a);
void FCEUPPU_OAMPoke(uint16 a,
	uint8 v,
	int hl);
