/*
 VSUni.c

 * MacFCEU 0.9a2
 *
 * Copyright notice for this file:
 *  vsuni.c Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "driver.h"
#include "FCEU.h"
#include "input.h"
#include "netplay.h"
#include "VSUni.h"
#include "state.h"
#include "PPU.h"
#include "video.h" //for rowsize

#define INESPRIV
#include "iNES.h"
#include "cart.h"

#define IOPT_GUN		0x1
#define IOPT_SWAPDIRAB		0x2
#define IOPT_PREDIP		0x10

typedef struct VSUNIENTRY
{
	char	*name;
	uint64	md5partial;
	uint16	PPU;
	uint8	mapper;
	uint8	mirroring;
	uint8	ioption;
	uint8	predip;
} VSUNIENTRY;

VSUNIENTRY	*curvs;

static int64	curmd5;

uint8		VSDIPS = 0;
uint8		vsdip = 0;

uint8		VSCoinOn[3];
uint16		VSServiceRC;
static uint8	coinct;
static int8	coinmotor;
uint8		watchdogtimer;

static uint8	*secptr;
static uint8	VSindex;
static uint8	is99, m99w;
static uint8	xevselect = 0;

/* Games that will probably not be supported ever (or for a long time),
 since they require dual-system:

	Balloon Fight
	VS Mahjong
	VS Tennis
	Wrecking Crew
*/

/* Games/PPU list.  Information copied from MAME.  ROMs are exchangable,
 so don't take this list as "this game must use this PPU".

 Checksum reflects zero padding in iNES form.


RP2C04-001:
- Baseball
- Freedom Force
- Gradius
- Hogan's Alley
- Mach Rider (Japan, Fighting Course)
- Pinball
- Platoon
- Super Xevious

RP2C04-002:
- Castlevania
- Ladies golf
- Mach Rider (Endurance Course)
- Raid on Bungeling Bay (Japan)
- Slalom
- Stroke N' Match Golf
- Wrecking Crew

RP2C04-003:
- Dr mario
- Excite Bike
- Goonies
- Soccer
- TKO Boxing

RP2c05-004:
- Clu Clu Land
- Excite Bike (Japan)
- Ice Climber
- Ice Climber Dual (Japan)
- Super Mario Bros.

Rcp2c03b:
- Battle City
- Duck Hunt
- Mahjang
- Pinball (Japan)
- Rbi Baseball
- Star Luster
- Stroke and Match Golf (Japan)
- Super Skykid
- Tennis
- Tetris

RC2C05-01:
- Ninja Jajamaru Kun (Japan)

RC2C05-02:
- Mighty Bomb Jack (Japan)

RC2C05-03:
- Gumshoe

RC2C05-04:
- Top Gun
*/

/* 151 is incorrect! */

VSUNIENTRY VSUniGames[]	=
{
 {"Baseball",			_int64val(0x691d4200L,0xea42be45L),RP2C04_0001, 99, 2,0},
 {"Battle City",		_int64val(0x8540949dL,0x74c4d0ebL),RP2C04_0001, 99, 2,0},
 {"Battle City(Bootleg)",	_int64val(0x8093cbe7L,0x137ac031L),RP2C04_0001, 99, 2,0},

 {"Clu Clu Land",		_int64val(0x1b812321L,0x8f62b1eeL),RP2C04_0004, 99, 2,IOPT_SWAPDIRAB},
 {"Dr Mario",			_int64val(0xe1af09c4L,0x77dc0081L),RP2C04_0003,  1, 0,IOPT_SWAPDIRAB},
 {"Duck Hunt",			_int64val(0x47735d1eL,0x5f1205bbL),RCandP2C03B, 99, 2,IOPT_GUN},
 {"Excitebike",			_int64val(0x3dcd1401L,0xbcafde77L),RP2C04_0003, 99, 2,0},
 {"Excitebike (J)",		_int64val(0x7ea51c9dL,0x007375f0L),RP2C04_0004, 99, 2,0},
 {"Freedom Force",		_int64val(0xed96436bL,0xd1b5e688L),RP2C04_0001,  4, 0,IOPT_GUN}, /* Wrong color in game select screen? */
 {"Stroke and Match Golf",	_int64val(0x61232560L,0x6e82bc66L),RP2C04_0002, 99, 2,IOPT_SWAPDIRAB|IOPT_PREDIP,0x01},

 {"Goonies",			_int64val(0x3b0085c4L,0xff29098eL),RP2C04_0003,151, 1,0},
 {"Gradius",			_int64val(0x50687ae6L,0x3bdad976L),RP2C04_0001,151, 1,IOPT_SWAPDIRAB},
 {"Gumshoe",			_int64val(0xb8500780L,0xbf69ce29L),  RC2C05_03, 99, 2,IOPT_GUN},
 {"Hogan's Alley",		_int64val(0xd78b7f0bL,0xb621fb45L),RP2C04_0001, 99, 2,IOPT_GUN},
 {"Ice Climber",		_int64val(0xd21e9995L,0x13435e2aL),RP2C04_0004, 99, 2,IOPT_SWAPDIRAB},
 {"Ladies Golf",		_int64val(0x781b24beL,0x57ef6785L),RP2C04_0002, 99, 2,IOPT_SWAPDIRAB|IOPT_PREDIP,0x1},

 {"Mach Rider",			_int64val(0x01567261L,0x8af06441L),RP2C04_0002, 99, 2,0},
 {"Mach Rider (J)",		_int64val(0xa625afb3L,0x99811a8aL),RP2C04_0001, 99, 2,0},
 {"Mighty Bomb Jack",		_int64val(0xe6a89f48L,0x73fac37bL),  RC2C05_02,  0, 2,0},
 {"Ninja Jajamaru Kun",		_int64val(0xb26a2c31L,0x474099c0L),  RC2C05_01, 99, 2,IOPT_SWAPDIRAB},
 {"Pinball",			_int64val(0xc5f49d3dL,0xe7f2e9b8L),RP2C04_0001, 99, 2,IOPT_PREDIP,0x01},
 {"Pinball (J)",		_int64val(0x66ab1a38L,0x28cc901cL),RCandP2C03B, 99, 2,IOPT_PREDIP,0x1},
 {"Platoon",			_int64val(0x160f2373L,0x51c19f1fL),RP2C04_0001, 68, 1,0},
 {"RBI Baseball",		_int64val(0x6a02d345L,0x812938afL),RP2C04_0001,  4, 1,IOPT_SWAPDIRAB},
 {"Soccer",			_int64val(0xd4e7a905L,0x8780eda3L),RP2C04_0003, 99, 2,IOPT_SWAPDIRAB},
 {"Star Luster",		_int64val(0x8360e134L,0xb316d94cL),RCandP2C03B, 99, 2,0},
 {"Stroke and Match Golf (J)",	_int64val(0x869bb83eL,0x02509747L),RCandP2C03B, 99, 2,IOPT_SWAPDIRAB|IOPT_PREDIP,0x1},
 {"Super Sky Kid",		_int64val(0x78d04c1dL,0xd4ec0101L),RCandP2C03B,  4, 1,IOPT_SWAPDIRAB|IOPT_PREDIP,0x20},

 {"Super Xevious",		_int64val(0x2d396247L,0xcf58f9faL),RP2C04_0001,206, 0,0},
 {"Tetris",			_int64val(0x531a5e8eL,0xea4ce157L),RCandP2C03B, 99, 2,IOPT_PREDIP,0x20},
 {"Top Gun",			_int64val(0xf1dea36eL,0x6a7b531dL),  RC2C05_04,  2, 0,0},
 {"VS Castlevania",		_int64val(0x92fd6909L,0xc81305b9L),RP2C04_0002,  2, 1,0},
 {"VS Slalom",			_int64val(0x4889b5a5L,0x0a623215L),RP2C04_0002,  0, 1,0},
 {"VS Super Mario Bros",	_int64val(0x39d8cfa7L,0x88e20b6cL),RP2C04_0004, 99, 2,0},
 {"VS TKO Boxing",		_int64val(0x6e1ee061L,0x71d8ce3aL),RP2C04_0003,  4, 1,IOPT_PREDIP,0x00},
 {0}
};

static uint8	security[2][32] =
{
	{
		0xff, 0xbf, 0xb7, 0x97, 0x97, 0x17, 0x57, 0x4f,
		0x6f, 0x6b, 0xeb, 0xa9, 0xb1, 0x90, 0x94, 0x14,
		0x56, 0x4e, 0x6f, 0x6b, 0xeb, 0xa9, 0xb1, 0x90,
		0xd4, 0x5c, 0x3e, 0x26, 0x87, 0x83, 0x13, 0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0xb4, 0x00, 0x00, 0x00,
		0x00, 0x6F, 0x00, 0x00, 0x00, 0x00, 0x94, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	}
};

SFORMAT FCEUVSUNI_STATEINFO[] =
{
	{&vsdip,1,"vsdp"},
	{VSCoinOn,3,"vscn"},
	{&coinct,1,"vscc"},
	{&VSServiceRC,2|FCEUSTATE_RLSB,"vssc"},
	{&VSindex,1,"vsin"},
	{0}
};

/* ---- code */

void Mapper99_init(void)
{ /* in 40KB case, second bank at #4 instead of #1 so first 32KB can init. as-is */
	ROM_BANK32(0);
}

void FCEUI_VSUniToggleDIPView(void)
{
	VSDIPS = !VSDIPS;
}

void FCEU_VSUniToggleDIP(int w)
{
	vsdip^= 1<<w;
}

void FCEUI_VSUniSetDIP(int w,
	int state)
{
	if ((vsdip>>w&1)!=state)
		FCEUI_VSUniToggleDIP(w);
}

uint8 FCEUI_VSUniGetDIPs(void)
{
	return vsdip;
}

static DECLFR(VSSecRead)
{
	if (A==0x5E00)
		{VSindex = 0; return A>>8;} //XFrame.DB
	else if (A==0x5E01)
		return secptr[VSindex++&0x1F];
	return 0x00;
}

void FCEU_VSUniCoin(int w)
{
	if (w>2u)
		return;
	VSCoinOn[w] = 5;
}

void FCEU_VSUniFrame(void)
{
	if (VSCoinOn[0])
		--VSCoinOn[0];
	if (VSCoinOn[1])
		--VSCoinOn[1];
	if (VSCoinOn[2])
	{ /* basically charges RC for 4 frames */
		--VSCoinOn[2];
		VSServiceRC+= 0x4940;
	}
	VSServiceRC-= VSServiceRC>>2;
	if (VSServiceRC>0xB640)
		VSServiceRC = 0xB640;
	else if (VSServiceRC<3)
		VSServiceRC = 3;
	if (coinmotor>0)
		--coinmotor;
	if (watchdogtimer && !--watchdogtimer)
	{
		FCEU_ResetHW(0);
		coinmotor = 0;
	}
	
}

void VSUniResetWatchDog(void)
{
	watchdogtimer = 72; /* 1.2s */
}

static DECLFW(VSCoinCtWrite)
{
(void)A;
	if (V&1)
	{
		if (!coinmotor)
		{
			if (!++coinct)
				coinct = 255;
			coinmotor =-1;
		}
	} else
		coinmotor = 3;
}

static DECLFR(VSCoinCtRead)
{
	A>>= 8;
	if (!fceuindbg)
		VSCoinCtWrite(A,A);
	return A;
}

/* xxxx xCRS
 S = status
 R = secondary on DualSystem, connects to 2nd CPU IRQ, RAM for 6000-7FFF
 C = 8KB CHR select (mapper 99); for 40KB PRG games, swaps 8000-9FFF PRG
*/

void VSUniCtlBank(uint8 V)
{
	if (!is99)
		return;
	m99w = V;
	//if (V&1)
	//	return;
	VROM_BANK8(V>>2&1);
	ROM_BANK8(0x8000,V&4);
}

void FCEU_VSUniSwap(uint8 *j0,
	uint8 *j1)
{
	if ((curvs && curvs->ioption&IOPT_SWAPDIRAB) || is99) /* should really be default */
	{
		uint16		t = *j0;

		*j0 = (*j0&0xC)|(*j1&0xF3);
		*j1 = (*j1&0xC)|(t&0xF3);
	}
}

/* Super Xevious */

static DECLFR(XevRead)
{
//printf("%04x\n",A);
	if (A==0x54FF)
		return 5;
	else if (A==0x5678)
		return xevselect ? 0:1;
	else if (A==0x578F)
		return xevselect ? 0xD1:0x89;
	else if (A==0x5567)
		return (xevselect^=1)!=0 ? 0x37:0x3E;
	return A>>8; //XFrame.DB
}

static DECLFW(VSUni_BWRAM)
{
	if (m99w&2)
		WRAM[A&0x7FF] = V;
}

static DECLFR(VSUni_AWRAM)
{
	return WRAM[A&0x7FF];
}

void FCEU_VSUniPower(void)
{
	uint32		A;

	VSCoinOn[0] = VSCoinOn[1] = coinct = VSindex = 0;
	A = 0x4020;
	do
	{
		SetReadHandler(A,A+(255-32),VSCoinCtRead);
		SetWriteHandler(A,A+(255-32),VSCoinCtWrite);
	} while((A+=0x100)<0x5FFF);
	/* these will override coin mirror */
	if (secptr)
		SetReadHandler(0x5E00,0x5E01,VSSecRead);
	if (int64eqhl(curmd5,0x2d396247L,0xcf58f9faL)) /* Super Xevious */
		SetReadHandler(0x5400,0x57FF,XevRead);
	if (is99)
	{
		SetReadHandler(0x6000,0x7FFF,VSUni_AWRAM);
		SetWriteHandler(0x6000,0x7FFF,VSUni_BWRAM);
	}
}

int FCEU_VSUniCheck(uint64 md5partial,
	int *MapperNo,
	uint8 *Mirroring,
	uint16 *defPPU)
{
	VSUNIENTRY	*vs;

	for(vs=VSUniGames; vs->name; ++vs)
	{
		if (!int64eq(md5partial,vs->md5partial))
			continue;
		//puts(vs->name);
		*MapperNo = vs->mapper;
		*Mirroring = vs->mirroring;
		*defPPU = vs->PPU;
		goto defo;
	}
	is99 = 0;
	if (!MapperNo || !Mirroring)
		return -1;

	if (*defPPU&~(PPU2C02|PPU2C07|PPU6538)) /* final case; use what the header points to */
		{vs = 0; goto defo;}
	else if (*MapperNo==99 || *MapperNo==151)
	{
		*defPPU = RP2C03B, vs = 0;

	defo:	FCEUGameInfo->type = GIT_VSUNI;
		FCEUGameInfo->cspecial = SIS_VSUNISYSTEM;
		FCEUGameInfo->inputfc = SI_NONE;
		int64set(curmd5,md5partial);

		if (int64eqhl(md5partial,0x6e1ee061L,0x71d8ce3aUL))
			secptr = security[0];
		else if (int64eqhl(md5partial,0x6a02d345,0x812938afUL))
			secptr = security[1];
		else
			secptr = 0;

		vsdip = 0x0;
		if ((curvs=vs)!=0 && vs->ioption&IOPT_PREDIP)
			vsdip = vs->predip;
		if (vs && vs->ioption&IOPT_GUN)
		{
			FCEUGameInfo->input[0] = SI_ZAPPER;
			FCEUGameInfo->input[1] = SI_NONE;
		} else
			FCEUGameInfo->input[0] = FCEUGameInfo->input[1] = SI_GAMEPAD;
		is99 = *MapperNo==99;
		return 1;
	}
	return 0;
}

void FCEU_VSUniDraw(PLB_pel *XBuf)
{
	char		buf[32], *s;
	uint32		*d, dke4;
	int		i;

	if (!VSDIPS)
		return;
	buf[sizeof(buf)-1] = 0;
	s = FCEU_PrintDec(coinct,buf,buf+sizeof(buf)-1,0);
	i = 11*FCEUSCRN_ROWSIZE+128+8u*(s-(buf+sizeof(buf)));
	DrawTextTrans(XBuf+i,FCEUSCRN_ROWSIZE,(const uint8*)s, PAL_LTGRAY);
#if FCEUSCRN_ROWSIZE==512
	DrawTextTrans(XBuf+i+(FCEUSCRN_ROWSIZE>>1),FCEUSCRN_ROWSIZE,(const uint8*)s, PAL_LTGRAY);
#endif
	d = (uint32*)(XBuf+FCEUSCRN_ROWSIZE*10+132); /* was 164 */
	/* base - top line */
	i = FCEUSCRN_ROWSIZE/256;
	do
		*d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ =
			*d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ =
			PAL_LTGRAY*0x1010101;
	while(d+=(256-72)>>2,--i);
	/* base and shadow */
	i = 7*FCEUSCRN_ROWSIZE/256;
	do /*x=72>>2 (18)*/
		*d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ =
			*d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ = *d++ =
			PAL_LTGRAY*0x1010101, *(PLB_pel*)d|= epal_dke;
	while(d+=(256-72)>>2,--i);
	/* base bottom shadow */
	dke4 = epal_dke, dke4|= dke4<<8, dke4|= dke4<<16;
	i = FCEUSCRN_ROWSIZE/256;
	do
		*d|= dke4&_be32_native(0xFFFFFF), d[1]|= dke4, d[2]|= dke4, d[3]|= dke4,
		d[4]|= dke4, d[5]|= dke4, d[6]|= dke4, d[7]|= dke4, d[8]|= dke4,
		d[9]|= dke4, d[10]|= dke4, d[11]|= dke4, d[12]|= dke4, d[13]|= dke4,
		d[14]|= dke4, d[15]|= dke4, d[16]|= dke4, d[17]|= dke4,
		*(PLB_pel*)&d[18]|= epal_dke;
	while(d+=256>>2,--i);
	/* short segment */
	((PLB_pel*)d)+= 6-FCEUSCRN_ROWSIZE;
	i = 1*FCEUSCRN_ROWSIZE/256;
	do
		*d = d[2] = d[4] = d[6] = d[8] = d[10] = d[12] = d[14] = PAL_WHITE*0x1010101;
	while(d+=256>>2,--i);
	/* extension and shadow */
	i = 0;
	do
	{
		((PLB_pel*)d)[4]|= epal_dke;
		((PLB_pel*)d)[4+256]|= epal_dke;
		if (vsdip>>i&1)
		{
			*d|= dke4&_be32_native(0xFFFFFF);
		#if FCEUSCRN_ROWSIZE==512
			d[256>>2]|= dke4&_be32_native(0xFFFFFF);
		#endif
			continue;
		}
		*d = d[256>>2] = d[256*2>>2] = d[256*3>>2] = PAL_WHITE*0x1010101;
		((PLB_pel*)d)[4+256*2]|= epal_dke;
		((PLB_pel*)d)[4+256*3]|= epal_dke;
		d[256*4>>2]|= dke4&_be32_native(0xFFFFFF), ((PLB_pel*)d)[256*4+4]|= epal_dke;
	#if FCEUSCRN_ROWSIZE==512
		d[256*5>>2]|= dke4&_be32_native(0xFFFFFF), ((PLB_pel*)d)[256*5+4]|= epal_dke;
	#endif
	} while(d+=2,++i<8);
}
