/*
 boards/Bandai.c
  Bandai

 * MacFCEU 0.9a1
 *
 * Copyright notice for this file:
 *  karaoke.c Copyright (C) 2002 Xodnizel
 *  bandai.c Copyright (C) 2007 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "Bandai.h"
#include "../cheat.h"

static uint8	DReg[12], is153, is24C02, IRQa;
static int16	IRQCount, IRQLatch;

static uint8 *WRAM = NULL;
static uint32 WRAMSIZE;

static SFORMAT StateRegs[] =
{
	{DReg,10,"REGS"},
	{&IRQa,1,"IRQA"},
	{&IRQCount,2,"IRQC"},
	{&IRQLatch,2,"IRQL"},	// need for Famicom Jump II - Saikyou no 7 Nin (J) [!]
	{0}
};

// x24C0x interface

enum
{
	BLMode_Idle,
	BLMode_Load,
	BLMode_Addr,
	BLMode_Read,
	BLMode_Write,
	BLMode_Ack,
	BLMode_NotAck,
	BLMode_AckWait
};

#define EEPROM GameMemBlock
static uint8 BL_state, BL_out, BL_SDA, BL_SCL;
static uint8 BL_addr, BL_word, BL_latch, BL_shift;

static SFORMAT x24C0xStateRegs[] =
{
	{ &BL_addr,1,"ADDR" },
	{ &BL_word,1,"WORD" },
	{ &BL_latch,1,"LATC" },
	{ &BL_shift,1,"BITC" },
	{ &BL_SDA,1,"SDA" },
	{ &BL_SCL,1,"SCL" },
	{ &BL_out,1,"OUT" },
//	{ &BL_oe,1,"OE" },
	{ &BL_state,1,"STAT" },
	{ 0 }
};

static void x24C0x_init(void)
{
	BL_addr = BL_word = BL_latch = BL_shift = BL_SDA = BL_SCL = 0;
	BL_state = BLMode_Idle;
	BL_out = 1;
}

static DECLFW(x24C0x_write)
{
	uint8		SDA = V>>6&1;
	uint8		SCL = V>>5&1;
(void)A;
	if (BL_SCL && SCL)
	{
		if (BL_SDA && !SDA)
		{ // start
			BL_state = BLMode_Load;
			BL_shift = BL_addr = 0;
		} else if (!BL_SDA && SDA) // stop
			BL_state = BLMode_Idle;
	} else if (!BL_SCL && SCL)
	{ // rising
		switch(BL_state)
		{
		case BLMode_Load:
			if (BL_shift<7)
			{
				BL_addr<<= 1;
				BL_addr|= SDA;
			} else
			{
				if (!is24C02)
					BL_word = BL_addr;
				if (SDA) // read
					BL_state = BLMode_Read;
				else if (is24C02)
					BL_state = BLMode_Addr;
				else
					BL_state = BLMode_Write;
			}
			++BL_shift;
			break;
		case BLMode_Addr:
			if (BL_shift==8) // ack
				BL_word = BL_out = 0;
			else
			{ // address input
				BL_word<<= 1;
				BL_word|= SDA;
				if(BL_shift==16)
				{ // end address input
					BL_shift = 7;
					BL_state = BLMode_Write;
				}
			}
			++BL_shift;
			break;
		case BLMode_Read:
			if (BL_shift==8)
			{ // ack
				BL_latch = EEPROM[BL_word];
				BL_shift = 0;
				BL_out = 0;
			} else
			{ // advance
				BL_latch<<= 1;
				if (++BL_shift==8)
					++BL_word;
				BL_out = BL_latch>>7;
			}
			break;
		case BLMode_Write:
			if (BL_shift==8) // ack
				BL_latch = BL_shift = BL_out = 0;
			else
			{ // read data
				BL_latch<<= 1;
				BL_latch|= SDA;
				if (++BL_shift==8)
				{
					EEPROM[BL_word] = BL_latch;
					++BL_word;
				}
			}
			break;
		case BLMode_Ack:
			BL_out = 0;
		}
	}
	BL_SDA = SDA;
	BL_SCL = SCL;
}

static uint8 x24C0x_read(void)
{
	return BL_out<<4;
}

//

static void BLSync(int v)
{
	const static uint8	mirt[] = {MI_V,MI_H,MI_0,MI_1};

	if (is153)
	{
		v = DReg[0]<<4&0x10;
		setchr8(0);
		setprg16(0x8000,(DReg[8]&15)|v);
		setprg16(0xC000,15|v);
	} else
	{
		for(v=0; v<8; ++v)
			setchr1(v<<10,DReg[v]);
		setprg16(0x8000,DReg[8]);
		setprg16(0xC000,~0);
	}
	setmirror(mirt[DReg[9]&3]);
}

static DECLFW(BandaiWrite)
{
	switch(A&=15)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	case 9:
		DReg[A] = V;
		BLSync(0);
		break;
	case 10: X65xx_IRQEnd(FCEU_IQEXT); IRQa = V&1; IRQCount = IRQLatch; break;
	case 11: IRQLatch&= 0xFF00; IRQLatch|= V; break;
	case 12: IRQLatch&= 0xFF; IRQLatch|= V<<8; break;
	}
}

static DECLFR(BandaiRead)
{
	return (A>>8&0xEF)|x24C0x_read();
}

static void FP_FASTAPASS(1) BandaiIRQHook(int a)
{
	if (IRQa)
		if ((IRQCount-=a)<0)
		{
			X65xx_IRQBegin(FCEU_IQEXT);
			IRQa = 0;
			IRQCount =~0;
		}
}

static void BandaiPower(void)
{
	IRQa = 0;
	x24C0x_init();
	BLSync(0);
	SetReadHandler(0x6000,0x7FFF,BandaiRead);
	SetCartRead(0x8000,0xFFFF);
	SetWriteHandler(0x6000,0xFFFF,BandaiWrite);
	SetWriteHandlerI(0x10,0x600D,0xFFFD,x24C0x_write);
}

void Bandai_LZ93D50_24C02_Init(CartInfo *info)
{
	is24C02 = 1;
	is153 = 0;
	info->Power = BandaiPower;
	MapIRQHook = BandaiIRQHook;

	info->battery = 1;
	info->SaveGame[0] = EEPROM;
	info->SaveGameLen[0] = 256;
	AddExState(EEPROM,256,0,"DATA");

	GameStateRestore = BLSync;
	AddExState(&x24C0xStateRegs,~0,0,0);
	AddExState(&StateRegs,~0,0,0);
}

void Bandai_LZ93D50_24C01_Init(CartInfo *info)
{
	is24C02 = 0;
	is153 = 0;
	info->Power = BandaiPower;
	MapIRQHook = BandaiIRQHook;

	info->battery = 1;
	info->SaveGame[0] = EEPROM;
	info->SaveGameLen[0] = 128;
	AddExState(EEPROM,128,0,"DATA");

	GameStateRestore = BLSync;
	AddExState(&x24C0xStateRegs,~0,0,0);
	AddExState(&StateRegs,~0,0,0);
}

/* ---- mapper 153: Bandai FCG-1/2, BA-JUMP 2 */

// Famicom jump 2:
// 0-7: Lower bit of data selects which 256KB PRG block is in use.
// This seems to be a hack on the developers' part, so I'll make emulation
// of it a hack(I think the current PRG block would depend on whatever the
// lowest bit of the CHR bank switching register that corresponds to the
// last CHR address read).

static void M153Power(void)
{
	BLSync(0);
	setprg8r(0x10,0x6000,0);
	SetCartRead(0x6000,0xFFFF);
	SetWriteHandler(0x6000,0x7FFF,CartBW);
	SetWriteHandler(0x8000,0xFFFF,BandaiWrite);
	FCEU_CheatSetRAM(WRAMSIZE>>10,0x6000,WRAM);
}


static void M153Close(void)
{
	if (WRAM)
		FCEU_free(WRAM);
	WRAM = NULL;
}

void Mapper153_Init(CartInfo *info)
{
	is153 = 1;
	info->Power = M153Power;
	info->Close = M153Close;
	MapIRQHook = BandaiIRQHook;

	WRAMSIZE = 8192;
	WRAM = (uint8*)FCEU_gmalloc(WRAMSIZE);
	SetupCartPRGMapping(0x10,WRAM,WRAMSIZE,1);
	AddExState(WRAM,WRAMSIZE,0,"WRAM");

	if (info->battery) {
		info->SaveGame[0] = WRAM;
		info->SaveGameLen[0] = WRAMSIZE;
	}

	GameStateRestore = BLSync;
	AddExState(&StateRegs,~0,0,0);
}

/* ---- mapper 188: Karaoke Studio */

/* * needs to be updated * */

static DECLFR(MBandaiRead)
{
(void)A;
	return DReg[7];
}

static void KaraokeSync(int v)
{
(void)v;
	if (DReg[0])
		setprg16(0x8000,(DReg[0]&7)|(~DReg[0]>>1&8));
	else
		setprg16(0x8000,7); /* +(size>>18) */
}

static DECLFW(M188Write)
{
(void)A;
	DReg[0] = V;
	KaraokeSync(0);
}

static void KaraokePower(void)
{
	DReg[0] = 0;
	setchr8(0);
	setprg8(0xC000,14);
	setprg8(0xE000,15);
	KaraokeSync(0);
	SetReadHandler(0x6000,0x7FFF,MBandaiRead);
	SetCartRead(0x8000,0xFFFF);
	SetWriteHandler(0x8000,0xFFFF,M188Write);
}

void Bandai_Karaoke_Init(CartInfo *info)
{
	DReg[7] = 3; /* (input_buttons&7)^3 */
	info->Power = KaraokePower;
	GameStateRestore = KaraokeSync;
	AddExState(DReg,1,0,"DREG");
}
