/*
 boards/JYCompany.c
  mappers 90, 209, 211

 * MacFCEU 0.9a2
 *
 * Copyright notice for this file:
 *  90.c Copyright (C) 2002 Xodnizel
 *  90.c Copyright (C) 2005 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "mapinc.h"
#include "JYCompany.h"

static uint8	jycb;

static uint8	IRQMode, IRQModeHi, IRQModeLo, premask;
static uint8	IRQPre, IRQPreSize;
static uint8	IRQCount;
static uint8	IRQXOR;
static uint8	IRQa;

static uint8	mul[2];
static uint8	regie;

static uint8	tkcom[4];
static uint8	prgb[4];
static uint8	chrlow[8], chrhigh[8], chr[2];

static uint16	names[4];
static uint8	resetctr, resetinit;

static SFORMAT Tek_StateRegs[] =
{
	{&IRQMode,1,"IRQM"},
	{&IRQPre,1,"IRQP"},
	{&IRQPreSize,1,"IRQR"},
	{&IRQCount,1,"IRQC"},
	{&IRQXOR,1,"IRQX"},
	{&IRQa,1,"IRQA"},
	{mul,2,"MUL"},
	{&regie,1,"REGI"},
	{tkcom,4,"TKCO"},
	{prgb,4,"PRGB"},
	{chr,2,"CLTC"},
	{chrlow,4,"CHRL"},
	{chrhigh,8,"CHRH"},
	{&names[0],2|FCEUSTATE_RLSB,"NMS0"},
	{&names[1],2|FCEUSTATE_RLSB,"NMS1"},
	{&names[2],2|FCEUSTATE_RLSB,"NMS2"},
	{&names[3],2|FCEUSTATE_RLSB,"NMS3"},
	{&resetctr,1,"TEKR"},
	{0}
};

static void mira(void)
{
	if ((tkcom[0]&jycb)|resetinit)
	{
		int x;

		if (tkcom[0]&0x40) // Name tables are ROM-only
			for(x=0; x<4; ++x)
				setntamem(CHRptr[0]+(((names[x])&CHRmask1[0])<<10),0,x);
		else // Name tables can be RAM or ROM.
			for(x=0; x<4; ++x)
				if ((tkcom[1]&0x80)==(names[x]&0x80)) // RAM selected.
					setntamem(NTARAM+((names[x]&1)<<10),1,x);
				else
					setntamem(CHRptr[0]+(((names[x])&CHRmask1[0])<<10),0,x);
	} else
		setmirror(mirVH01[tkcom[1]&3]);
}

static void tekprom(void)
{
	uint32		bankmode = (tkcom[3]&6)<<5;

	switch(tkcom[0]&7)
	{
	case 0:
		if (tkcom[0]&0x80)
			setprg8(0x6000,(((prgb[3]<<2)+3)&63)|bankmode);
		setprg32(0x8000,15|(tkcom[3]&6)<<3);
		break;
	case 1: if (tkcom[0]&0x80)
			setprg8(0x6000,(((prgb[3]<<1)+1)&63)|bankmode);
		setprg16(0x8000,(prgb[1]&0x1F)|(tkcom[3]&6)<<4);
		setprg16(0xC000,31|(tkcom[3]&6)<<4);
		break;
	case 3: // bit reversion
	case 2: if (tkcom[0]&0x80)
			setprg8(0x6000,(prgb[3]&63)|bankmode);
		setprg8(0x8000,(prgb[0]&63)|bankmode);
		setprg8(0xA000,(prgb[1]&63)|bankmode);
		setprg8(0xC000,(prgb[2]&63)|bankmode);
		setprg8(0xE000,63|bankmode);
		break;
	case 4: if (tkcom[0]&0x80)
			setprg8(0x6000,(((prgb[3]<<2)+3)&63)|bankmode);
		setprg32(0x8000,(prgb[3]&0x0F)|(tkcom[3]&6)<<3);
		break;
	case 5:	if (tkcom[0]&0x80)
			setprg8(0x6000,(((prgb[3]<<1)+1)&63)|bankmode);
		setprg16(0x8000,(prgb[1]&0x1F)|(tkcom[3]&6)<<4);
		setprg16(0xC000,(prgb[3]&0x1F)|(tkcom[3]&6)<<4);
		break;
	case 7: // bit reversion
	case 6:	if (tkcom[0]&0x80)
			setprg8(0x6000,(prgb[3]&63)|bankmode);
		setprg8(0x8000,(prgb[0]&63)|bankmode);
		setprg8(0xA000,(prgb[1]&63)|bankmode);
		setprg8(0xC000,(prgb[2]&63)|bankmode);
		setprg8(0xE000,(prgb[3]&63)|bankmode);
	}
}

static void tekvrom(void)
{
	int		x, bank = 0, mask = 0xFFFF;

	if (!(tkcom[3]&0x20))
	{
		bank = (tkcom[3]&1)|(tkcom[3]&0x18)>>2;
		switch(tkcom[0]&0x18)
		{
		case 0x00: bank<<= 5; mask = 31; break;
		case 0x08: bank<<= 6; mask = 63; break;
		case 0x10: bank<<= 7; mask = 0x7F; break;
		case 0x18: bank<<= 8; mask = 0xFF; break;
		}
	}
	switch(tkcom[0]&0x18)
	{
	case 0x00:      // 8KB
		setchr8(((chrlow[0]|chrhigh[0]<<8)&mask)|bank);
		break;
	case 0x08:      // 4KB
		setchr4(0x0000,((chrlow[chr[0]]|chrhigh[chr[0]]<<8)&mask)|bank);
		setchr4(0x1000,((chrlow[chr[1]]|chrhigh[chr[1]]<<8)&mask)|bank);
		break;
	case 0x10:      // 2KB
		for(x=0; x<8; x+=2)
			setchr2(x<<10,((chrlow[x]|(chrhigh[x]<<8))&mask)|bank);
		break;
	case 0x18:      // 1KB
		for(x=0; x<8; ++x)
			setchr1(x<<10,((chrlow[x]|chrhigh[x]<<8)&mask)|bank);
	}
}

static DECLFW(JYCTekWrite)
{
	switch(A&0x5C03)
	{
	case 0x5800: mul[0] = V; break;
	case 0x5801: mul[1] = V; break;
	case 0x5803: regie = V;
	}
}

static DECLFR(JYCTekRead)
{
	switch(A&0x5C03)
	{
	case 0x5800: return mul[0]*mul[1];
	case 0x5801: return mul[0]*mul[1]>>8;
	case 0x5803: return regie;

	default: return resetctr;
	}
	return 0xFF;
}

static DECLFW(JYCPRGWrite)
{
	prgb[A&3] = V;
	tekprom();
}

static DECLFW(JYCCHRlowWrite)
{
	chrlow[A&7] = V;
	tekvrom();
}

static DECLFW(JYCCHRhiWrite)
{
	chrhigh[A&7] = V;
	tekvrom();
}

static DECLFW(JYCNTWrite)
{
	if (A&4)
	{
		names[A&3]&= 0xFF;
		names[A&3]|= V<<8;
	} else
	{
		names[A&3]&= 0xFF00;
		names[A&3]|= V;
	}
	mira();
}

static DECLFW(JYCIRQWrite)
{
	switch(A&7)
	{
	case 0:
		IRQa = V&1;
		if (!(V&1))
			XFrame.intlow&=~FCEU_IQEXT;
		break;
	case 1: IRQMode = V, IRQModeHi = V>>6, IRQModeLo = V&3;
		premask = IRQMode&4 ? 7:0xFF;
		break;
	case 2:	IRQa = 0;
		XFrame.intlow&=~FCEU_IQEXT;
		break;
	case 3:	IRQa = 1;
		break;
	case 4:	IRQPre = V^IRQXOR;
		break;
	case 5:	IRQCount = V^IRQXOR;
		break;
	case 6:	IRQXOR = V;
		break;
	case 7: IRQPreSize = V;
	}
}

static DECLFW(JYCModeWrite)
{
	tkcom[A&3] = V;
	tekprom();
	tekvrom();
	mira();
}

static DECLFW(JYCDummyWrite)
{
(void)A; (void)V;
}

static void ClockCounter(void)
{
	if (IRQModeHi==1) // count up
	{
		if (!(++IRQPre&premask))
			if (!++IRQCount && IRQa)
				XFrame.intlow|= FCEU_IQEXT;
	} else if (IRQModeHi==2) // count down
		if (!(~(--IRQPre)&premask))
			if (--IRQCount==0xFF && IRQa)
				XFrame.intlow|= FCEU_IQEXT;
}

static void CPUWrap(int a)
{
	int		x;

	if (!IRQModeLo)
		for(x=0; x<a; ++x)
			ClockCounter();
}

static void SLWrap(void)
{
	int		x;

	if (IRQModeLo==1)
		for(x=0; x<8; ++x)
			ClockCounter();
}

static uint32	lastread;

static void JYCPPU(uint32 A)
{
	if (IRQModeLo==2)
	{
		if (lastread!=A)
		{
			ClockCounter();
			ClockCounter();
		}
		lastread = A;
	}
	if (jycb)
	{
		uint8		l, h = A>>8;

		if (h<0x20 && !(~h&15))
		{
			l = A&0xF0;
			if (l==0xD0)
			{
				chr[(h&0x10)>>4] = (h&0x10)>>2;
				tekvrom();
			} else if (l==0xE0)
			{
				chr[(h&0x10)>>4] = ((h&0x10)>>2)+2;
				tekvrom();
			}
		}
	} else
	{
		chr[0] = 0;
		chr[1] = 4;
	}
}

static void JYCReset()
{
	resetctr+= 0x40;
	resetctr&= 0xC0;
	memset(tkcom,0,sizeof(tkcom));
	memset(prgb,0xFF,sizeof(prgb));
	tekprom();
	tekvrom();
}

static void JYCRestore(int v)
{
(void)v;
	IRQModeHi = IRQMode>>6, IRQModeLo = IRQMode&3;
	premask = IRQMode&4 ? 7:0xFF;
	tekprom();
	tekvrom();
	mira();
}

static void JYCPower(void)
{
	SetWriteHandler(0x5000,0x5FFF,JYCTekWrite);
	SetWriteHandler(0x8000,0x8FF0,JYCPRGWrite);
	SetWriteHandler(0x9000,0x9FFF,JYCCHRlowWrite);
	SetWriteHandler(0xA000,0xAFFF,JYCCHRhiWrite);
	SetWriteHandler(0xB000,0xBFFF,JYCNTWrite);
	SetWriteHandler(0xC000,0xCFFF,JYCIRQWrite);
	SetWriteHandler(0xD000,0xD5FF,JYCModeWrite);
	SetWriteHandler(0xE000,0xFFFF,JYCDummyWrite);

	SetReadHandler(0x5000,0x5FFF,JYCTekRead);
	SetReadHandler(0x6000,0xFFFF,CartBR);

	mul[0] = mul[1] = regie = 0xFF;

	memset(tkcom,0,sizeof(tkcom));
	memset(prgb,0xFF,sizeof(prgb));
	memset(chrlow,0xFF,sizeof(chrlow));
	memset(chrhigh,0xFF,sizeof(chrhigh));
	memset(names,0,sizeof(names));

	resetctr = resetinit;

	tekprom();
	tekvrom();
}

/* ---- mapper 90: J.Y.Company type A */

void JYCA_Init(CartInfo *info)
{
	resetinit = jycb = 0;
	info->Reset = JYCReset;
	info->Power = JYCPower;
	PPU_hook = JYCPPU;
	MapIRQHook = CPUWrap;
	GameHBIRQHook2 = SLWrap;
	GameStateRestore = JYCRestore;
	AddExState(Tek_StateRegs,~0,0,0);
}

/* ---- mapper 209 */
/* compicated hardware with decribed above features disabled by default */

void JYCB_Init(CartInfo *info)
{
	JYCA_Init(info);
	jycb = 0x20;
}

/* ---- mapper 211 */
/* same as mapper 209 but with forced nametable control */

void JYCC_Init(CartInfo *info)
{
	JYCA_Init(info); //B?
	resetinit = 0xC0;
}
