/*
 board/MMC1.c
  covers mappers 1, 155

 * MacFCEU 0.9a2
 *
 * Copyright notice for this file:
 *  mmc1.c Copyright (C) 1998 BERO
 *  mmc1.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC1.h"
#include "../cheat.h"

/* ---- constants */

#define SerialDelay	2

/* ---- globals */

uint8		MMC1DReg[4];

static uint8	*WRAM;
static uint8	Buffer, BufferShift;
static uint8	is155, MMC1opts;
static uint64	MMC1SerialNTime;

SFORMAT MMC1_StateRegs[] =
{
	{MMC1DReg,4,"DREG"},
	{&MMC1SerialNTime,8|FCEUSTATE_RLSB,"LRST"},
	{&Buffer,1,"M1BF"},
	{&BufferShift,1,"M1MR"},
	{0}
};

void (*MMC1CHRHook4)(uint32 A,
	uint8 V);
void (*MMC1PRGHook16)(uint32 A,
	uint8 V);

/* ---- code */

/* mmc1_a12 indicates that WRAM enabling can be affected by PPU VRAM */
/* exact nature unknown */

static DECLFW(MBWRAM)
{
extern uint32 RefreshAddr;
	if (MMC1PRG0&MMC1WRAMDisable && !is155)
		return;
	//FCEUPPU_LineUpdate();
	//if (!(RefreshAddr&0x6000))
	//	return;
	IOPage[A>>11][A] = V; // WRAM is enabled.
}

static DECLFR(MAWRAM)
{
	if (MMC1PRG0&MMC1WRAMDisable && !is155)
		return A>>8; //XFrame.DB -- WRAM is disabled
	//FCEUPPU_LineUpdate();
	//if (!(RefreshAddr&0x6000))
	//	return A>>8;
	return IOPage[A>>11][A];
}

void MMC1FixCHR(void)
{
	if (MMC1opts&4)
		setprg8r(0x10,0x6000,MMC1Ctl&MMC1BankMode ? MMC1Bank>>4&1:MMC1Bank>>3&1);
	if (MMC1Ctl&MMC1BankMode)
	{
		MMC1CHRHook4(0x0000,MMC1Bank);
		MMC1CHRHook4(0x1000,MMC1CHR1);
	} else
	{
		MMC1CHRHook4(0x0000,MMC1Bank&0xFE);
		MMC1CHRHook4(0x1000,MMC1Bank|1);
	}
}

void MMC1FixPRG(void)
{
	uint8		offs = MMC1Bank&MMC1BankSwap4;

	if (MMC1opts&0x80)
		return;
	switch(MMC1Ctl&0xC)
	{
	case 0x0:
	case 0x4:
		MMC1PRGHook16(0x8000,offs+=MMC1PRG0&~1);
		MMC1PRGHook16(0xC000,++offs);
		break;
	case 0x8:
		MMC1PRGHook16(0xC000,MMC1PRG0+offs);
		MMC1PRGHook16(0x8000,offs);
		break;
	case 0xC:
		MMC1PRGHook16(0x8000,MMC1PRG0+offs);
		MMC1PRGHook16(0xC000,0xF+offs);
	}
}

void MMC1FixMir(void)
{
	setmirror(mir01VH[MMC1Ctl&MMC1Mirror]);
}

void MMC1SerialClr(void)
{
	MMC1Ctl|= 0xC;
	BufferShift = Buffer = 0;
}

DECLFW(MMC1_write)
{
	uint64		t; /* for 64 bit backward-compatibility */

	/* The MMC1 is busy so ignore the write. */
	/* As of version FCE Ultra 0.81, the timestamp is only
	   increased before each instruction is executed(in other words
	   precision isn't that great), but this should still work to
	   deal with 2 writes in a row from a single RMW instruction.
	*/
	//if(lreset+2>timestampbase+timestamp) return;

	t = timestampbase; (void)uint64addu32(t,timestamp);
	if (int64gt(MMC1SerialNTime,t))
		return;
	if (V&MMC1SerialReset)
	{
		MMC1Ctl = 0;
		MMC1SerialClr();
		MMC1FixPRG();
		(void)uint64addu32(t,SerialDelay);
		MMC1SerialNTime = t;
		return;
	}

	Buffer|= (V&1)<<BufferShift++;

	if (BufferShift!=5)
		return;
	A = A>>13&3;
	MMC1DReg[A] = Buffer;
	BufferShift = Buffer = 0;

	//switch(n)
	if (!A)
		MMC1FixMir();
	if (A!=3)
		MMC1FixCHR();
	if (A!=2)
		MMC1FixPRG();
}

void MMC1_Restore(int v)
{
(void)v; /* timestampbase _is_ stored in MacFCEU state files */
	MMC1FixMir();
	MMC1FixCHR();
	MMC1FixPRG();
}

void GenMMC1Reset(void)
{
	_int64xset(MMC1SerialNTime,SerialDelay);
	Buffer = BufferShift = 0;
}

static int DetectMMC1WRAMSize(uint32 crc32)
{
	switch(crc32)
	{
	case 0xc6182024: /* Romance of the 3 Kingdoms */
	case 0x2225c20f: /* Genghis Khan */
	case 0x4642dda6: /* Nobunaga's Ambition */
	case 0x29449ba9: /* ""	"" (J) */
	case 0x2b11e0b0: /* ""	"" (J) */
		FCEU_printf(
" >8KB external WRAM present.  Use UNIF if you hack the ROM image.\n");
		return 16;

	default: return 8;
	}
}

void GenMMC1Power()
{
	if (MMC1opts&1)
	{
		FCEU_CheatSetRAM(8,0x6000,WRAM);
		if (MMC1opts&4)
			dwmemset(WRAM,0,8192);
		else if (!(MMC1opts&2))
			dwmemset(WRAM,0,8192);
	}

	SetWriteHandler(0x8000,0xFFFF,MMC1_write);
	SetCartRead(0x8000,0xFFFF);

	if (MMC1opts&1)
	{
		SetReadHandler(0x6000,0x7FFF,MAWRAM);
		SetWriteHandler(0x6000,0x7FFF,MBWRAM);
		setprg8r(0x10,0x6000,0);
	}

	GenMMC1Reset();

	MMC1Ctl = MMC1CtlPower;
	MMC1Bank = 0;
	MMC1CHR1 = 0; /* Should this be something other than 0? */
	MMC1PRG0 = 0; /* MMC1WRAMDisable for rev. B3 */

	MMC1FixMir();
	MMC1FixCHR();
	MMC1FixPRG();
	MMC1opts|= MMC1opts<<4&0x80;
}

void GenMMC1Close()
{
	DeleteCHRRAM(0);

	if (WRAM)
		FCEU_free(WRAM);
	WRAM = 0;
}

static void defprg16(uint32 A,
	uint8 V)
{
	setprg16(A,V);
}

static void defchr4(uint32 A,
	uint8 V)
{
	setchr4(A,V);
}

void GenMMC1_Init(CartInfo *info,
	int prg,
	int chr,
	int wram,
	int battery)
{
	is155 = MMC1opts = 0;

	info->Power = GenMMC1Power;
	info->Reset = GenMMC1Reset;
	info->Close = GenMMC1Close;
	MMC1PRGHook16 = defprg16;
	MMC1CHRHook4 = defchr4;
	PRGmask16[0]&= (prg>>14)-1;
	CHRmask4[0]&= (chr>>12)-1;
	CHRmask8[0]&= (chr>>13)-1;

	if (wram)
	{
		WRAM = (uint8*)FCEU_gmalloc(wram*1024);
		MMC1opts|= 1;
		if (wram>8)
			MMC1opts|= 4;
		SetupCartPRGMapping(0x10,WRAM,wram*1024,1);
		AddExState(WRAM,wram*1024,0,"WRAM");

		if (battery)
		{
			MMC1opts|= 2;
			info->SaveGame[0] = WRAM+(MMC1opts&4 ? 8192:0);
			info->SaveGameLen[0] = 8192;
		}
	}

	if (!chr && NewCHRRAM(0,8192))
		AddExState(CHRptr[0],CHRsize[0],0,"CHRR");

	GameStateRestore = MMC1_Restore;
	AddExState(MMC1_StateRegs,~0,0,0);
}

/* ---- mapper 1 */

void SxROM_Init(CartInfo *info)
{
	int		ws = DetectMMC1WRAMSize(info->CRC32);

	GenMMC1_Init(info,512,256,ws,info->battery);
}

/* ---- mapper 155: MMC1A */
/* Same as mapper 1, without respect for WRAM enable bit. */

void MMC1A_Init(CartInfo *info)
{
	GenMMC1_Init(info,512,256,8,info->battery);
	is155 = 1;
}

/* ---- generic */

void SAROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,128,64,8,info->battery);
}

void SBROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,128,64,0,0);
}

void SCROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,128,128,0,0);
}

void SEROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,32,64,0,0);
	MMC1opts|= 8;
}

/* P128/256, C<=64, W<=8 */
void SFROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,64,0,info->battery);
}

void SGROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,0,0,0);
}

void SHROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,32,128,0,0);
	MMC1opts|= 8;
}

void SJROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,64,8,info->battery);
}

void SKROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,64,8,info->battery);
}

void SLROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,128,0,0);
}

void SL1ROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,128,128,0,0);
}

void SL2ROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,256,0,0);
}

void SNROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,0,8,info->battery);
}

void SOROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,256,0,16,info->battery);
}

void SUROM_Init(CartInfo *info)
{
	GenMMC1_Init(info,512,0,16,info->battery);
}
