/*
 boards/MMC3and6.c
  MMC3 A/B/C base, and MMC6(HKROM)

 * MacFCEU 0.9a2
 *
 * Copyright notice for this file:
 *  mmc3.c Copyright (C) 1998 BERO
 *  mmc3.c Copyright (C) 2003 Xodnizel
 *  Mapper 12 code Copyright (C) 2003 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC3and6.h"
#include "../cheat.h"

uint8		MMC3cmd, MMC3IRQReload, MMC3resetmode, MMC3A000B, MMC3WRAM_ctl;
uint8		MMC3IRQCount, MMC3IRQLatch, MMC3IRQa;
uint8		MMC3Rev; /* rev. B/C */

static uint8	MMC6WA;

uint8		MMC3DReg[16];

void (*MMC3pwrap)(uint32 A,
	uint8 V);
void (*MMC3cwrap)(uint32 A,
	uint8 V);
void (*MMC3mwrap)(uint8 V);

static uint8	PPUCHRBus;
static uint8	TKSMIR[8];

uint8		*MMC3WRAM;
static uint32	WRAMsize;

static uint8	MMC3opts = 0; //bit #1: battery backup (don't zero out)
static uint8	isMMC6;

SFORMAT MMC3_StateRegs[] =
{
	{MMC3DReg,8,"REGS"},
	{&MMC3resetmode,1,"RMOD"},
	{&MMC3cmd,1,"CMD"},
	{&MMC3A000B,1,"A000"},
	{&MMC3WRAM_ctl,1,"A001"},
	{&MMC3IRQReload,1,"IRQR"},
	{&MMC3IRQCount,1,"IRQC"},
	{&MMC3IRQLatch,1,"IRQL"},
	{&MMC3IRQa,1,"IRQA"},
	{0}
};

void FixMMC3PRG(int V)
{
	if (V&0x40)
	{
		MMC3pwrap(0x8000,~1);
		MMC3pwrap(0xC000,MMC3DReg[6]);
	} else
	{
		MMC3pwrap(0x8000,MMC3DReg[6]);
		MMC3pwrap(0xC000,~1);
	}
	MMC3pwrap(0xA000,MMC3DReg[7]);
	MMC3pwrap(0xE000,~0);
}

void FixMMC3CHR(int V)
{
	int		cbase = V<<5&0x1000;

	MMC3cwrap(cbase,MMC3DReg[0]&~1);
	MMC3cwrap(cbase^0x400,MMC3DReg[0]|1);
	MMC3cwrap(cbase^0x800,MMC3DReg[1]&~1);
	MMC3cwrap(cbase^0xC00,MMC3DReg[1]|1);

	MMC3cwrap(cbase^0x1000,MMC3DReg[2]);
	MMC3cwrap(cbase^0x1400,MMC3DReg[3]);
	MMC3cwrap(cbase^0x1800,MMC3DReg[4]);
	MMC3cwrap(cbase^0x1C00,MMC3DReg[5]);
}

DECLFW(MMC3_8CMDWrite)
{
	int		x;

	if (!(A&1))
	{ /* 0x8000 */
		if ((V^MMC3cmd)&0x40) /* PRG change */
			FixMMC3PRG(V);
		if ((V^MMC3cmd)&0x80)
			FixMMC3CHR(V);
		MMC3cmd = V;
		return;
	}
	A = MMC3cmd<<5&0x1000;
	x = MMC3cmd&7;

	MMC3DReg[x] = V;
	switch(x)
	{
	case 0:
		MMC3cwrap(A^0x000,V&~1);
		MMC3cwrap(A^0x400,V|1);
		break;
	case 1:
		MMC3cwrap(A^0x800,V&~1);
		MMC3cwrap(A^0xC00,V|1);
		break;
	case 2:
		MMC3cwrap(A^0x1000,V);
		break;
	case 3:
		MMC3cwrap(A^0x1400,V);
		break;
	case 4:
		MMC3cwrap(A^0x1800,V);
		break;
	case 5:
		MMC3cwrap(A^0x1C00,V);
		break;
	case 6:
		if (MMC3cmd&0x40)
			MMC3pwrap(0xC000,V);
		else
			MMC3pwrap(0x8000,V);
		break;
	case 7:
		MMC3pwrap(0xA000,V);
	}
}

DECLFW(MMC3_AMWSWrite)
{
	if (!(A&1))
	{ /* 0xA000 */
		if (MMC3mwrap)
			MMC3mwrap(V&1);
	} else /* 0xA001 */
		MMC3WRAM_ctl = V; //WRAM security set
}

static DECLFW(MMC6_A001)
{
(void)A;
	if ((MMC6WA&1)|(MMC3cmd&0x20))
		MMC6WA = V|1;
}

DECLFW(MMC3_CIRQWrite)
{
	if (!(A&1)) /* 0xC000 */
		MMC3IRQLatch = V;
	else /* 0xC001 */
		MMC3IRQReload = 1;
}

DECLFW(MMC3_EIRQWrite)
{
(void)V;
	if ((MMC3IRQa=A&1)==0) /* 0xE000, 0xE001 */
		XFrame.intlow&=~FCEU_IQEXT, MapIRQHook = NULL; //X6502_IRQEnd
}

/* Star Trek: 25th Anniversary (without hacks) requires rev. C */

void ClockMMC3Counter(void)
{
	uint8		hadct;

	if ((hadct=MMC3IRQCount)==0 || MMC3IRQReload)
	{
		if (!MMC3Rev && MMC3IRQReload)
			hadct = 1;
		MMC3IRQCount = MMC3IRQLatch, MMC3IRQReload = 0;
	} else
		--MMC3IRQCount;
	if ((hadct|MMC3Rev) && !MMC3IRQCount && MMC3IRQa)
	{
		XFrame.intlow|= FCEU_IQEXT; //X6502_IRQBegin
		if (!MMC3IRQCount && isMMC6)
			--MMC3IRQCount;
	}
}

static void MMC3_hb_KickMasterHack(void)
{
	if (scanline==238)
		ClockMMC3Counter();
	ClockMMC3Counter();
}

static void MMC3_hb_PALStarWarsHack(void)
{
	if (scanline==240)
		ClockMMC3Counter();
	ClockMMC3Counter();
}

void GenMMC3Restore(int v)
{
(void)v;
	if (MMC3mwrap)
		MMC3mwrap(MMC3A000B&1);
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void GENPWRAP(uint32 A,
	uint8 V)
{
	setprg8(A,V&0x3F);
}

static void GENMWRAP(uint8 V)
{
	MMC3A000B = V, V^= 1;
	setmirror(V);
}

static void GENNOMWRAP(uint8 V)
{
	MMC3A000B = V;
}

static DECLFW(MBWRAM)
{
	if (MMC3WRAM)
		MMC3WRAM[A-0x6000] = V;
}

static DECLFR(MAWRAM)
{
	return MMC3WRAM[A-0x6000];
}

static DECLFW(MBWRAMMMC6)
{
	if (MMC3WRAM && !(~(MMC6WA>>(A>>8&2))&0x30))
		MMC3WRAM[A&0x3FF] = V;
}

static DECLFR(MAWRAMMMC6)
{
	if (!(MMC6WA&0xA0)) /* either bank enabled? */
		return A>>8;
	return MMC6WA>>(A>>8&2)&0x20 ? MMC3WRAM[A&0x3FF]:0;
}

/* ---- general */

void GenMMC3Reset(void)
{
	MMC3IRQCount = MMC3IRQLatch = MMC3IRQa = MMC3cmd = 0;

	MMC3DReg[0] = MMC3DReg[6] = 0;
	MMC3DReg[1] = 2;
	MMC3DReg[2] = 4;
	MMC3DReg[3] = 5;
	MMC3DReg[4] = 6;
	MMC3DReg[5] = 7;
	MMC3DReg[7] = 1;

	FixMMC3PRG(0);
	FixMMC3CHR(0);
}

void GenMMC3Power(void)
{
	/* needs a 16 CPU cc delay...somewhere */
	PPU_A12hook = ClockMMC3Counter;

	MMC3WRAM_ctl = MMC3A000B = 0;
	setmirror(~MMC3A000B&1);

	SetWriteHandler(0x8000,0x9FFF,MMC3_8CMDWrite);
	SetWriteHandler(0xA000,0xBFFF,MMC3_AMWSWrite);
	SetWriteHandler(0xC000,0xDFFF,MMC3_CIRQWrite);
	//if (!isTEROM)
	SetWriteHandler(0xE000,0xFFFF,MMC3_EIRQWrite);
	SetCartRead(0x8000,0xFFFF);

	if (WRAMsize)
	{
		if (isMMC6)
		{ /* MMC6 */
			SetReadHandler(0x7000,0x7FFF,MAWRAMMMC6);
			SetWriteHandler(0x7000,0x7FFF,MBWRAMMMC6);
			SetWriteHandlerI(2,0xA001,0xBFFF,MMC6_A001);
			PRGBank2[14] = 0x800;
			PRGBank2[15] = 0x801;
			MMC6WA = 0;
			FCEU_CheatSetRAM(1,0x7000,MMC3WRAM);
		} else
		{
			SetReadHandler(0x6000,0x6000-1+WRAMsize,MAWRAM);
			SetWriteHandler(0x6000,0x6000-1+WRAMsize,MBWRAM);
			PRGBank2[12] = 0x800;
			PRGBank2[13] = WRAMsize>0x800 ? 0x801:-1;
			PRGBank2[14] = WRAMsize>0x1000 ? 0x802:-1;
			PRGBank2[15] = WRAMsize>0x1800 ? 0x803:-1;
			FCEU_CheatSetRAM(WRAMsize>>10,0x6000,MMC3WRAM);
		}
		if (!(MMC3opts&2))
			dwmemset(MMC3WRAM,0,WRAMsize);
	}

	GenMMC3Reset();

	EraseCHRRAM(0,0);
	EraseCHRRAM(0x10,0);
}

void GenMMC3Close(void)
{
	if (MMC3WRAM)
		{FCEU_free(MMC3WRAM); MMC3WRAM = 0;}
	DeleteCHRRAM(0);
	DeleteCHRRAM(0x10);
	MMC3EXPREGS[5] = 0; /* for debug.c */
	MMC3opts = isMMC6 = 0; /* reset */
}

static void SetIRQHook(CartInfo *info)
{
	void (*hook)(void) = ClockMMC3Counter;

	switch(info->CRC32)
	{
	case 0x5104833E:
		if (info->MD5[15]==0xFA) // Kick Master (1991)
			hook = MMC3_hb_KickMasterHack;
		break;
	case 0x5A6860F1:
	case 0xAE280E20: // Shougi Meikan '92/'93
		hook = MMC3_hb_KickMasterHack;
		break;
	case 0xFCD772EB: // PAL Star Wars, similar problem as Kick Master.
		hook = MMC3_hb_PALStarWarsHack;
		break;
	case 0x326AB3B6:
		if (info->MD5[15]==0x1F) // 8 eyes (1990)
			MMC3Rev = 0;
		break;
	case 0xF312D1DE:
		if (info->MD5[15]==0x6A) // 5.MMC3_rev_A.nes
			MMC3Rev = 0;
		break;
	case 0x52E2B5E0:
		if (info->MD5[15]==0xA3) // Super Mario Bros. 3 (J)
			MMC3Rev = 0;
	
	}
	GameHBIRQHook = hook;
}

static void GENCWRAP(uint32 A,
	uint8 V)
{
	setchr1(A&0xFFFF,V);
}

void GenMMC3_Init(CartInfo *info,
	int prg,
	int chr,
	int wram,
	int battery)
{
	MMC3pwrap = GENPWRAP;
	MMC3cwrap = GENCWRAP;
	MMC3mwrap = GENMWRAP;
	MMC3Rev = 1; /* most boards are B/C */
	MMC3opts = WRAMsize==1024;

	PRGmask8[0]&= (prg>>13)-1;
	CHRmask1[0]&= (chr>>10)-1;
	CHRmask2[0]&= (chr>>11)-1;

	if ((WRAMsize=wram*1024)!=0)
	{
		MMC3WRAM = (uint8*)FCEU_gmalloc(WRAMsize);
		AddExState(MMC3WRAM,WRAMsize,0,"WRAM");
	}
	if (battery)
	{
		MMC3opts|= 2;
		info->SaveGame[0] = MMC3WRAM;
		info->SaveGameLen[0] = WRAMsize;
	}
	if (!chr)
		if (NewCHRRAM(0,8192))
			AddExState(CHRptr[0],CHRsize[0],0,"CHRR");
	info->Power = GenMMC3Power;
	info->Reset = GenMMC3Reset;
	info->Close = GenMMC3Close;
	SetIRQHook(info);
	GameStateRestore = GenMMC3Restore;
	AddExState(MMC3_StateRegs,~0,0,0);
	if (!MMC3Rev)
		FCEU_printf("  MMC3 A board detected\n");
}

/* ---- mapper 12: Rex Soft DBZ5 */

static void M12CW(uint32 A,
	uint8 V)
{
	setchr1(A,(MMC3EXPREGS[A>>12&1]<<8)+V);
}

static DECLFW(M12Write)
{
(void)A;
	MMC3EXPREGS[0] = V&1;
	MMC3EXPREGS[1] = V>>4&1;
}

static void M12Power(void)
{
	MMC3EXPREGS[0] = MMC3EXPREGS[1] = 0;
	GenMMC3Power();
	SetWriteHandler(0x4100,0x5FFF,M12Write);
}

void RexSoftDBZ5_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3cwrap = M12CW;
	info->Power = M12Power;
	AddExState(MMC3EXPREGS,2,0,"EXPR");
}

/* ---- mapper 37: ZZ */

static void M37PW(uint32 A,
	uint8 V)
{
	V&= MMC3EXPREGS[0]<<1|7;
	V|= MMC3EXPREGS[0]<<2&0x10;
	if (!(uint8)(~MMC3EXPREGS[0]&3))
		V|= 8;
	setprg8(A,V);
}

static void M37CW(uint32 A,
	uint8 V)
{
	uint32		NV = V&0x7F;

	NV|= MMC3EXPREGS[0]<<5&0x80;
	setchr1(A,NV);
}

static DECLFW(M37_write)
{
(void)A;
	V&= 7;
	if (MMC3EXPREGS[0]==V)
		return;
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M37Reset(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Reset();
}

static void M37Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M37_write);
}

void CustomPAL_ZZ_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3pwrap = M37PW;
	MMC3cwrap = M37CW;
	info->Reset = M37Reset;
	info->Power = M37Power;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 47: QJ */

static void M47PW(uint32 A,
	uint8 V)
{
	V&= 0xF;
	V|= MMC3EXPREGS[0]<<4;
	setprg8(A,V);
}

static void M47CW(uint32 A,
	uint8 V)
{
	uint32		NV = V&0x7F;

	NV|= MMC3EXPREGS[0]<<7;
	setchr1(A,NV);
}

static DECLFW(M47Write)
{
(void)A;
	MMC3EXPREGS[0] = V&1;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M47Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M47Write);
	SetReadHandler(0x6000,0x7FFF,0);
}

void QJ_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3pwrap = M47PW;
	MMC3cwrap = M47CW;
	info->Power = M47Power;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 115: Kasing (standard) */

static void M115PW(uint32 A,
	uint8 V)
{
	setprg8(A,V);
	if (MMC3EXPREGS[0]&0x80)
		setprg16(0x8000,MMC3EXPREGS[0]&7);
}

static void M115CW(uint32 A,
	uint8 V)
{
	setchr1(A,(uint32)V|(MMC3EXPREGS[1]&1)<<8);
}

static DECLFW(M115Write)
{
	if (A==0x6000)
		MMC3EXPREGS[0] = V;
	else if (A==0x6001)
		MMC3EXPREGS[1] = V;
	FixMMC3PRG(MMC3cmd);
}

static void M115Power(void)
{
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M115Write);
	SetReadHandler(0x6000,0x7FFF,0);
}

void Kasing_STD_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,0,0);
	MMC3cwrap = M115CW;
	MMC3pwrap = M115PW;
	info->Power = M115Power;
	AddExState(MMC3EXPREGS,2,0,"EXPR");
}

/* ---- TxxROM boards */

static void FP_FASTAPASS(1) TKSPPU(uint32 A)
{
	PPUCHRBus = A = A>>10&7;
	setmirror(MI_0+TKSMIR[A]);
}

static void TKSWRAP(uint32 A,
	uint8 V)
{
	TKSMIR[A>>10] = V>>7;
	setchr1(A,V&0x7F);
	if (PPUCHRBus==(A>>10))
		setmirror(MI_0+(V>>7));
}

static int	hackm4 = 0; /* For Karnov, maybe others.  BLAH.  Stupid iNES format.*/

static void M4Power(void)
{
	GenMMC3Power();
	MMC3A000B =~hackm4&1;
	setmirror(hackm4);
}

/* mapper 4 */
void TxROM_Init(CartInfo *info)
{
	int	ws = 8;

	if (info->CRC32==0x93991433 || info->CRC32==0xAF65AA84)
	{
		FCEU_printf(
"Low-G-Man can not work normally in the iNES format.\n"
"This game has been recognized by its CRC32 value,\n"
"and the appropriate changes will be made so it will run.\n"
"If you wish to hack this game, you should use the UNIF format.\n"
"\n");
		ws = 0;
	}

	GenMMC3_Init(info,512,256,ws,info->battery);
	info->Power = M4Power;
	hackm4 = info->mirror;
}

void TBROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,64,32,0,0);
}

void TEROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,32,32,0,0);
	MMC3mwrap = NULL; /* TEROM/TFROM have hardwired mirroring */
}

void TFROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,64,0,0);
	MMC3mwrap = NULL;
}

void TGROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,0,0,0);
}

void TNROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,0,8,0);
}

void TKROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
}

void TLROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,0,0);
}

void TSROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,0);
}

void TLSROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,0);
	MMC3cwrap = TKSWRAP;
	MMC3mwrap = GENNOMWRAP;
	PPU_hook = TKSPPU;
	AddExState(&PPUCHRBus,1,0,"PPUC");
}

/* mapper 118 */
void TKSROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	MMC3cwrap = TKSWRAP;
	MMC3mwrap = GENNOMWRAP;
	PPU_hook = TKSPPU;
	AddExState(&PPUCHRBus,1,0,"PPUC");
}

/* mapper 119 */
static void TQWRAP(uint32 A,
	uint8 V)
{
	setchr1r(V>>2&0x10,A,V&0x3F);
}

void TQROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,64,0,0);
	MMC3cwrap = TQWRAP;
	if (NewCHRRAM(0x10,8192))
		AddExState(CHRptr[0x10],CHRsize[0x10],0,"CHRR");
}

/* MMC6 board */
/* two known titles by Nintendo: Star Tropics 1 and 2 */

void HKROM_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,512,1,info->battery);
	/* on IRQs, behaves like B except counter is changed after zero */
	isMMC6 = 1;
}
