/*
 boards/MMC5.c

 MacFCEU 0.9a2

 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/* None of this code should use any of the iNES bank switching wrappers. */

#include "mapinc.h"
#include "MMC5.h"
#include "../cheat.h"

/* according to Nestopia, it's bit #0 */
#define RAW_DISABLE		0x41

/* ---- globals */

uint8 MMC5Ex;		/* MMC5Hack */
uint8 MMC5VROMMask;
uint8 *MMC5ExRAM;	/* MMC5HackExNTARAMPtr */
uint8 *MMC5VROMPtr;	/* MMC5HackVROMPTR */
uint8 MMC5ExMode;	/* MMC5HackCHRMode */
uint8 MMC5SplitCtrl;	/* MMC5HackSPMode */
uint8 MMC5CHRHigh;	/* MMC50x5130 */
uint8 MMC5SplitY;	/* MMC5HackSPScroll */
uint8 MMC5SplitBank;	/* MMC5HackSPPage */

static uint8 PRGBanks[4];
static uint8 WRAMPage;
static uint16 CHRBanksA[8], CHRBanksB[4];
static uint8 WRAMMaskEnable[2];
uint8 MMC5ABMode; /* A=0, B=1 */

static uint8 IRQScanline, IRQEnable;
static uint8 CHRMode, NTAMirroring, NTFill, ATFill;

static uint8 MMC5IRQR;
static uint8 MMC5LineCounter;
static uint8 mmc5psize, mmc5vsize;
static uint8 mul[2];

static uint32 WRAMSIZE;
static uint8 *WRAM;
static uint8 *MMC5fill;
static uint8 *ExRAM;

static uint8 MMC5WRAMsize;
static uint8 MMC5WRAMIndex[8];

static uint8 MMC5ROMWrProtect[4];
static uint8 MMC5MemIn[5];

static void MMC5CHRA(void);
static void MMC5CHRB(void);

static void (*sfun)(int P);
static void (*psfun)(void);

/* ---- code */

static INLINE void MMC5SPRVROM_BANK2(uint32 A,
	uint32 V)
{
	if (CHRptr[0])
		MMC5SPRVPage[(A)>>10] = MMC5SPRVPage[((A)>>10)+1] = &CHRptr[0][(V&CHRmask2[0])<<11]-(A);
}
static INLINE void MMC5BGVROM_BANK2(uint32 A,
	uint32 V)
{
	if (CHRptr[0])
		MMC5BGVPage[(A)>>10] = MMC5BGVPage[((A)>>10)+1] = &CHRptr[0][(V&CHRmask2[0])<<11]-(A);
}

static INLINE void MMC5SPRVROM_BANK4(uint32 A,
	uint32 V)
{
	if (CHRptr[0])
		MMC5SPRVPage[(A)>>10] = MMC5SPRVPage[((A)>>10)+1] =
			MMC5SPRVPage[((A)>>10)+2] = MMC5SPRVPage[((A)>>10)+3] =
			&CHRptr[0][(V&CHRmask4[0])<<12]-(A);
}
static INLINE void MMC5BGVROM_BANK4(uint32 A,
	uint32 V)
{
	if (CHRptr[0])
		MMC5BGVPage[(A)>>10] = MMC5BGVPage[((A)>>10)+1] =
			MMC5BGVPage[((A)>>10)+2] = MMC5BGVPage[((A)>>10)+3] =
			&CHRptr[0][(V&CHRmask4[0])<<12]-(A);
}

static INLINE void MMC5SPRVROM_BANK8(uint32 V)
{
	if (CHRptr[0])
		MMC5SPRVPage[0] = MMC5SPRVPage[1] = MMC5SPRVPage[2] = MMC5SPRVPage[3] =
			MMC5SPRVPage[4] = MMC5SPRVPage[5] = MMC5SPRVPage[6] = MMC5SPRVPage[7] =
			&CHRptr[0][(V&CHRmask8[0])<<13];
}
static INLINE void MMC5BGVROM_BANK8(uint32 V)
{
	if (CHRptr[0])
		MMC5BGVPage[0] = MMC5BGVPage[1] = MMC5BGVPage[2] = MMC5BGVPage[3] =
			MMC5BGVPage[4] = MMC5BGVPage[5] = MMC5BGVPage[6] = MMC5BGVPage[7] =
			&CHRptr[0][(V&CHRmask8[0])<<13];
}

typedef struct mmc5cartw
{
	uint32		crc32;
	uint8		size, md5f;
} mmc5cartw;

/* FCEUX list */
const mmc5cartw	MMC5CartList[] =
{
	{0x6f4e4312,4,0xde}, /* Aoki Ookami to Shiroki Mejika - Genchou Hishi */
	{0x15fe6d0f,2,0x65}, /* Bandit Kings of Ancient China */
	{0x671f23a8,0,0xf7}, /* Castlevania III - Dracula's Curse (E) */
	{0xcd4e7430,0,0x05}, /* Castlevania III - Dracula's Curse (KC) */
	{0xed2465be,0,0xbf}, /* Castlevania III - Dracula's Curse (U) */
	{0xfe3488d1,2}, /* Daikoukai Jidai */
	{0x0ec6c023,1,0xe7}, /* Gemfire */
	{0x0afb395e,0,0x26}, /* Gun Sight */
	{0x1ced086f,2,0xad}, /* Ishin no Arashi */
	{0x9cbadc25,1,0xc1}, /* Just Breed (J) */
	{0x6396b988,2,0xb0}, /* L'Empereur (J) */
	{0x9c18762b,2,0xf7}, /* L'Empereur (U) */
	{0xb0480ae9,0,0x0d}, /* Laser Invasion */
	{0xb4735fac,0,0x76}, /* Metal Slader Glory */
	{0xf540677b,4,0x25}, /* Nobunaga no Yabou - Bushou Fuuun Roku (J) */
	{0xeee9a682,20,0x5c}, /* Nobunaga no Yabou - Sengoku Gunyuu Den (J) (PRG0) */
	{0xf9b4240f,2}, /* Nobunaga no Yabou - Sengoku Gunyuu Den (J) (PRG1) */
	{0x8ce478db,2,0x60}, /* Nobunaga's Ambition 2 */
	{0xf011e490,4,0xe5}, /* Romance of The Three Kingdoms II */
	{0xbc80fb52,1}, /* Royal Blood */
	{0x184c2124,4}, /* Sangokushi II (J) (PRG0) */
	{0xee8e6553,4}, /* Sangokushi II (J) (PRG1) */
	{0xd532e98f,1,0xf3}, /* Shin 4 Nin Uchi Mahjong - Yakuman Tengoku */
	{0x39f2ce4b,2}, /* Suikoden - Tenmei no Chikai */
	{0xbb7f829a,0,0xbd}, /* Uchuu Keibitai SDF */
	{0xaca15643,2,0x7f}  /* Uncharted Waters */
};

static int DetectMMC5WRAMSize(CartInfo *info)
{
	const mmc5cartw	*l;
	uint32		crc32 = info->CRC32;
	int		x;

	for(l=MMC5CartList,x=0; x<sizeof(MMC5CartList)/sizeof(MMC5CartList[0]); ++x,++l)
		if (crc32==l->crc32 && (info->MD5[0]==l->md5f || !l->md5f))
		{
			if (l->size>1)
				FCEU_printf(
" >8KB external WRAM present.  Use UNIF if you hack the ROM image.\n");
			return l->size*8u;
		}
	return 64;
}

static void BuildWRAMSizeTable(void)
{
	int x;

	for(x=0; x<8; ++x)
		switch(MMC5WRAMsize)
		{
		case 0: MMC5WRAMIndex[x] = 255; break;		 //X,X,X,X,X,X,X,X
		case 1: MMC5WRAMIndex[x] = x>3 ? 255:0; break;	 //0,0,0,0,X,X,X,X
		case 2: MMC5WRAMIndex[x] = x>>2&1; break;	 //0,0,0,0,1,1,1,1
		case 4: MMC5WRAMIndex[x] = x>3 ? 255:x&3; break; //0,1,2,3,X,X,X,X
		case 8: MMC5WRAMIndex[x] = x;			 //0,1,2,3,4,5,6,7
		}
}

static void MMC5CHRA(void)
{
	int		x;

	switch(mmc5vsize&3)
	{
	case 0:	setchr8(CHRBanksA[7]);
		MMC5SPRVROM_BANK8(CHRBanksA[7]);
		break;
	case 1:	setchr4(0x0000,CHRBanksA[3]);
		setchr4(0x1000,CHRBanksA[7]);
		MMC5SPRVROM_BANK4(0x0000,CHRBanksA[3]);
		MMC5SPRVROM_BANK4(0x1000,CHRBanksA[7]);
		break;
	case 2:	setchr2(0x0000,CHRBanksA[1]);
		setchr2(0x0800,CHRBanksA[3]);
		setchr2(0x1000,CHRBanksA[5]);
		setchr2(0x1800,CHRBanksA[7]);
		MMC5SPRVROM_BANK2(0x0000,CHRBanksA[1]);
		MMC5SPRVROM_BANK2(0x0800,CHRBanksA[3]);
		MMC5SPRVROM_BANK2(0x1000,CHRBanksA[5]);
		MMC5SPRVROM_BANK2(0x1800,CHRBanksA[7]);
		break;
	case 3:	for(x=0; x<8; ++x)
		{
			uint32		A = x<<10;

			setchr1(A,CHRBanksA[x]);
			if (CHRptr[0])
				MMC5SPRVPage[A>>10] = &CHRptr[0][(CHRBanksA[x]&CHRmask1[0])<<10]-A;
		}
	}
}

static void MMC5CHRB(void)
{
	int		x;

	switch(mmc5vsize&3)
	{
	case 0:	setchr8(CHRBanksB[3]);
		MMC5BGVROM_BANK8(CHRBanksB[3]);
		break;
	case 1:	setchr4(0x0000,CHRBanksB[3]);
		setchr4(0x1000,CHRBanksB[3]);
		MMC5BGVROM_BANK4(0x0000,CHRBanksB[3]);
		MMC5BGVROM_BANK4(0x1000,CHRBanksB[3]);
		break;
	case 2:	setchr2(0x0000,CHRBanksB[1]);
		setchr2(0x0800,CHRBanksB[3]);
		setchr2(0x1000,CHRBanksB[1]);
		setchr2(0x1800,CHRBanksB[3]);
		MMC5BGVROM_BANK2(0x0000,CHRBanksB[1]);
		MMC5BGVROM_BANK2(0x0800,CHRBanksB[3]);
		MMC5BGVROM_BANK2(0x1000,CHRBanksB[1]);
		MMC5BGVROM_BANK2(0x1800,CHRBanksB[3]);
		break;
	case 3:	for(x=0; x<8; ++x)
		{
			uint32		A = x<<10;

			setchr1(A,CHRBanksB[x&3]);
			if (CHRptr[0])
				MMC5BGVPage[A>>10] = &CHRptr[0][(CHRBanksB[x&3]&CHRmask1[0])<<10]-A;
		}
	}
}

static void MMC5WRAM(uint32 A,
	uint32 V)
{
	V = MMC5WRAMIndex[V&7];
	if (V!=255)
	{
		setprg8r(0x10,A,V);
		FCEU_CheatSetRAM(8,0x6000,WRAM+((V*8192u)&(WRAMSIZE-1)));
		MMC5MemIn[(A-0x6000)>>13] = 1;
	} else
		MMC5MemIn[(A-0x6000)>>13] = 0;
}

static void MMC5PRG(void)
{
	int		x;

	switch(mmc5psize&3)
	{
	case 0:	MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = MMC5ROMWrProtect[2] = MMC5ROMWrProtect[3] = 1;
		setprg32(0x8000,((PRGBanks[1]&0x7F)>>2));
		for(x=0; x<4; ++x)
			MMC5MemIn[1+x] = 1;
		break;
	case 1:	if (PRGBanks[1]&0x80)
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 1;
			setprg16(0x8000,PRGBanks[1]>>1);
			MMC5MemIn[1] = MMC5MemIn[2] = 1;
		} else
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 0;
			MMC5WRAM(0x8000,PRGBanks[1]&(7&0xFE));
			MMC5WRAM(0xA000,(PRGBanks[1]&(7&0xFE))+1);
		}
		MMC5MemIn[3] = MMC5MemIn[4] = 1;
		MMC5ROMWrProtect[2] = MMC5ROMWrProtect[3] = 1;
		setprg16(0xC000,(PRGBanks[3]&0x7F)>>1);
		break;
	case 2:	if (PRGBanks[1]&0x80)
		{
			MMC5MemIn[1] = MMC5MemIn[2] = 1;
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 1;
			setprg16(0x8000,(PRGBanks[1]&0x7F)>>1);
		} else
		{
			MMC5ROMWrProtect[0] = MMC5ROMWrProtect[1] = 0;
			MMC5WRAM(0x8000,PRGBanks[1]&(7&0xFE));
			MMC5WRAM(0xA000,(PRGBanks[1]&(7&0xFE))+1);
		}
		if (PRGBanks[2]&0x80)
		{
			MMC5ROMWrProtect[2] = 1;
			MMC5MemIn[3] = 1;
			setprg8(0xC000,PRGBanks[2]&0x7F);
		} else
		{
			MMC5ROMWrProtect[2] = 0;
			MMC5WRAM(0xC000,PRGBanks[2]&7);
		}
		MMC5MemIn[4] = 1;
		MMC5ROMWrProtect[3] = 1;
		setprg8(0xE000,PRGBanks[3]&0x7F);
		break;
	case 3:	for(x=0; x<3; ++x)
			if (PRGBanks[x]&0x80)
			{
				MMC5ROMWrProtect[x] = 1;
				setprg8(0x8000+(x<<13),PRGBanks[x]&0x7F);
				MMC5MemIn[1+x] = 1;
			} else
			{
				MMC5ROMWrProtect[x] = 0;
				MMC5WRAM(0x8000+(x<<13),PRGBanks[x]&7);
			}
		MMC5MemIn[4] = 1;
		MMC5ROMWrProtect[3] = 1;
		setprg8(0xE000,PRGBanks[3]&0x7F);
	}
}

static DECLFW(MMC5W5100)
{
	uint32		x;

	switch(A)
	{
	case 0x5100:
		mmc5psize = V;
		MMC5PRG();
		break;
	case 0x5101:
		mmc5vsize = V;
		if (!MMC5ABMode)
		{
			MMC5CHRB();
			MMC5CHRA();
		} else
		{
			MMC5CHRA();
			MMC5CHRB();
		}
		break;
	case 0x5102:
		WRAMMaskEnable[0] = V;
		break;
	case 0x5103:
		WRAMMaskEnable[1] = V;
		break;
	case 0x5104:
		CHRMode = V;
		MMC5ExMode = V&3;
		break;
	case 0x5105:
		for(x=0; x<4; ++x)
		{
			switch((V>>(x<<1))&3)
			{
			case 0: PPUNTARAM|= 1<<x; vnapage[x] = NTARAM; break;
			case 1: PPUNTARAM|= 1<<x; vnapage[x] = NTARAM+0x400; break;
			case 2: PPUNTARAM|= 1<<x; vnapage[x] = ExRAM; break;
			case 3: PPUNTARAM&=~(1<<x); vnapage[x] = MMC5fill; break;
			}
		}
		NTAMirroring = V;
		break;
	case 0x5106:
		if (V!=NTFill)
		{
			NTFill = V;
			x = V|V<<8, x|= x<<16;
			FCEU_dwmemset(MMC5fill,x,0x3C0);
		}
		break;
	case 0x5107:
		if (V!=ATFill)
		{
			ATFill = V;
			x = V&3, x|= x<<2, x|= x<<4, x|= x<<8, x|= x<<16;
			FCEU_dwmemset(MMC5fill+0x3C0,x,0x40);
		}
		break;
	case 0x5113:
		WRAMPage = V;
		MMC5WRAM(0x6000,V&7);
		break;
	case 0x5114:
	case 0x5115:
	case 0x5116:
	case 0x5117:
		PRGBanks[A&3] = V;
		MMC5PRG();
		break;
	case 0x5120:
	case 0x5121:
	case 0x5122:
	case 0x5123:
	case 0x5124:
	case 0x5125:
	case 0x5126:
	case 0x5127:
		MMC5ABMode = 0;
		CHRBanksA[A&7] = V|(MMC5CHRHigh&3)<<8;
		MMC5CHRA();
		break;
	case 0x5128:
	case 0x5129:
	case 0x512a:
	case 0x512b:
		MMC5ABMode = 1;
		CHRBanksB[A&3] = V|(MMC5CHRHigh&3)<<8;
		MMC5CHRB();
		break;
	case 0x5130: MMC5CHRHigh = V;
	}
}

static DECLFW(MMC5W5200)
{
	switch(A)
	{
	case 0x5200: MMC5SplitCtrl = V; break;
	case 0x5201: MMC5SplitY = V>>3&0x1F; break;
	case 0x5202: MMC5SplitBank = V&0x3F; break;
	case 0x5203: X65xx_IRQEnd(FCEU_IQEXT); IRQScanline = V; break;
	case 0x5204: X65xx_IRQEnd(FCEU_IQEXT); IRQEnable = V&0x80; break;
	case 0x5205: mul[0] = V; break;
	case 0x5206: mul[1] = V;
	}
}

static DECLFR(MMC5_ReadROMRAM)
{
	if (MMC5MemIn[(A-0x6000)>>13])
		return IOPage[A>>11][A];
	else
		return A>>8;
}

static DECLFW(MMC5_WriteROMRAM)
{
	if (A>=0x8000 && (MMC5ROMWrProtect[(A-0x8000)>>13]))
		return;
	if (MMC5MemIn[(A-0x6000)>>13])
		if (((WRAMMaskEnable[0]&3)|(WRAMMaskEnable[1]&3)<<2)==6)
			IOPage[A>>11][A] = V;
}

static DECLFW(MMC5_ExRAMWr)
{
	if (MMC5ExMode!=3)
		ExRAM[A&0x3FF] = V;
}

static DECLFR(MMC5_ExRAMRd)
{
	return ExRAM[A&0x3FF];
}

static DECLFR(MMC5R5200)
{
	uint8		x;

	switch(A)
	{
	case 0x5204:
		X65xx_IRQEnd(FCEU_IQEXT);
		x = MMC5IRQR;
		if (!fceuindbg)
			MMC5IRQR&= 0x40;
		return x;
	case 0x5205:
		return mul[0]*mul[1];
	case 0x5206:
		return mul[0]*mul[1]>>8;
	}
	return A>>8;
}

static void MMC5Synco(void)
{
	int		x;

	MMC5PRG();
	for(x=0; x<4; ++x)
		switch(NTAMirroring>>(x<<1)&3)
		{
		case 0: PPUNTARAM|= 1<<x; vnapage[x] = NTARAM; break;
		case 1: PPUNTARAM|= 1<<x; vnapage[x] = NTARAM+0x400; break;
		case 2: PPUNTARAM|= 1<<x; vnapage[x] = ExRAM; break;
		case 3: PPUNTARAM&=~(1<<x); vnapage[x] = MMC5fill; break;
		}
	MMC5WRAM(0x6000,WRAMPage&7);
	if (!MMC5ABMode)
	{
		MMC5CHRB();
		MMC5CHRA();
	} else
	{
		MMC5CHRA();
		MMC5CHRB();
	}

	x = NTFill, x|= x<<8, x|= x<<16;
	FCEU_dwmemset(MMC5fill,x,0x3C0);
	x = ATFill&3, x|= x<<2, x|= x<<4, x|= x<<8, x|= x<<16;
	FCEU_dwmemset(MMC5fill+0x3C0,x,0x40);

	MMC5ExMode = CHRMode&3;

	//X65xx_IRQEnd(FCEU_IQEXT);
}

void MMC5_hb(int scanline)
{
	if (!(PPUReg[1]&0x18) || scanline>239)
	{ /* everything IRQ off and zero when rendering disabled or out of frame; 2017.04.23 */
	  /* fixes credits in Metal Slader Glory */
		MMC5IRQR&=~0x40;
		goto rstl;
	}
	if (!(MMC5IRQR&0x40))
	{
		MMC5IRQR|= 0x40;
	rstl:	MMC5IRQR&=~0x80;
		MMC5LineCounter = 0;
		X65xx_IRQEnd(FCEU_IQEXT);
	} else if (++MMC5LineCounter==IRQScanline)
	{
		MMC5IRQR|= 0x80;
		if (IRQEnable&0x80)
			X65xx_IRQBegin(FCEU_IQEXT);
	}

}

static void MMC5_StateRestore(int v)
{
(void)v;
	MMC5Synco();
}

typedef struct
{
	uint16	wl[2];
	uint8	env[2];
	uint8	enable, running;
	uint8	raw;
	uint8	rawcontrol;
	int32	dcount[2];
	int32	vcount[2];
	int32	BC[3];
} MMC5APU;

static MMC5APU	MMC5Sound;

static int32	volume;


void FCEU_MMC5SetVolume(int16 v)
{
#if HAVE_HQ
	if (v<0 && FSettings.soundq)
		v =-v;
#endif
	volume = (v<<(8-2))/25;
}

static void Do5PCM(void)
{
	int32		V, start, end, amp;

	start = MMC5Sound.BC[2], end = (SOUNDTS<<16)/soundtsinc;
	if (end<=start)
		return;
	MMC5Sound.BC[2] = end, V = start;

	if (!(MMC5Sound.rawcontrol&RAW_DISABLE) && (amp=MMC5Sound.raw*volume>>3)!=0)
		do
			FCEUWave[V>>4]+= amp;
		while(++V<end);
}

static int	tal[4] = {1,2,4,6};

static void Do5SQ(int P)
{
	int32		V, start, end, wl;

	start = MMC5Sound.BC[P], end = (SOUNDTS<<16)/soundtsinc;
	if (end<=start)
		return;
	MMC5Sound.BC[P] = end;

	wl = MMC5Sound.wl[P]+1;
	if (wl>7 && MMC5Sound.running&(P+1))
	{
		int32		amp, rthresh;
		int		dc,vc;

		wl<<= 18;
		amp = (MMC5Sound.env[P]&15)*volume;
		rthresh = tal[MMC5Sound.env[P]>>6]; //&0xC0)>>6

		dc = MMC5Sound.dcount[P];
		vc = MMC5Sound.vcount[P];

		for(V=start; V<end; ++V)
		{
			if (dc<rthresh)
				FCEUWave[V>>4]+= amp;
			vc-= nesincsize;
			while(vc<=0)
				vc+= wl, dc = (dc+1)&7;
		}
		MMC5Sound.dcount[P] = dc;
		MMC5Sound.vcount[P] = vc;
	}
}

#ifdef HAVE_HQ
static void Do5SQHQ(int P)
{
	int32		V, wl;

	wl = MMC5Sound.wl[P]+1;
	if (wl>7 && MMC5Sound.running&(P+1))
	{
		int32		end = SOUNDTS;
		int		amp, rthresh, dc, vc;

		wl+= wl;
		amp = (MMC5Sound.env[P]&0xF)*volume;
		rthresh = tal[MMC5Sound.env[P]>>6]; //&0xC0)>>6
		dc = MMC5Sound.dcount[P];
		vc = MMC5Sound.vcount[P];
		end = SOUNDTS;
		for(V=MMC5Sound.BC[P]; V<end; ++V)
		{
			if (dc<rthresh)
				FCEUWave[V]+= amp;
			if (--vc<=0) /* Less than zero when first started. */
				vc = wl, dc = (dc+1)&7;
		}
		MMC5Sound.dcount[P] = dc;
		MMC5Sound.vcount[P] = vc;
	}
	MMC5Sound.BC[P] = SOUNDTS;
}

static void Do5PCMHQ(void)
{
	int32		V, amp;

	if (!(MMC5Sound.rawcontrol&RAW_DISABLE) && (amp=MMC5Sound.raw*volume>>3)!=0)
		for(V=MMC5Sound.BC[2]; V<SOUNDTS; ++V)
			FCEUWave[V]+= amp;
	MMC5Sound.BC[2] = SOUNDTS;
}

static void MMC5RunSoundHQ(void)
{
	Do5SQHQ(0);
	Do5SQHQ(1);
	Do5PCMHQ();
}

void MMC5HiSync(int32 ts)
{
	int32		*d = MMC5Sound.BC;

	*d = d[1] = d[2] = ts;
}
#endif

static void MMC5RunSound(int32 Count)
{
	int32		*d = MMC5Sound.BC;

	Do5SQ(0);
	Do5SQ(1);
	Do5PCM();
	*d = d[1] = d[2] = Count;
}

static DECLFW(MMC5W5)
{
	A&= 0x1F;

	GameExpSound.Fill[SCMMC5] = MMC5RunSound;
#ifdef HAVE_HQ
	GameExpSound.HiFill[SCMMC5] = MMC5RunSoundHQ;
#endif
	switch(A)
	{
	case 16:
		if (psfun)
			psfun();
		MMC5Sound.rawcontrol = V;
		break;
	case 17:
		if (psfun)
			psfun();
		MMC5Sound.raw = V;
		break;
	case 0:
	case 4:
		if (sfun)
			sfun(A>>2);
		MMC5Sound.env[A>>2] = V;
		break;
	case 2:
	case 6:
		if (sfun)
			sfun(A>>2);
		MMC5Sound.wl[A>>2]&=~0xFF;
		MMC5Sound.wl[A>>2]|= V&0xFF;
		break;
	case 3:
	case 7:
		MMC5Sound.wl[A>>2]&=~0x700;
		MMC5Sound.wl[A>>2]|= V<<8&0x700;
		MMC5Sound.running|= 1<<(A>>2);
		break;
	case 0x15:
		if (sfun)
		{
			sfun(0);
			sfun(1);
		}
		MMC5Sound.running&= V;
		MMC5Sound.enable = V;
	}
}

static void Mapper5_ESI(void)
{
	GameExpSound.RChange[SCMMC5] = Mapper5_ESI;
	if (FSettings.SndRate)
	{
	#if HAVE_HQ
		if (FSettings.soundq>=1)
		{
			sfun = Do5SQHQ;
			psfun = Do5PCMHQ;
		} else
	#endif
		{
			sfun = Do5SQ;
			psfun = Do5PCM;
		}
	} else
	{
		sfun = 0;
		psfun = 0;
	}
	memset(MMC5Sound.BC,0,sizeof(MMC5Sound.BC));
	memset(MMC5Sound.vcount,0,sizeof(MMC5Sound.vcount));
#ifdef HAVE_HQ
	GameExpSound.HiSync = MMC5HiSync;
#endif
}

void NSFMMC5_Init(void)
{
	memset(&MMC5Sound,0,sizeof(MMC5Sound));
	mul[0] = mul[1] = 0;
	ExRAM = (uint8*)FCEU_gmalloc(1024);
	Mapper5_ESI();
	SetWriteHandler(0x5C00,0x5FEF,MMC5_ExRAMWr);
	SetReadHandler(0x5C00,0x5FEF,MMC5_ExRAMRd);
	MMC5ExMode = 2;
	SetWriteHandler(0x5000,0x5015,MMC5W5);
	SetWriteHandler(0x5205,0x5206,MMC5W5200);
	SetReadHandler(0x5205,0x5206,MMC5R5200);
}

void GenMMC5Close(void)
{
	MMC5Ex = 0;
	if (WRAM)
		FCEU_free(WRAM);
	WRAM = NULL;
	if (ExRAM)
		{FCEU_free(ExRAM); ExRAM = NULL;}
}

static void GenMMC5Reset(void)
{
	int		x;

	for(x=0; x<4; ++x)
		PRGBanks[x] =~0;
	for(x=0; x<8; ++x)
		CHRBanksA[x] =~0;
	for(x=0; x<4; ++x)
		CHRBanksB[x] =~0;
	WRAMMaskEnable[0] = WRAMMaskEnable[1] =~0;

	mmc5psize = mmc5vsize = 3;
	CHRMode = 0;

	NTAMirroring = NTFill = ATFill = 0xFF;

	MMC5Synco();

	SetWriteHandler(0x5000,0x5015,MMC5W5);
	SetWriteHandler(0x5100,0x5130,MMC5W5100);
	SetWriteHandler(0x5200,0x5206,MMC5W5200);
	SetReadHandler(0x5200,0x5206,MMC5R5200);
	SetWriteHandler(0x5C00,0x5FFF,MMC5_ExRAMWr);
	SetReadHandler(0x5C00,0x5FFF,MMC5_ExRAMRd);
	SetWriteHandler(0x6000,0xFFFF,MMC5_WriteROMRAM);
	SetReadHandler(0x6000,0xFFFF,MMC5_ReadROMRAM);

//	GameHBIRQHook=MMC5_hb;
//	FCEU_CheatSetRAM(8,0x6000,WRAM);
	FCEU_CheatSetRAM(1,0x5C00,ExRAM);

	MMC5Sound.rawcontrol = 1; /* initially disabled */
}

static SFORMAT MMC5_StateRegs[] =
{
	{PRGBanks,4,"PRGB"},
	{CHRBanksA,16,"CHRA"},
	{CHRBanksB,8,"CHRB"},
	{&WRAMPage,1,"WRMP"},
	{WRAMMaskEnable,2,"WRME"},
	{&MMC5ABMode,1,"ABMD"},
	{&IRQScanline,1,"IRQS"},
	{&IRQEnable,1,"IRQE"},
	{&CHRMode,1,"CHRM"},
	{&NTAMirroring,1,"NTAM"},
	{&NTFill,1,"NTFL"},
	{&ATFill,1,"ATFL"},

	{mul,2,"MUL"},
	{&MMC5IRQR,1,"IRQR"},
	{&MMC5LineCounter,1,"LCTR"},
	{&mmc5psize,1,"PSIZ"},
	{&mmc5vsize,1,"VSIZ"},
	{MMC5ROMWrProtect,4,"WRPR"},
	{MMC5MemIn,5,"MEMI"},

	{&MMC5Sound.wl[0],2|FCEUSTATE_RLSB,"SDW0"},
	{&MMC5Sound.wl[1],2|FCEUSTATE_RLSB,"SDW1"},
	{MMC5Sound.env,2,"SDEV"},
	{&MMC5Sound.enable,1,"SDEN"},
	{&MMC5Sound.running,1,"SDRU"},
	{&MMC5Sound.raw,1,"SDRW"},
	{&MMC5Sound.rawcontrol,1,"SDRC"},
	{&MMC5Sound.dcount[0],4|FCEUSTATE_RLSB,"DCT0"},
	{&MMC5Sound.dcount[1],4|FCEUSTATE_RLSB,"DCT1"},
	{&MMC5Sound.vcount[0],4|FCEUSTATE_RLSB,"VCT0"},
	{&MMC5Sound.vcount[1],4|FCEUSTATE_RLSB,"VCT1"},
	{&MMC5Sound.BC[0],4|FCEUSTATE_RLSB,"BC00"},
	{&MMC5Sound.BC[1],4|FCEUSTATE_RLSB,"BC01"},
	{&MMC5Sound.BC[2],4|FCEUSTATE_RLSB,"BC02"},
	{0}
};

static void GenMMC5_Init(CartInfo *info,
	int wsize,
	int battery)
{
	GenMMC5Close();
	if (wsize)
	{
		WRAM = (uint8*)FCEU_gmalloc(wsize*1024u);
		SetupCartPRGMapping(0x10,WRAM,wsize*1024u,1);
		AddExState(WRAM,wsize*1024u,0,"WRAM");
	}

	MMC5fill = (uint8*)FCEU_gmalloc(1024);
	ExRAM = (uint8*)FCEU_gmalloc(1024);

	AddExState(ExRAM,1024,0,"ERAM");
	AddExState(&MMC5SplitCtrl,1,0,"SPLM");
	AddExState(&MMC5SplitY,1,0,"SPLS");
	AddExState(&MMC5SplitBank,1,0,"SPLP");
	AddExState(&MMC5CHRHigh,1,0,"5130");
	AddExState(MMC5_StateRegs,~0,0,0);

	MMC5WRAMsize = wsize>>3;
	BuildWRAMSizeTable();
	GameStateRestore = MMC5_StateRestore;
	info->Power = GenMMC5Reset;
	info->Close = GenMMC5Close;

	if (battery)
	{
		info->SaveGame[0] = WRAM;
		if (wsize<17)
			info->SaveGameLen[0] = 8192;
		else
			info->SaveGameLen[0] = 32768;
	}

	MMC5VROMMask = CHRmask4[0]; //&0x3F
	MMC5ExRAM = ExRAM;
	MMC5Ex = 1;
	MMC5VROMPtr = CHRptr[0];
	MMC5ExMode = 0;
	MMC5SplitCtrl = MMC5SplitY = MMC5SplitBank = 0;
	Mapper5_ESI();

//	FFCEUX_PPURead = mmc5_PPURead;
//	FFCEUX_PPUWrite = mmc5_PPUWrite;
}

void ExROM_Init(CartInfo *info)
{
	WRAMSIZE = DetectMMC5WRAMSize(info);
	GenMMC5_Init(info,WRAMSIZE,info->battery);
}

// ELROM seems to have 0KB of WRAM
// EKROM seems to have 8KB of WRAM, battery-backed
// ETROM seems to have 16KB of WRAM, battery-backed
// EWROM seems to have 32KB of WRAM, battery-backed

void ELROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,0,0);
}

void EKROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,8,info->battery);
}

void ETROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,16,info->battery);
}

void EWROM_Init(CartInfo *info)
{
	GenMMC5_Init(info,32,info->battery);
}
