/*
 boards/SomeriTeam.c

 * MacFCEU 0.9d8
 *
 * Copyright notice for this file:
 *  mmc3.c Copyright (C) 1998 BERO
 *  mmc3.c Copyright (C) 2003 Xodnizel
 *  116.c Copyright (C) 2011 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "MMC1.h"
#include "MMC3and6.h"

/* ---- constants */

enum
{
	modeVRCII,
	modeMMC3,
	modeMMC1 /* 2/3 */
};

/* ---- globals */

/* just do what FCEUXMM does, and redundantly handle VRCII functions here */
static uint8 V2CHR[8];
#define mode MMC3DReg[10]
#define V2MIR MMC3DReg[11]
#define V2PRG (MMC3DReg+12)

static void SL12Sync(int v);
static DECLFW(SL12VRC2Write8);
static DECLFW(SL12VRC2WriteA);
static DECLFW(SL12VRC2WriteC);

static writefunc	write8 = SL12VRC2Write8;
static writefunc	writeA = SL12VRC2WriteA;
static writefunc	writeC = SL12VRC2WriteC, writeE = SL12VRC2WriteC;

/* ---- mapper 116: Someri Team SL-12 */

static void SL12SyncPRG(void)
{
	if (mode==modeVRCII)
	{
		setprg8(0x8000,V2PRG[0]);
		setprg8(0xA000,V2PRG[1]);
		setprg16(0xC000,~0);
	} else if (mode==modeMMC3)
	{ /* unsure if MMC3 actually does this */
		uint32		A = MMC3cmd>>5&2;

		//setprg8(A,V&0x3F);
		setprg8(0x8000,MMC3DReg[6+A]);
		setprg8(0xA000,MMC3DReg[7]);
		setprg8(0xC000,MMC3DReg[6+(A^2)]);
		setprg8(0xE000,MMC3DReg[9]);
	} else
	{
		uint8		bank = MMC1PRG0&15;

		if (MMC1Ctl&MMC1PRGSwap16)
			if (MMC1Ctl&MMC1PRGSwapLow)
			{
				setprg16(0x8000,bank);
				setprg16(0xC000,15);
			} else
			{
				setprg16(0x8000,0);
				setprg16(0xC000,bank);
			}
		else
			setprg32(0x8000,bank>>1);
	}
}

static void SL12SyncCHR(void)
{
	if (mode==modeVRCII)
	{
		uint32		hi = mode<<6&0x100;

		setchr1(0x0000,hi|V2CHR[0]);
		setchr1(0x0400,hi|V2CHR[1]);
		setchr1(0x0800,hi|V2CHR[2]);
		setchr1(0x0C00,hi|V2CHR[3]);
		setchr1(0x1000,hi|V2CHR[4]);
		setchr1(0x1400,hi|V2CHR[5]);
		setchr1(0x1800,hi|V2CHR[6]);
		setchr1(0x1C00,hi|V2CHR[7]);
		return;
	}
	if (mode==modeMMC3)
	{
		FixMMC3CHR(MMC3cmd);
		return;
	}
	setchr8r(0x10,0); /* ?? */
	if (MMC1Ctl&MMC1BankMode)
	{
		setchr4(0x0000,MMC1DReg[1]);
		setchr4(0x1000,MMC1DReg[2]);
	} else
		setchr8(MMC1DReg[1]>>1);
}

static void SL12SyncMIR(void)
{
	if (mode<modeMMC1)
		setmirror(mode ? MMC3A000B&1:~V2MIR&1);
	else
		MMC1FixMir();
}

static DECLFW(SL12VRC2WriteC)
{
	int32		x = (((A>>1&1)|A>>11)+2)&7;
	int32		h = A<<2&4;

	V2CHR[x] = (V2CHR[x]&0xF0>>h)|(V&15)<<h;
	SL12SyncCHR();
}

static DECLFW(SL12VRC2WriteA)
{
	if (A>=0xB000)
	{
		SL12VRC2WriteC(A,V);
		return;
	}
	V2PRG[1] = V; SL12SyncPRG();
}

static DECLFW(SL12VRC2Write8)
{
	if (A<0x9000)
		{V2PRG[0] = V; SL12SyncPRG();}
	else
		{V2MIR = V; SL12SyncMIR();}
}

static void SL12SetMode(void)
{
	if (mode>=modeMMC1)
	{
		write8 = writeA = writeC = writeE = MMC1_write;
		return;
	}
	if (mode==modeMMC3)
	{
		write8 = MMC3_8CMDWrite;
		writeA = MMC3_AMWSWrite;
		writeC = MMC3_CIRQWrite;
		writeE = MMC3_EIRQWrite;
		return;
	}
	write8 = SL12VRC2Write8;
	writeA = SL12VRC2WriteA;
	writeC = writeE = SL12VRC2WriteC;
}

static DECLFW(SL12Write4)
{
//	if(A==0x4100) {MMC3EXPREGS[0]=V; setmirror(V&1);}
//	else if(A==0x4141) MMC3EXPREGS[1]=V;
//	else if(A==0xA000) MMC3EXPREGS[2]=V;
//	else if(A==0x4106) MMC3EXPREGS[3]=V;
	V&= 3;
	if (mode==V)
		return;
	mode = V;
	if (A&1)
		MMC1SerialClr();
	if (V!=modeMMC3)
	{
		if (MMC3IRQa)
			X65xx_IRQEnd(FCEU_IQEXT);
		MMC3IRQa = 0;
	}
	SL12Sync(0);
//	lastp = lastc = ~0;
}

static DECLFW(SL12Write8)
{
	write8(A,V);
}

static DECLFW(SL12WriteA)
{
	writeA(A,V);
}

static DECLFW(SL12WriteC)
{
	writeC(A,V);
}

static DECLFW(SL12WriteE)
{
	writeE(A,V);
}

static void SL12Sync(int v)
{
(void)v;
	mode&= 3;
	SL12SetMode();
	SL12SyncPRG();
	SL12SyncCHR();
	SL12SyncMIR();
}

static void SL12HBIRQ(void)
{
	if (mode!=modeMMC3)
		return;
	ClockMMC3Counter();
}

static void SL12Power(void)
{
	int			x;

	memset(MMC3DReg,0,sizeof(MMC3DReg));
	V2CHR[0] = V2CHR[1] = V2CHR[2] = V2CHR[3] =~0; /* #3 varies */
	V2CHR[4] = 4;
	V2CHR[5] = 5;
	V2CHR[6] = 6;
	V2CHR[7] = 7;
	V2PRG[1] = 1;
	MMC3WRAM_ctl = MMC3A000B = 0;
	MMC3DReg[1] = 2;
	MMC3DReg[2] = 4;
	MMC3DReg[3] = 5;
	MMC3DReg[4] = 6;
	MMC3DReg[5] = 7;
	MMC3DReg[6] =~3;
	MMC3DReg[7] =~2;
	MMC3DReg[8] =~1;
	MMC3DReg[9] =~0;

	GenMMC1Reset();
	MMC1Ctl = MMC1CtlPower;

	x = 0x4100;
	do
		SetWriteHandler(x,x+0xFF,SL12Write4);
	while((x+=0x200)<0x6000);
	SetWriteHandler(0x8000,0x9FFF,SL12Write8);
	SetWriteHandler(0xA000,0xBFFF,SL12WriteA);
	SetWriteHandler(0xC000,0xDFFF,SL12WriteC);
	SetWriteHandler(0xE000,0xE003,SL12WriteE);
	SetCartRead(0x8000,0xFFFF);

	SL12Sync(0);
}

//extern SFORMAT MMC3_StateRegs[], MMC1_StateRegs[];

void SomeriTeam_SL12_Init(CartInfo *info)
{
	GenMMC3_Init(info,128,512,0,0); //512,256
	GenMMC1_Init(info,128,512,0,info->battery);
//	MMC3Persist = 1;
	info->Power = SL12Power;
	GameHBIRQHook = PPU_A12hook = SL12HBIRQ;
	GameStateRestore = SL12Sync;
//	MMC3cwrap = SL12CW;
//	MMC3pwrap = SL12PW;
	AddExState(V2CHR,8,0,"VRCC");
//	AddExState(MMC3_StateRegs,~0,0,0);
	AddExState(MMC3EXPREGS,6,0,"EXPR");
//	AddExState(MMC1_StateRegs,~0,0,0);
	if (NewCHRRAM(0x10,8192))
		AddExState(CHRptr[0x10],CHRsize[0x10],0,"CHRR");
}
