/*
 boards/Tengen.c
  mappers 64, 158

 * MacFCEU 0.9a1
 *
 * Copyright notice for this file:
 *  tengen.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "Tengen.h"

uint8	TengenDReg[11]; //0:C0,C2,C4,C5,C6,C7,P0,P1,C1,C3,F:P2
uint8	TengenCmd;

static uint8	mirr, domirA, rmode, IRQmode, IRQCount, IRQa, IRQLatch;

static void RAMBO1_CHR(void);
static void (*TengenCHR)(void) = RAMBO1_CHR;

static SFORMAT StateRegs[] =
{
	{TengenDReg, 11, "REGS"},
	{&TengenCmd, 1, "CMDR"},
	{&mirr, 1, "MIRR"},
	{&rmode, 1, "RMOD"},
	{&IRQmode, 1, "IRQM"},
	{&IRQCount, 1, "IRQC"},
	{&IRQa, 1, "IRQA"},
	{&IRQLatch, 1, "IRQL"},
	{0}
};

static void RAMBO1_IRQHook(int a)
{
	static int32 smallcount;

	if (!IRQmode)
		return;
	smallcount+= a;
	while(smallcount>=4)
	{
		smallcount-= 4;
		if (--IRQCount==0xFF && IRQa)
			X65xx_IRQBegin(FCEU_IQEXT);
	}
}

static void RAMBO1_HBHook(void)
{
	if (!IRQmode && scanline!=240)
	{
		rmode = 0;
		if (--IRQCount==0xFF && IRQa)
		{
			rmode = 1;
			X65xx_IRQBegin(FCEU_IQEXT);
		}
	}
}

static void RAMBO1_CHR(void)
{
	unsigned	x, chrxor = (unsigned int)TengenCmd<<5&0x1000;

	if (TengenCmd&0x20)
	{
		setchr1(chrxor^0,TengenDReg[0]);
		setchr1(chrxor^0x400,TengenDReg[8]);
		setchr1(chrxor^0x800,TengenDReg[1]);
		setchr1(chrxor^0xC00,TengenDReg[9]);
	} else
	{
		setchr2(chrxor^0,TengenDReg[0]>>1);
		setchr2(chrxor^0x800,TengenDReg[1]>>1);
	}
	x = 4;
	do /* 0x1000-0x1C00, 2-5 */
		setchr1(chrxor^x<<10,TengenDReg[x-2]);
	while(++x<8);
}

static void RAMBO1_Sync(int v)
{
(void)v;
	TengenCHR();
	setprg8(0x8000,TengenDReg[6]);
	setprg8(0xA000,TengenDReg[7]);
	setprg8(0xC000,TengenDReg[10]);
	setprg8(0xE000,~0);
	if (domirA)
		setmirror(mirr);
}

static DECLFW(RAMBO1_Write)
{
	switch(A&0xF001)
	{
	case 0xA000: mirr =~V&1; RAMBO1_Sync(0); break;
	case 0x8000: TengenCmd = V; break;
	case 0x8001:
		A = TengenCmd&15;
		if (A<10)
			TengenDReg[A] = V;
		else if (A==15)
			TengenDReg[10] = V;
		RAMBO1_Sync(0);
		break;
	case 0xC000:
		IRQLatch = V;
		if (rmode==1)
			IRQCount = IRQLatch;
		break;
	case 0xC001:
		rmode = 1;
		IRQCount = IRQLatch;
		IRQmode = V&1;
		break;
	case 0xE000:
		IRQa = 0;
		X65xx_IRQEnd(FCEU_IQEXT);
		if (rmode==1)
			IRQCount = IRQLatch;
		break;
	case 0xE001:
		IRQa = 1;
		if (rmode==1)
			IRQCount = IRQLatch;
	}
}

static void RAMBO1_Power(void)
{
	TengenCmd = mirr = 0;
	TengenDReg[0] = TengenDReg[1] = TengenDReg[2] = TengenDReg[3] =
		TengenDReg[4] = TengenDReg[5] = TengenDReg[6] = TengenDReg[7] =
		TengenDReg[8] = TengenDReg[9] = TengenDReg[10] =~0;
	RAMBO1_Sync(0);
	SetReadHandler(0x8000,0xFFFF,CartBR);
	SetWriteHandler(0x8000,0xFFFF,RAMBO1_Write);
}

/* ---- mapper 64: Tengen 800032 */

void Tengen800032_Init(CartInfo *info)
{
	domirA = 1;
	TengenCHR = RAMBO1_CHR;

	info->Power = RAMBO1_Power;
	MapIRQHook = RAMBO1_IRQHook;
	GameHBIRQHook = RAMBO1_HBHook;
	GameStateRestore = RAMBO1_Sync;
	AddExState(&StateRegs,~0,0,0);
}

/* ---- mapper 158: Tengen 800037 */

static void CHR8_00037(void)
{
	RAMBO1_CHR();
	if (TengenCmd&0x80)
		setmirrorw(TengenDReg[2]>>7^1,
			TengenDReg[3]>>7^1,
			TengenDReg[4]>>7^1,
			TengenDReg[5]>>7^1);
	else
		setmirrorw(TengenDReg[0]>>7^1,
			TengenDReg[0]>>7^1,
			TengenDReg[1]>>7^1,
			TengenDReg[1]>>7^1);
}

void Tengen800037_Init(CartInfo *info)
{
	Tengen800032_Init(info);
	domirA = 0;
	TengenCHR = CHR8_00037;
//	setchr1wrap = MirWrap;
//	PPU_hook = MirrorFear;
}

/*
 MIMIC1 boards T800002, T800004, T800030 are actually Sunsoft4...
*/
