/*
 boards/latch.c
  mappers 2, 3, 7, 11/144, 13, 38, 59, 66, 70, 78, 79, 93, 94,
	107, 140, 152, 174, 240, 241

 * MacFCEU 0.9a1
 *
 * Copyright notice for this file:
 *  59.c, 79.c, datalatch.c (C) 2002 Xodnizel
 *  addrlatch.cpp (C) 2006 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "latch.h"
//#include "cheat.h"

#define CHRRAM (GameMemBlock)

uint8		latch_byte;//, bus_conflict;
uint16		latch_addr;

static uint32	latch_init;
static uint32	addrlohi;
static void	(*WSync)(void);
static readfunc	rfunc;

static uint8	*WRAM, wramz;

DECLFW(Latch_Write)
{
	latch_addr = A;
	latch_byte = V;
	WSync();
}

DECLFW(Latch_BusWrite)
{
	latch_addr = A;
	latch_byte = CartBR(A); /* was `ARead[A](A)`; 2010.04.12 */
	//should be able to get how far into the instruction the CPU is
	if (V!=latch_byte && (XFrame.count>0 || int64lo(timestampbase)&2))
		latch_byte = V; /* FCEUX does `CartBR(A)&V` */
	WSync();
}

static DECLFW(Latch_CmdWrite)
{
	if ((latch_addr=A)&latch_init>>16)
	{
		latch_addr&= 0xFF00, latch_addr|= V;
		WSync();
	}// else if (bus_conflict)
	//	latch_byte = V&CartBR(A);
	else
		latch_byte = V;
}

static void LatchRestore(int v)
{
(void)v;
	WSync();
}

static void M11Sync(void); //as default

void Latch_Power(void)
{
	writefunc	wf = latch_init&0x7F0000U ? Latch_CmdWrite
		:latch_init&1<<23 ? Latch_BusWrite:Latch_Write;

//	bus_conflict = 0;
	wf(latch_init&0xFFFF,latch_init>>24);
	if (rfunc==CartBR)
		if (WRAM)
		{
			SetCartRead(0x6000,0xFFFF);
			SetWriteHandler(0x6000,0x7FFF,CartBW);
			//FCEU_CheatSetRAM(wramz,0x6000,WRAM); --not working
		} else
			SetCartRead(0x8000,0xFFFF);
	else
		SetReadHandler(0x8000,0xFFFF,rfunc);
	SetWriteHandler(addrlohi>>16,addrlohi&0xFFFF,wf);
}

void Latch_Close(void)
{
	if (WRAM)
		FCEU_free(WRAM);
	WRAM = NULL;
}

void Latch_Init(CartInfo *info,
	void (*wsyncproc)(void),
	readfunc func,
	uint32 init,
	uint32 lohi,
	uint8 wram)
{
	latch_init = init;
	addrlohi = lohi;

	if ((WSync=wsyncproc)==0)
		WSync = M11Sync;

	AddExState(&latch_byte,1,0,init&0x7F0000U ? "CMD":"LATC");
	if (init&0xFFFF)
		AddExState(&latch_addr,2,1,"LATA");
	if (wram)
	{
		int		z = wram<<10;

		wramz = wram;
		WRAM = FCEU_gmalloc(z);
		AddExState(WRAM,z,0,"WRAM");
		SetupCartPRGMapping(0x10,WRAM,z,1);
		if (info->battery)
		{
			info->SaveGame[0] = WRAM;
			info->SaveGameLen[0] = z;
		}
	}
	info->Power = Latch_Power;
	info->Close = Latch_Close;

	if (!func)
		func = CartBR;
	rfunc = func;
	GameStateRestore = LatchRestore;
}

static void CapHIPower(void)
{
	setprg16(0xC000,~0);
	Latch_Power();
}

/* ---- mapper 2: UxROM */

static void UNROMSync(void)
{
	setprg16(0x8000,latch_byte);
	setchr8(0);
}

void UxROM_Init(CartInfo *info)
{
	Latch_Init(info,UNROMSync,0,0,0x8000FFFFU,0);
	info->Power = CapHIPower;
//	bus_conflict = 1;
//	if(ROM_size<256>>4) PRGmask16[0]&=7;
//	AddExState(CHRRAM,8192,0,"CHRR");
}

/* ---- mapper 3: CNROM */

static void CNROMSync(void)
{
	setchr8(latch_byte); //&=3
	setprg16(0x8000,0);
	setprg16(0xC000,1);
}

void CNROM_Init(CartInfo *info)
{
	Latch_Init(info,CNROMSync,0,0,0x8000FFFFU,0);
//	bus_conflict = 1;
}

/* ---- mapper 7: ANROM/AOROM */

static void AxROMSync(void)
{
	setprg32(0x8000,latch_byte&15);
	setmirror(MI_0+(latch_byte>>4&1));
	setchr8(0); /* 2015.09.13 */
}

void AxROM_Init(CartInfo *info)
{
	Latch_Init(info,AxROMSync,0,0,0x8000FFFFU,0);
}

/* ---- mapper 11 (AGCI-47516/ColorDreams 74*377), 144 (AGCI 50282) */

static void M11Sync(void)
{
	setprg32(0x8000,latch_byte&15);
	setchr8(latch_byte>>4);
}

void AGCI47516_Init(CartInfo *info)
{
	Latch_Init(info,0,0,0,0x8000FFFFU,0);
}

void AGCI50282_Init(CartInfo *info)
{
	Latch_Init(info,0,0,0,0x8001FFFFU,0);
}

/* ---- mapper 13: 32KB VRAM CPROM */

static void CPROMSync(void)
{
//	setvram4(0x1000,CHRRAM+((latch_byte&=3)<<12));
	setchr4r(0x10,0x1000,latch_byte&3);
	setprg32(0x8000,latch_byte>>4&3);
}

/*static void CPROMPower(void)
{
//	setvram8(0);
	Latch_Power();
}
*/

void CPROM_Init(CartInfo *info)
{
	Latch_Init(info,CPROMSync,0,0,0x8000FFFFU,0);
//	info->Power = CPROMPower;
	AddExState(CHRRAM,16384,0,"CHRR");
	SetupCartCHRMapping(0x10,CHRRAM,16384,1);
}

/* ---- mapper 34: BxROM or AVE NINA-001 */

static void BNROMSync(void)
{
	setprg32(0x8000,latch_byte);
	setchr8r(0x10,0);
}

void BNROM_Init(CartInfo *info)
{
	Latch_Init(info,BNROMSync,0,0,0x8000FFFF,0);
	AddExState(CHRRAM,8192,0,"CHRR");
	SetupCartCHRMapping(0x10,CHRRAM,8192,1);
}

/*void BxROM_Init(CartInfo *info)
{
	if (VROM_size)
		{AVE_NINA001_Init(info); return;}
	BNROM_Init(info);
}*/

/* ---- mapper 38: 74*161/138 */

static void M38Sync(void)
{
	setprg32(0x8000,latch_byte&3);
	setchr8(latch_byte>>2);
}

void Discrete_74_161_138_Init(CartInfo *info)
{
	Latch_Init(info,M38Sync,0,0,0x70007FFFU,0);
}

/* ---- mapper 59: T3H53 */

static void M59Sync(void)
{
	unsigned	A = latch_addr;

	setprg32(0x8000,A>>4&7);
	setchr8(A&7);
	setmirror(A>>3&1); //MIRROR_SET2
}

static DECLFR(M59R8000)
{
	if (latch_addr&0x100)
		return 0;
	else
		return CartBR(A);
}

void T3H53_Init(CartInfo *info)
{
	Latch_Init(info,M59Sync,M59R8000,0x8000,0x8000FFFFU,0);
}

/* ---- mapper 66, 140/240: GNROM compatible, MHROM */

/*static void M66Sync2(void)
{
	setprg32(latch_byte>>4);
	VRAM_BANK8(latch_byte&3);
}
*/

static void GNROMSync(void)
{
//	latch_byte&= 0x33;
	setprg32(0x8000,latch_byte>>4);
	setchr8(latch_byte);
}

void GNROM_Init(CartInfo *info)
{ /* notey: this can have bus conflicts */
	Latch_Init(info,GNROMSync,0,0,0x8000FFFFU,0);
}

void MHROM_Init(CartInfo *info)
{
	GNROM_Init(info);
//	PRGmask32[0]&= 1;
//	CHRmask8[0]&= 1;
}

void JF14x_Init(CartInfo *info)
{
	Latch_Init(info,GNROMSync,0,0,0x60007FFFU,0);
}

void CandE_SHLZ_Init(CartInfo *info)
{
	Latch_Init(info,GNROMSync,0,0,0x40205FFFU,0);
}

/* ---- mapper 70(type A), 152(type B): 74*161/161/32 */

static void M70Sync(void)
{
	setprg16(0x8000,latch_byte>>4);
	setchr8(latch_byte&15);
}

void Discrete_74_161_161_32A_Init(CartInfo *info)
{
	Latch_Init(info,M70Sync,0,0,0x8000FFFFU,0);
	info->Power = CapHIPower;
}

static void M152Sync(void)
{
	setprg16(0x8000,latch_byte>>4&7);
	setchr8(latch_byte&15);
	setmirror(MI_0+(latch_byte>>7&1));
}

void Discrete_74_161_161_32B_Init(CartInfo *info)
{
	if (info->mirror<=MI_V)
		{Discrete_74_161_161_32A_Init(info); return;}
	Latch_Init(info,M152Sync,0,0,0x8000FFFFU,0);
	info->Power = CapHIPower;
}

/* ---- mapper 78: Jaleco JF-16 */

static void J16Sync(void)
{
	setprg16(0x8000,latch_byte); //&7
	setchr8(latch_byte>>4);
	setmirror(MI_0+(latch_byte>>3&1));
}

void JF16_Init(CartInfo *info)
{
	Latch_Init(info,J16Sync,0,1<<23,0x8000FFFFU,0);
	info->Power = CapHIPower;
}

static void HDSync(void)
{
	setprg16(0x8000,latch_byte); //&7
	setchr8(latch_byte>>4);
	setmirror(latch_byte>>3&1 ? MI_V:MI_H);
}

void HolyDiver_Init(CartInfo *info)
{
	Latch_Init(info,HDSync,0,1<<23,0x8000FFFFU,0);
	info->Power = CapHIPower;
}

/* ---- mapper 79: AVE NINA-06 */
/* some older state files may not be compatible */

static void M79Sync(void)
{
	setprg32(0x8000,latch_byte>>3);
	setchr8(latch_byte);
}

static void M79Power(void)
{
	int		i;

	Latch_Power();
	for(i=0x4300; i<0x6000; i+=0x200)
		SetWriteHandler(i,i+0xFF,Latch_Write);
}

void AVE_NINA06_Init(CartInfo *info)
{
	Latch_Init(info,M79Sync,0,0,0x410041FF,0);
	info->Power = M79Power;
}

/* ---- mapper 87: 74*139/74 */

static void M87Sync(void)
{
	setprg16(0x8000,0);
	setprg16(0xC000,1);
	setchr8((latch_byte>>1&1)|(latch_byte<<1&2));
}

void Discrete_74_139_74_Init(CartInfo *info)
{
	Latch_Init(info,M87Sync,0,0xFFU<<24,0x6000FFFFU,0);
}

/* ---- mapper 93: Sunsoft-2A - "Shanghi" original version mapper */

static void Sunsoft2ASync(void)
{
	setprg16(0x8000,latch_byte>>4);
	setchr8(0);
}

void Sunsoft2A_Init(CartInfo *info)
{
	Latch_Init(info,Sunsoft2ASync,0,0,0x8000FFFFU,0);
	info->Power = CapHIPower;
}

/* ---- mapper 94: UN1ROM */

static void UN1ROMSync(void)
{
	setprg16(0x8000,(/*ARead[latch_addr](latch_addr)&*/latch_byte)>>2);
	setchr8(0);
}

void UN1ROM_Init(CartInfo *info)
{
	Latch_Init(info,UN1ROMSync,0,0,0x8000FFFFU,0);
	info->Power = CapHIPower;
}

/* ---- mapper 107: Magic Series */

static void M107Sync(void)
{
	setprg32(0x8000,latch_byte>>1&3);
	setchr8(latch_byte&7);
}

void Magic_Series_Init(CartInfo *info)
{
	Latch_Init(info,M107Sync,0,0xFFU<<24,0x8000FFFFU,0);
}

/* ---- mapper 241: TXC / MXMDHTWO */

static void M241Sync(void)
{ /* still not known how 5xxx addresses play in */
	setchr8(0);
	setprg8r(0x10,0x6000,0); /* SRAM */
	setprg32(0x8000,latch_byte);//|(latch_byte>>4&8));
}

void TXC_MXMDHTWO_Init(CartInfo *info)
{
	Latch_Init(info,M241Sync,0,0,0x8000FFFFU,8);
}

/* ---- mapper 174: NTDec 5-in-1 */
/* have not gotten to menu with the dump! */

static void M174Sync(void)
{
	if (latch_addr&0x80)
	{
		int		x = latch_addr>>4&7;

		setprg16(0x8000,x);
		setprg16(0xC000,x);
	} else
		setprg32(0x8000,latch_addr>>5&3);
	setchr8(latch_addr>>1&7);
	setmirror(~latch_addr&1);
}

void NTDec5in1_Init(CartInfo *info)
{
	Latch_Init(info,M174Sync,0,0xFF00,0x8000FFFFU,0);
//	info->Power = M174Power;
}
