/*
 boards/unlicensed.c
  short unlicensed, also bootleg boards

 * MacFCEU 0.9a2
 *
 * Copyright notice for this file:
 *  mmc3.cpp Copyright (C) 1998 BERO
 *  mmc3.cpp Copyright (C) 2003 Xodnizel
 *  01-222.cpp Copyright (C) 2006 CaH4e3
 *  cityfighter.c Copyright (C) 2007 CaH4e3
 *  bb.c Copyright (C) 2007 CaH4e3
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"
#include "latch.h"
#include "MMC3and6.h"
#include "unlicensed.h"

static uint8	reg[8];

/* ---- BMC Golden 190-in-1 */

static void BMC190in1Sync(void)
{
	int		v = latch_byte;

	setmirror(~v&1);
	setprg16(0x8000,v>>=2);
	setprg32(0xC000,v);
	setchr8(v);
}

void BMC_190in1_Init(CartInfo *info)
{
	Latch_Init(info,BMC190in1Sync,0,0,0x8000FFFFU,0);
}

/* ---- UNL-CC21 */

static void UNLCC21Sync(void)
{
	setprg32(0x8000,0);
	setchr8(latch_addr);
	setmirror(MI_0+(latch_addr>>1&1));
}

void UNL_CC21_Init(CartInfo *info)
{
	Latch_Init(info,UNLCC21Sync,0,0x8000,0x8000FFFFU,0);
}

/* ---- mapper 28: INL-ROM (Action 53) */

#define csel reg[0]
#define psel reg[1]
#define mode reg[2]
#define bsel reg[3]
#define exp reg[4]

static void M28Sync(int version)
{
	int		lo = (bsel&63)*2u, hi = lo|1, k;
(void)version;
	switch(mode>>2&15)
	{
	/* fixed-low */
	case 2: hi = lo|(psel&1); break;
	case 6: hi = (lo&~2)|(psel&3); break;
	case 10: hi = (lo&~6)|(psel&7); break;
	case 14: hi = (lo&~14)|(psel&15); break;
	/* fixed-high */
	case 3: lo|= psel&1; break;
	case 7: lo = (lo&~2)|(psel&3); break;
	case 11: lo = (lo&~6)|(psel&7); break;
	case 15: lo = (lo&~14)|(psel&15); break;
	/* 32KB */
	case 4: case 5: lo = (lo&~2)|(psel<<1&2), hi = lo|1; break;
	case 8: case 9: lo = (lo&~6)|(psel<<1&6), hi = lo|1; break;
	case 12: case 13: lo = (lo&~14)|(psel<<1&14);
	default: hi = lo|1;
	}
	k = PRGsize[0]-1;
	setprg16(0x8000,lo&k);
	setprg16(0xC000,hi&k);
	setchr8(csel&3);
	setmirror(mir01VH[mode&3]);
}

static DECLFW(M28W5)
{
(void)A;
	exp = V&0x81;
}

static DECLFW(M28W8)
{
(void)A;
	if (!exp)
	{ //00
		csel = V;
		if (!(mode&2))
			mode = (mode&~1)|(V>>4&1);
		setmirror(mir01VH[mode&3]);
	} else if (!(exp-1))
	{ //01
		psel = V;
		if (!(mode&2))
			mode = (mode&~1)|(V>>4&1);
		M28Sync(V);
	} else if (exp==0x80)
		{mode = V; M28Sync(0);}
	else //81
		{bsel = V; M28Sync(0);}
}

static void M28Reset(void)
{
	psel = 15, bsel = 63;
	M28Sync(0);
}

static void M28Power(void)
{
	//setprg16r(0x10,0x6000,0);
	//SetWriteHandler(0x6000,0x7FFF,FCEU_BWRAM);
	SetWriteHandler(0x5000,0x5FFF,M28W5);
	SetCartRead(0x6000,0xFFFF);
	SetWriteHandler(0x8000,0xFFFF,M28W8);
	M28Reset();
}

void INL28_Init(CartInfo *info)
{
	info->Reset = M28Reset;
	info->Power = M28Power;
	GameStateRestore = M28Sync;
	AddExState(reg,5,0,"REG");
	//SetupCartPRGMapping(0x10,GameMemBlock,8192,1);
	//if (info->battery)
	//	info->SaveGame[0] = GameMemBlock, info->SaveGameLen[0] = 8192;
}

/* ---- mapper 44: BMC SuperBig 7-in-1 */

static void M44PW(uint32 A,
	uint8 V)
{
	uint32		NV = V;

	if (MMC3EXPREGS[0]<6)
		NV&= 0xF;
	else
		NV&= 0x1F;
	NV|= MMC3EXPREGS[0]<<4;
	setprg8(A,NV);
}

static void M44CW(uint32 A,
	uint8 V)
{
	uint32		NV = V;

	if (MMC3EXPREGS[0]<6)
		NV&= 0x7F;
	NV|= MMC3EXPREGS[0]<<7;
	setchr1(A,NV);
}

static DECLFW(M44_write)
{
(void)A;
	V&= 7;
	if (MMC3EXPREGS[0]==V)
		return;
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M44Power(void)
{
	MMC3EXPREGS[0] = 0;
	GenMMC3Power();
	SetWriteHandlerI(2,0xA001,0xBFFF,M44_write);
}

void BMC_SuperBig7in1_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	info->Power = M44Power;
	MMC3cwrap = M44CW;
	MMC3pwrap = M44PW;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- mapper 52: BMC Mario Party 7-in-1 */

static void M52PW(uint32 A,
	uint8 V)
{
	uint32		NV = V;

	NV&= 0x1F^(MMC3EXPREGS[0]<<1&0x10);
	NV|= ((MMC3EXPREGS[0]&6)|(MMC3EXPREGS[0]>>3&MMC3EXPREGS[0]&1))<<4;
	setprg8(A,NV);
}

static void M52CW(uint32 A,
	uint8 V)
{
	uint32	NV = V;

	NV&=~(MMC3EXPREGS[0]<<1&0x80);
	NV|= ((MMC3EXPREGS[0]>>3&4)|(MMC3EXPREGS[0]>>1&2)
		|(MMC3EXPREGS[0]>>6&MMC3EXPREGS[0]>>4&1))<<7;
	setchr1(A,NV);
}

static DECLFW(M52_write)
{
	if (MMC3EXPREGS[1]) //update once, then set latch
	{
		MMC3WRAM[A-0x6000] = V;
		return;
	}
	MMC3EXPREGS[1] = 1;
	MMC3EXPREGS[0] = V;
	FixMMC3PRG(MMC3cmd);
	FixMMC3CHR(MMC3cmd);
}

static void M52Power(void)
{
	MMC3EXPREGS[0] = MMC3EXPREGS[1] = 0;
	GenMMC3Power();
	SetWriteHandler(0x6000,0x7FFF,M52_write);
}

void BMC_MarioParty7in1_Init(CartInfo *info)
{
	GenMMC3_Init(info,512,256,8,info->battery);
	info->Power = M52Power;
	MMC3cwrap = M52CW;
	MMC3pwrap = M52PW;
	AddExState(MMC3EXPREGS,2,0,"EXPR");
}

/* ---- mapper 58: BMC-GK192 */

static void BMC_GK192_Sync(void)
{
	if (latch_addr&0x40)
	{
		int		x = latch_addr&7;

		setprg16(0x8000,x);
		setprg16(0xC000,x);
	} else
		setprg32(0x8000,latch_addr>>1&3);
	setchr8(latch_addr>>3&7);
	setmirror(latch_addr>>7^1);
}

void BMC_GK192_Init(CartInfo *info)
{
	Latch_Init(info,BMC_GK192_Sync,0,0x8000,0x8000FFFFU,0);
}

/* ---- mapper 132: UNL 22211 type A */

#define mapix	reg[4]

static void UNL22211_Sync(void)
{
	setprg32(0x8000,reg[2]>>2); //&1);
	if (mapix==172)
	{ // 1991 DU MA Racing CHR bank sequence is probably WRONG, so rearranging
	// CHR banks for normal UNIF board and mapper 172 may be unneccessary
	 	int	b = (latch_byte^reg[2])>>3;

		setchr8((b&2)|(b>>2&1));
	} else
		setchr8(reg[2]&3);
}

static DECLFW(UNL22211W4)
{
	reg[A&3] = V;
}

static DECLFR(UNL22211R4)
{
(void)A;
	return (reg[1]^reg[2])|(mapix==173 ? 0x41:0x40);
}

static void UNL22211Power(void)
{
	{
		uint8		*p = reg;

		*p = p[1] = p[2] = p[3] = 0;
	}
	SetReadHandler(0x4100,0x4100,UNL22211R4);
	SetWriteHandler(0x4100,0x4103,UNL22211W4);
	Latch_Power();
}

void UNL_22211A_Init(CartInfo *info)
{
	mapix = 132;
	Latch_Init(info,UNL22211_Sync,0,0,0x8000FFFFU,0);
	info->Power = UNL22211Power;
	AddExState(reg,4,0,"REG");
}

/* ---- mapper 172: UNL 22211 type B */

void UNL_22211B_Init(CartInfo *info)
{
	UNL_22211A_Init(info);
	mapix = 172;
}

/* ---- mapper 173: UNL 22211 type C */

void UNL_22211C_Init(CartInfo *info)
{
	UNL_22211A_Init(info);
	mapix = 173;
}

/* ---- mapper 235: GoldenGame 150-in-1/260-in-1 */

#define ggopen	reg[0]
#define ggsel	reg[1]
#define ggbankx	0xE4901000U
#define ggopenx	0x0457

static void M235Sync(void)
{
	uint32		A = latch_addr;
	uint32		s = ggsel+(A>>8&3);
	uint32		b = (ggbankx>>s+s&3)<<5;

	if (A&0x400)
		setmirror(MI_0);
	else
		setmirror(~A>>13&1);
	if (A&0x800)
	{
		b+= b+(A>>12&1);
		setprg16(0x8000,b);
		setprg16(0xC000,b);
	} else
		setprg32(0x8000,b);
	ggopen = ggopenx>>s&1;
}

static DECLFR(M235_read)
{
	return ggopen ? A>>8:CartBR(A);
}

static void M235Power(void)
{
	ggopen = 0;
	if (ggsel!=3)
		SetReadHandler(0x8000,0xFFFF,M235_read);
	Latch_Power();
}

void BMC_GoldenGamexx0in1_Init(CartInfo *info)
{
	if ((ggsel=ROM_size-64>>4)>12)
		ggsel =~0; /*150 or 260*/
	ggsel&= 0xC;
	Latch_Init(info,M235Sync,0,0x8000,0x8000FFFFU,0);
	info->Power = M235Power;
	AddExState(reg,1,0,"OPN");
}

/* ---- mapper 254: bootleg Pikachu Y2K */
/* there's also an Ai Senshi Nicol pirate, but piracy is illegal... */

static DECLFR(M254_read)
{
	return MMC3WRAM[A-0x6000]^MMC3EXPREGS[1];
}

static DECLFW(M254_write)
{
	MMC3EXPREGS[0] = 0;
	MMC3_8CMDWrite(A,V);
}

static void M254Reset(void)
{
	MMC3EXPREGS[0] =~0;
	GenMMC3Reset();
}

static void M254Power(void)
{
	MMC3EXPREGS[0] =~0;
	GenMMC3Power();
	SetReadHandler(0x6000,0x7FFF,M254_read);
	SetWriteHandlerI(2,0x8000,0x9FFE,M254_write);
}

void BTL_PikachuY2K_Init(CartInfo *info)
{
	GenMMC3_Init(info,128,128,8,info->battery);
	info->Power = M254Power;
	info->Reset = M254Reset;
	AddExState(MMC3EXPREGS,1,0,"EXPR");
}

/* ---- City Fighter (VRC hack) */

extern void RawVRC6_Init(void);
extern void RawVRC6_Set(uint8 v);

#define MMCCHR reg
static int32 IRQCount;
static uint8 IRQa;
static uint8 MMCPRG, MMCpmode, MMCmir;

static SFORMAT UNLCF_StateRegs[] =
{
	{&IRQa,1,"IRQA"},
	{&IRQCount,4,"IRQC"},
	{&MMCPRG,1,"PREG"},
	{&MMCpmode,1,"PMOD"},
	{&MMCmir,1,"MIRR"},
	{MMCCHR,8,"CREG"},
	{0}
};

static void PRGSync(void)
{
	setprg32(0x8000,MMCPRG>>2);
	if (!MMCpmode)
		setprg8(0xC000,MMCPRG);
}

static void CHRSync(void)
{
	int		A, x;

	for(x=A=0; x<8; ++x,A+=0x400)
		setchr1(A,MMCCHR[x]);
	setmirror(mirVH01[MMCmir&3]);
}

static DECLFW(UNLCFWrite9)
{
	if (A&12)
		if (A&0x800)
			RawVRC6_Set(V);
		else
			goto prg;
	else
	{
		MMCmir = V&3;
		CHRSync();
	prg:	MMCPRG = V&0xC;
		PRGSync();
	}
}

static DECLFW(UNLCFWriteC)
{
(void)A;
	MMCpmode = V&1;
	PRGSync();
}

static DECLFW(UNLCFWriteA)
{
	static const int8 atab[] = {2,4,0,0,6};

	int		x = atab[(A>>12)-10]|(A>>3&1), sh = A&4;

	MMCCHR[x] = (MMCCHR[x]&0xF0>>sh)|(V&15)<<sh;
	CHRSync();
}

static DECLFW(UNLCFWriteF)
{
	V&= 15;
	if (!(A&=12))
		IRQCount = (IRQCount&0x1E0)|(V+V);
	else if (!(A-=4))
		IRQCount = (IRQCount&0x1E)|V<<5;
	else if (!(A-=4))
		{IRQa = V&2; X65xx_IRQEnd(FCEU_IQEXT);}
}

static void UNLCFIRQ(int a)
{
	if (IRQa && (IRQCount-=a)<=0)
		X65xx_IRQBegin(FCEU_IQEXT);
}

static void UNLCFRestore(int v)
{
(void)v;
	PRGSync();
	CHRSync();
}

static void UNLCFPower(void)
{
	RawVRC6_Init();
	MMCPRG = 0;
	UNLCFRestore(0);
	SetReadHandler(0x8000,0xFFFF,CartBR);
	SetWriteHandler(0x8000,0x9FFF,UNLCFWrite9);
	SetWriteHandler(0xA000,0xBFFF,UNLCFWriteA);
	SetWriteHandler(0xC000,0xCFFF,UNLCFWriteC);
	SetWriteHandler(0xD000,0xEFFF,UNLCFWriteA);
	SetWriteHandler(0xF000,0xFFFF,UNLCFWriteF);
}

void UNL_CF_Init(CartInfo *info)
{
	info->Power = UNLCFPower;
	MapIRQHook = UNLCFIRQ;
	GameStateRestore = UNLCFRestore;
	AddExState(UNLCF_StateRegs,~0,0,0);
}

/* ---- Bubble Bobble (FDS hack) */

static SFORMAT UNLBB_StateRegs[] =
{
	{&reg[0],1,"REG"},
	{&reg[1],1,"CHR"},
	{0}
};

static void UNLBBSync(int v)
{
(void)v;
	setprg8(0x6000,reg[0]&3);
	setprg32(0x8000,~0);
	setchr8(reg[1]&3);
}

static DECLFW(UNLBBWrite)
{
	A&= 0x9000;
	if (A==0x8000)
		reg[0] = reg[1] = V;
	else
		reg[1] = V&1; /* ProWres */
	UNLBBSync(0);
}

static void UNLBBPower(void)
{
	reg[0] =~0;
	reg[1] = 0;
	SetCartRead(0x6000,0xFFFF);
	SetWriteHandler(0x8000,0xFFFF,UNLBBWrite);
	UNLBBSync(0);
}

void UNL_BB_Init(CartInfo *info)
{
	info->Power = UNLBBPower;
	GameStateRestore = UNLBBSync;
	AddExState(UNLBB_StateRegs,~0,0,0);
}
