/*
 cart.c

 * Copyright notice for this file:
 *  cart.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "driver.h"
#include "FCEU.h"
#include "PPU.h"

#include "cart.h"
#include "memory.h"

#include "general.h"

/*
   This file contains all code for coordinating the mapping in of the
 address space external to the NES.
   It's also (ab)used by the NSF code.
*/

uint32	ROM_size, VROM_size; /* totals (prev. in ines.c); *16KB PRG, *8KB CHR */

int16	PRGBank2[32];
uint8	*IOPage[32], *VPage[8];
uint8	**VPageR = VPage;
uint8	*MMC5SPRVPage[8];
uint8	*MMC5BGVPage[8];

uint8	PRGIsRAM[32]; /* This page is/is not PRG RAM. */

/* 16 are (sort of) reserved for UNIF/iNES and 16 to map other stuff. */
uint8	CartCHRram[32];
uint8	CartPRGram[32];

uint8	*PRGptr[32];
uint8	*CHRptr[32];

uint32	PRGsize[32];
uint32	CHRsize[32];

uint32	PRGmask2[32];
uint32	PRGmask4[32];
uint32	PRGmask8[32];
uint32	PRGmask16[32];
uint32	PRGmask32[32];

uint32	CHRmask1[32];
uint32	CHRmask2[32];
uint32	CHRmask4[32];
uint32	CHRmask8[32];

/* can't we make it 16 entries instead of 32? */
DECLFR(CartBR00) {return IOPage[0][A];} DECLFR(CartBR08) {return IOPage[1][A];}
DECLFR(CartBR10) {return IOPage[2][A];} DECLFR(CartBR18) {return IOPage[3][A];}
DECLFR(CartBR20) {return IOPage[4][A];} DECLFR(CartBR28) {return IOPage[5][A];}
DECLFR(CartBR30) {return IOPage[6][A];} DECLFR(CartBR38) {return IOPage[7][A];}
DECLFR(CartBR40) {return IOPage[8][A];} DECLFR(CartBR48) {return IOPage[9][A];}
DECLFR(CartBR50) {return IOPage[10][A];} DECLFR(CartBR58) {return IOPage[11][A];}
DECLFR(CartBR60) {return IOPage[12][A];} DECLFR(CartBR68) {return IOPage[13][A];}
DECLFR(CartBR70) {return IOPage[14][A];} DECLFR(CartBR78) {return IOPage[15][A];}
DECLFR(CartBR80) {return IOPage[16][A];} DECLFR(CartBR88) {return IOPage[17][A];}
DECLFR(CartBR90) {return IOPage[18][A];} DECLFR(CartBR98) {return IOPage[19][A];}
DECLFR(CartBRA0) {return IOPage[20][A];} DECLFR(CartBRA8) {return IOPage[21][A];}
DECLFR(CartBRB0) {return IOPage[22][A];} DECLFR(CartBRB8) {return IOPage[23][A];}
DECLFR(CartBRC0) {return IOPage[24][A];} DECLFR(CartBRC8) {return IOPage[25][A];}
DECLFR(CartBRD0) {return IOPage[26][A];} DECLFR(CartBRD8) {return IOPage[27][A];}
DECLFR(CartBRE0) {return IOPage[28][A];} DECLFR(CartBRE8) {return IOPage[29][A];}
DECLFR(CartBRF0) {return IOPage[30][A];} DECLFR(CartBRF8) {return IOPage[31][A];}

void FASTAPASS(2) SetCartRead(uint16 a,
	uint16 e)
{
	uint16		v, n;

	if (a>e)
		return;
	switch(v=a&0xF800)
	{
	case 0x0000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR00);
		if (n) break; a = ++v;
	case 0x0800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR08);
		if (n) break; a = ++v;
	case 0x1000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR10);
		if (n) break; a = ++v;
	case 0x1800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR18);
		if (n) break; a = ++v;
	case 0x2000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR20);
		if (n) break; a = ++v;
	case 0x2800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR28);
		if (n) break; a = ++v;
	case 0x3000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR30);
		if (n) break; a = ++v;
	case 0x3800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR38);
		if (n) break; a = ++v;
	case 0x4000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR40);
		if (n) break; a = ++v;
	case 0x4800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR48);
		if (n) break; a = ++v;
	case 0x5000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR50);
		if (n) break; a = ++v;
	case 0x5800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR58);
		if (n) break; a = ++v;
	case 0x6000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR60);
		if (n) break; a = ++v;
	case 0x6800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR68);
		if (n) break; a = ++v;
	case 0x7000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR70);
		if (n) break; a = ++v;
	case 0x7800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR78);
		if (n) break; a = ++v;
	case 0x8000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR80);
		if (n) break; a = ++v;
	case 0x8800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR88);
		if (n) break; a = ++v;
	case 0x9000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR90);
		if (n) break; a = ++v;
	case 0x9800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBR98);
		if (n) break; a = ++v;
	case 0xA000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRA0);
		if (n) break; a = ++v;
	case 0xA800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRA8);
		if (n) break; a = ++v;
	case 0xB000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRB0);
		if (n) break; a = ++v;
	case 0xB800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRB8);
		if (n) break; a = ++v;
	case 0xC000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRC0);
		if (n) break; a = ++v;
	case 0xC800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRC8);
		if (n) break; a = ++v;
	case 0xD000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRD0);
		if (n) break; a = ++v;
	case 0xD800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRD8);
		if (n) break; a = ++v;
	case 0xE000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRE0);
		if (n) break; a = ++v;
	case 0xE800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRE8);
		if (n) break; a = ++v;
	case 0xF000: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRF0);
		if (n) break; a = ++v;
	case 0xF800: SetReadHandler(a,(v|=0x7FF)<e ? (n=0,v):(n=1,e),CartBRF8);
	}
}

DECLFR(CartBR)
{
	return IOPage[A>>11][A];
}

DECLFW(CartBW)
{
	uint32		AB = A>>11;

	if (PRGIsRAM[AB] && IOPage[AB])
		IOPage[AB][A] = V;
}

DECLFR(CartBROB)
{
	uint32		AB = A>>11;

	if (!IOPage[AB])
		return A>>8; //XFrame.DB
	return IOPage[AB][A];
}

#define setpageptr(s,A,p,V,r) \
	do{uint8 *_p; uint32 _AB=A>>11; int _x,_r=CartPRGram[r]; \
	if((s)>>1>0) \
		if(_AB+=(_x=((s)>>1))-1,(_p=p)==0) \
			do \
				PRGIsRAM[_AB]=0,IOPage[_AB]=0,PRGBank2[_AB]=-1; \
			while(--_AB,--_x); \
		else { \
			_p-=A; \
			do \
				PRGIsRAM[_AB]=_r,IOPage[_AB]=_p,PRGBank2[_AB]=V; \
			while(--_AB,--_x); \
		} \
}while(0)

static uint8	nothing[8192];

void ResetCartMapping(void)
{
	int		x;

	memset(PRGBank2,~0,sizeof(PRGBank2));
	x = 31;
	do
		IOPage[x] = nothing-(x<<11),
		PRGptr[x] = CHRptr[x] = 0,
		PRGsize[x] = CHRsize[x] = 0;
	while(--x>=0);
	x = 7;
	do
		MMC5SPRVPage[x] = MMC5BGVPage[x] = VPageR[x] = nothing-(x<<10);
	while(--x>=0);
}

void SetupCartPRGMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram)
{
	PRGptr[chip] = p;
	PRGsize[chip] = size;

	PRGmask2[chip] = (size>>11)-1;
	PRGmask4[chip] = (size>>12)-1;
	PRGmask8[chip] = (size>>13)-1;
	PRGmask16[chip] = (size>>14)-1;
	PRGmask32[chip] = (size>>15)-1;

	CartPRGram[chip] = ram ? 1:0;
}

void SetupCartCHRMapping(int chip,
	uint8 *p,
	uint32 size,
	int ram)
{
	chip&= 31; /* say NO to crash happiness; 2010.04.12 */
	CHRptr[chip] = p;
	CHRsize[chip] = size;

	CHRmask1[chip] = (size>>10)-1;
	CHRmask2[chip] = (size>>11)-1;
	CHRmask4[chip] = (size>>12)-1;
	CHRmask8[chip] = (size>>13)-1;

	CartCHRram[chip] = ram;
}

//added these 3; 2010.03.12
int NewCHRRAM(int w,
	uint32 z)
{
	uint8		*r;

	if ((r=(uint8*)FCEU_gmalloc(z))==0)
		return 0;
	SetupCartCHRMapping(w,r,z,1);

	return 1;
}

void DeleteCHRRAM(int w)
{
	if (CHRptr[w] && CartCHRram[w])
		{FCEU_free(CHRptr[w]); CHRptr[w] = nothing, CartCHRram[w] = 0;}
}

void EraseCHRRAM(int w,
	int32 v)
{
	if (CHRptr[w] && CartCHRram[w])
		memset(CHRptr[w],v,CHRsize[w]);
}

void FASTAPASS(3) setprg4r(int r,
	unsigned int A,
	unsigned int V)
{
	V&= PRGmask4[r];
	setpageptr(4,A,PRGptr[r]==0 ? 0:&PRGptr[r][V<<12],V,r);
}

void FASTAPASS(3) setprg8r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<13)
	{
		V&= PRGmask8[r];
		setpageptr(8,A,p==0 ? 0:&p[V<<13],V,r);
	} else
	{
		uint32		VA = V<<2;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,p==0 ? 0:&p[((VA+x)&m)<<11],V,r);
		while(A+=1<<11,++x<4);
	}
}

void FASTAPASS(3) setprg16r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<14)
	{
		V&= PRGmask16[r];
		setpageptr(16,A,p==0 ? 0:&p[V<<14],V,r);
	} else
	{
		uint32		VA = V<<3;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,p==0 ? 0:&p[((VA+x)&m)<<11],V,r);
		while(A+=1<<11,++x<8);
	}
}

void FASTAPASS(3) setprg32r(int r,
	unsigned int A,
	unsigned int V)
{
	uint8		*p = PRGptr[r];

	if (PRGsize[r]>=1<<15)
	{
		V&= PRGmask32[r];
		setpageptr(32,A,p==0 ? 0:&p[V<<15],V,r);
	} else
	{
		uint32		VA = V<<4;
		int		x = 0, m = PRGmask2[r];

		do
			setpageptr(2,A,p==0 ? 0:&p[((VA+x)&m)<<11],V,r);
		while(A+=1<<11,++x<16);
	}
}

void FASTAPASS(3) setchr1r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (CHRptr[r]==0)
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask1[r];
	VPageR[AB] = &CHRptr[r][V<<10]-A;

	if (CartCHRram[r])
		PPUCHRRAM|= 1<<AB;
	else
		PPUCHRRAM&=~(1<<AB);
}

void FASTAPASS(3) setchr2r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (CHRptr[r]==0)
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask2[r];
	VPageR[AB] = VPageR[AB+1] = &CHRptr[r][V<<11]-(A);

	if (CartCHRram[r])
		PPUCHRRAM|= 3<<AB;
	else
		PPUCHRRAM&=~3<<AB;
}

void FASTAPASS(3) setchr4r(int r,
	unsigned int A,
	unsigned int V)
{
	uint32		AB;

	if (CHRptr[r]==0)
		return;
	FCEUPPU_LineUpdate();

	AB = A>>10, V&= CHRmask4[r];
	VPageR[AB] = VPageR[AB+1] =
		VPageR[AB+2] = VPageR[AB+3] = &CHRptr[r][V<<12]-(A);

	if (CartCHRram[r])
		PPUCHRRAM|= 15<<AB;
	else
		PPUCHRRAM&=~(15<<AB);
}

void FASTAPASS(2) setchr8r(int r,
	unsigned int V)
{
	if (CHRptr[r]==0)
		return;

	FCEUPPU_LineUpdate();

	V&= CHRmask8[r], V<<= 13;
	setvram8(&CHRptr[r][V]);

	if (CartCHRram[r])
		PPUCHRRAM|= 0xFF;
	else
		PPUCHRRAM&=~0xFF;
}

void FASTAPASS(2) setvram4(uint32 A,
	uint8 *p)
{
	uint32		AB = A>>10;
	register uint8	**w = AB+VPageR;

	p-= A;
	*w++ = p, *w++ = p, *w++ = p, *w = p;
	PPUCHRRAM|= 15<<AB;
}

void FASTAPASS(1) setvram8(uint8 *p)
{
	register uint8	**w = VPageR;

	*w++ = p, *w++ = p, *w++ = p, *w++ = p,
	*w++ = p, *w++ = p, *w++ = p, *w = p;
	PPUCHRRAM|= 0xFF;
}

void FASTAPASS(3) setvramb1(uint8 *p,
	uint32 A,
	uint32 b)
{
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10;
	VPageR[AB] = p+(b<<10)-A;
	PPUCHRRAM|= 1<<AB;
}

void FASTAPASS(3) setvramb2(uint8 *p,
	uint32 A,
	uint32 b)
{
	uint8		**w;
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10, w = AB+VPageR;
	w[0] = w[1] = p+(b<<11)-A;
	PPUCHRRAM|= 3<<AB;
}

void FASTAPASS(3) setvramb4(uint8 *p,
	uint32 A,
	uint32 b)
{
	uint8		**w;
	uint32		AB;

	FCEUPPU_LineUpdate();

	AB = A>>10, w = AB+VPageR;
	w[0] = w[1] = w[2] = w[3] = p+(b<<12)-A;
	PPUCHRRAM|= 15<<AB;
}

void FASTAPASS(2) setvramb8(uint8 *p,
	uint32 b)
{
	uint8		**w;

	FCEUPPU_LineUpdate();

	w = VPageR;
	w[0] = w[1] = w[2] = w[3] = w[4] = w[5] = w[6] = w[7] = p+(b<<13);

	PPUCHRRAM|= 0xFF;
}

/* This function can be called without calling SetupCartMirroring(). */

void FASTAPASS(3) setntamem(uint8 *p,
	int ram,
	uint32 b)
{
	FCEUPPU_LineUpdate();

	vnapage[b] = p;
	PPUNTARAM&=~(1<<b);
	if (ram)
		PPUNTARAM|= 1<<b;
}

static int	mirrorhard = 0;

void setmirrorw(int a,
	int b,
	int c,
	int d)
{
	FCEUPPU_LineUpdate();

	vnapage[0] = NTARAM+a*0x400;
	vnapage[1] = NTARAM+b*0x400;
	vnapage[2] = NTARAM+c*0x400;
	vnapage[3] = NTARAM+d*0x400;
}

void FASTAPASS(1) setmirror(int t)
{
	FCEUPPU_LineUpdate();

	if (mirrorhard)
		return;
//	t = 0xF053>>(t<<2&12)&3^(t&12);
//	vnapage[0] = NTARAM+(t<<10&0x400);
//	vnapage[1] = NTARAM+(t<<9&0x400);
//	vnapage[2] = NTARAM+(t<<8&0x400);
//	vnapage[3] = NTARAM+(t<<7&0x400);

	switch(t)
	{
	case MI_H:
		vnapage[0] = vnapage[1] = NTARAM;
		vnapage[2] = vnapage[3] = NTARAM+0x400;
		break;
	case MI_V:
		vnapage[0] = vnapage[2] = NTARAM;
		vnapage[1] = vnapage[3] = NTARAM+0x400;
		break;
	case MI_0:
		vnapage[0] = vnapage[1] = vnapage[2] = vnapage[3] = NTARAM;
		break;
	case MI_1:
		vnapage[0] = vnapage[1] = vnapage[2] = vnapage[3] = NTARAM+0x400;
	}
	PPUNTARAM = 15;
}

void SetupCartMirroring(int m,
	int hard,
	uint8 *extra)
{
	if (m<4)
	{
		mirrorhard = 0;
		setmirror(m);
	} else
	{
		vnapage[0] = NTARAM;
		vnapage[1] = NTARAM+0x400;
		vnapage[2] = extra;
		vnapage[3] = extra+0x400;
		PPUNTARAM = 15;
	}

	mirrorhard = hard;
}

void FCEU_SaveGameSave(CartInfo *LocalHWInfo)
{
	FILE		*sp;
	char		*soot;

	if (!LocalHWInfo->battery || LocalHWInfo->SaveGame[0]==0)
		return;
	if ((soot=FCEU_MakeFName(FCEUMKF_SAV,0,"sav"))==0)
		return;
	sp = FCEUD_UTF8fopen(soot,"wb");
	free(soot);
	if (sp)
	{
		int	x = 0;

		do
			if (LocalHWInfo->SaveGame[x])
				fwrite(LocalHWInfo->SaveGame[x],1,
					LocalHWInfo->SaveGameLen[x],sp);
		while(++x<4);
		fclose(sp);
	} else
		FCEU_PrintError("WRAM file \"%s\" cannot be written to.\n",soot);
}

void FCEU_LoadGameSave(CartInfo *LocalHWInfo)
{
	FILE		*sp;
	char		*soot;

	if (!LocalHWInfo->battery || LocalHWInfo->SaveGame[0]==0)
		return;

	if ((soot=FCEU_MakeFName(FCEUMKF_SAV,0,"sav"))==0)
		return;
	sp = FCEUD_UTF8fopen(soot,"rb");
	free(soot);
	if (sp)
	{
		int	x = 0;

		do
			if (LocalHWInfo->SaveGame[x])
				fread(LocalHWInfo->SaveGame[x],1,
					LocalHWInfo->SaveGameLen[x],sp);
		while(++x<4);
		fclose(sp);
	}
}
