/*
 fsound.h

 * MacFCEU 0.9a2
 *
 * Copyright notice for this file:
 *  sound.h Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* ---- constants */

/* expansion indices */

#define SCVRC6			0
#define SCVRC7			1
#define SCNFDS			2
#define SCMMC5			3
#define SCN163			4
#define SCSS5B			5

/* APU-specific */

/* rectangular envelope: 4000/4/C */
#define REnvSpeedMask		0xF
#define REnvSpeedAmp		0x10 /* amplitude is speed register, not amp sweep reg. */
#define REnvelopeLoop		0x20
#define RDutyTypeShift		6
/* rectangle sweep: 4001/5h */
#define RSweepEnable		0x80
#define kRSweepPeriod		4 /* bits 4-6 */
#define RSweepPeriodMask	7
#define RSweepMask		7
#define RSweepDir		8 /* direction - 1=subtract; 0=add */
/* rectangle wave time: 4002/3/6/7 */
#define RWaveLnMin		8u
#define RLnCtrShift		3

/* triangle: 4008h */
#define TLoadMask		0x7F
#define TStartMask		0x80

/* rectangular noise: 400Eh */
#define NShiftShort		0x80

/* 4010h */
#define DMCFEnable		0x80
#define DMCFLoop		0x40
#define DMCFMask		0xF

/* 4015h reads */
#define DMCSIRQ			0x80
#define FrameSIRQ		0x40
#define RdyClSIRQ		0x20

/* 4017h */
#define FMode5Step		0x80
#define FModeNoIRQ		0x40
#define FModeMask		0xC0

/* ---- typedefs */

typedef void (*voidfunc)(void);
typedef void (*fillfunc)(int32 c);

typedef struct EXPSOUND
{
	fillfunc Fill[6]; /* low/mid-quality expansion sound */
	voidfunc HiFill[6];
	fillfunc HiSync[6];
	voidfunc RChange[6];

	/* NeoFill is for sound devices (only VRC7, currently)
	   that are emulated in a more high-level manner in HQ mode.
	   Interestingly, this device has slightly better sound quality
	   (updated more often) in lq mode than in high-quality mode.
	   Maybe that should be fixed. :)
	*/
	void (*NeoFill)(int16 *Wave,
		int Count);
	void (*Kill)(void);
} EXPSOUND;

/* ---- externs */

extern EXPSOUND GameExpSound;

extern uint32 RecCurFreq[2];
extern int32 nesincsize, Soundfccnt, APU_DMCacc;
extern uint8 PSGReg[16];
extern uint8 SIRQStat, DMADualRead, REnvLoopEnable, NShShortEnable, RNoDutyReset;
extern uint8 APUSilenceWatch, APUFlatWave;

/* WaveFinal's static */
extern int32 FCEUWave[13472];

extern uint32 soundtsinc;
extern uint32 soundtsoffs;

#define SOUNDTS			(timestamp+soundtsoffs)

/* ---- prototypes */

#define APU_DMCClock(x) do{ \
	if ((APU_DMCacc-=x)<=0) \
		APU_DMCClockDAC(); \
}while(0)

/* this may be removed, due to the fact that the APU has no IRQ clock */
#define APU_IRQClock(x) do{ \
	if (!((XFStatL&=0x8000FFFF)<0) && (XFStatL-=x)<0) \
		APU_FrameIRQ(); \
}while(0)

#define APU_FrameClock(x) do{ \
	if ((Soundfccnt-=x)<0) \
		APU_FrameUpdate(); /* updates Soundfccnt */ \
}while(0)

void FCEU_SetSoundVariables(void);

/* sound output is now int16, as data was already clipped that range */
int16 *FCEU_GetSoundBuffer(int32 *c);
int FlushEmulateSound(void);

void APU_MapIO(void);
void APU_FrameUpdate(void);
void APU_FrameIRQ(void);
void APU_DMCClockDAC(void);
void APU_DMCDMA(void);

void APU_Power(void);
void APU_Reset(void);
void FCEUAPU_SaveState(void);
void FCEUAPU_LoadState(int version);

uint8 FCEUAPU_RegPeek(uint16 A);
void FCEUAPU_RegPoke(uint16 A,
	uint8 V,
	int hl);
