/*
 input/FKB.c

 * Copyright notice for this file:
 *  FKB.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "share.h"
#include "FKB.h"

#define AK2(x,y)	((FKB_##x)|(FKB_##y <<8))
#define AK(x)		FKB_##x

uint8		bufit[0x66];
uint8		ksmode, ksindex, capsstrobe;

static uint8	fkbmatrix[9][2][4]=
{
{{AK(F8),AK(RETURN),AK(LBRACKET),AK(RBRACKET)},   
	{AK(KANA),AK(RIGHTSHIFT),AK(BACKSLASH),AK(STOP)}},
{{AK(F7),AK(AT),AK(COLON),AK(SEMICOLON)}, 		
	{AK(UNDERSCORE),AK(SLASH),AK(MINUS),AK(CARET)}},
{{AK(F6),AK(O),AK(L),AK(K)},   
	{AK(PERIOD),AK(COMMA),AK(P),AK(0)}},
{{AK(F5),AK(I),AK(U),AK(J)},   	
	{AK(M),AK(N),AK(9),AK(8)}},
{{AK(F4),AK(Y),AK(G),AK(H)},  
	{AK(B),AK(V),AK(7),AK(6)}},
{{AK(F3),AK(T),AK(R),AK(D)},  
	{AK(F),AK(C),AK(5),AK(4)}},
{{AK(F2),AK(W),AK(S),AK(A)},  
	{AK(X),AK(Z),AK(E),AK(3)}},
{{AK(F1),AK(ESCAPE),AK(Q),AK(CONTROL)},    
	{AK(LEFTSHIFT),AK(GRAPH),AK(1),AK(2)}},
{{AK(CLEAR),AK(UP),AK(RIGHT),AK(LEFT)},    
	{AK(DOWN),AK(SPACE),AK(DELETE),AK(INSERT)}},
};

static void FP_FASTAPASS(1) FKB_Write(uint8 v)
{
	v>>= 1;
	if (v&2)
		if (ksmode && !(v&1))
			++ksindex, ksindex&= ((int)ksindex-9)>>8;
	ksmode = v&1;
}

static uint8 FP_FASTAPASS(2) FKB_Read(int w,
	uint8 ret)
{
//printf("$%04x, %d, %d\n",w+0x4016,ksindex,ksmode);
	if (w)
	{
		int		x;

		x = 0, ret&=~0x1E;
		do
			if (bufit[fkbmatrix[ksindex][ksmode][x]])
				ret|= 1<<(x+1);
		while(++x<4);
	}
	ret^= 0x1E;

	return ret;
}

static void FKB_Strobe(void)
{
	ksmode = ksindex = 0;
	//printf("strobe\n");
}

static void FP_FASTAPASS(2) FKB_Update(void *data,
	int arg)
{
(void)arg;
	memcpy(bufit+1,data,0x48);
}

static INPUTCFC FKB = {FKB_Read,FKB_Write,FKB_Strobe,FKB_Update};

INPUTCFC *FCEU_InitFKB(void)
{
	memset(bufit,0,sizeof(bufit));
	ksmode = ksindex = 0;
	return &FKB;
}

#undef AK
#undef AK2

/* ---- Subor variant */
/* similar to Family Keyboard, but has a different table and index depth */

#define AK(x)		SKB_##x

static uint8 skbmatrix[13][2][4] =
{
{ { AK(4), AK(G), AK(F), AK(C) },
	{ AK(F2), AK(E), AK(5), AK(V) } },
{ { AK(2), AK(D), AK(S), AK(END) },
	{ AK(F1), AK(W), AK(3), AK(X) } },
{ { AK(INSERT), AK(BACK), AK(NEXT), AK(RIGHT) },
	{ AK(F8), AK(PRIOR), AK(DELETE), AK(HOME) } },
{ { AK(9), AK(I), AK(L), AK(COMMA) },
	{ AK(F5), AK(O), AK(0), AK(PERIOD) } },
{ { AK(RBRACKET), AK(RETURN), AK(UP), AK(LEFT) },
	{ AK(F7), AK(LBRACKET), AK(BACKSLASH), AK(DOWN) } },
{ { AK(Q), AK(CAPITAL), AK(Z), AK(TAB) },
	{ AK(ESCAPE), AK(A), AK(1), AK(LCONTROL) } },
{ { AK(7), AK(Y), AK(K), AK(M) },
	{ AK(F4), AK(U), AK(8), AK(J) } },
{ { AK(MINUS), AK(SEMICOLON), AK(APOSTROPHE), AK(SLASH) },
	{ AK(F6), AK(P), AK(EQUALS), AK(LSHIFT) } },
{ { AK(T), AK(H), AK(N), AK(SPACE) },
	{ AK(F3), AK(R), AK(6), AK(B) } },
{ { AK(NUMPAD6), AK(GRETURN), AK(NUMPAD4), AK(NUMPAD8) },
 // baibaidino actually uses diferent layot
	{ AK(NUMPAD2), AK(RCONTROL), AK(RSHIFT), AK(RMENU) } }, /* last three here are filler... */
{ { AK(LMENU), AK(NUMPAD4), AK(NUMPAD7), AK(F11) },
	{ AK(F12), AK(NUMPAD1), AK(NUMPAD2), AK(NUMPAD8) } },
{ { AK(SUBTRACT), AK(ADD), AK(MULTIPLY), AK(NUMPAD9) },
	{ AK(F10), AK(NUMPAD5), AK(DIVIDE), AK(NUMLOCK) } },
{ { AK(GRAVE), AK(NUMPAD6), AK(PAUSE), AK(SPACE) },
	{ AK(F9), AK(NUMPAD3), AK(DECIMAL), AK(NUMPAD0) } },
};

static void FP_FASTAPASS(1) SuborKB_Write(uint8 v)
{
	uint8		reset = v&1;

	v>>= 1;
	if (v&2) /* command */
	{
		if (ksmode && !(v&1)) /* scan */
			++ksindex, ksindex&= ((int)ksindex-13)>>8;
		if (reset)
			ksindex = 0;
	}
	ksmode = v&1;
}

static uint8 FP_FASTAPASS(2) SuborKB_Read(int w,
	uint8 ret)
{
	if (w)
	{
		int		x, v;

		x = 0, ret&=~0x1E;
		do
		{
			v = skbmatrix[ksindex][ksmode][x];
			if (v==AK(CAPITAL))
				v = capsstrobe;
			else
				v = bufit[v];
			if (v)
				ret|= 1<<(x+1);
		} while(++x<4);
		ret^= 0x1E;
	}
	return ret;
}

static void FP_FASTAPASS(2) SuborKB_Update(void *data,
	int w)
{
	uint8		old = bufit[AK(CAPITAL)];
(void)w;
	memcpy(bufit+1,data,sizeof(bufit)-1);
	capsstrobe = old!=bufit[AK(CAPITAL)];
}

static INPUTCFC SuborKB = {SuborKB_Read, SuborKB_Write, FKB_Strobe, SuborKB_Update};

INPUTCFC *FCEU_InitSuborKB(void)
{
	FCEU_InitFKB();
	return &SuborKB;
}
