/*
 macos/memview.c

  Created: 2015.07.12
  Revised: 2015.10.29

Sunday, July 12, 2015
  created.  based on ideas with the FCEUX RAM editor.
*/

#include <stdio.h>

#include <MacWindows.h>
#include <Events.h>
#include <Controls.h>
#include <ControlDefinitions.h>

#include "../types.h"
#include "../fsound.h"
#include "../PPU.h"

#include "shell.h"
#include "shellmenus.h"
#include "macvideo.h"
#include "memview.h"
#include "listwindow.h"
#include "../utility/disasm6502.h"

/* ---- constants */

#define PASCAL_MEMVIEW_TITLE		"\pMemory View"
#define PASCAL_MEMVIEW_FONTNAME		"\pMonaco"
#define TEXT_FONT_SIZE			9

#define FIRST_DATA_CHAR			10
#define MAXIMUM_LINECOUNT		(0x10000/16)
#define INIT_MEMVIEW_HEIGHT		(11*11+15)

#define MEMVIEW_MINIMUM_HEIGHT		(INIT_MEMVIEW_HEIGHT-24)
#define MEMVIEW_MAXIMUM_HEIGHT		(4096+32)
#define MEMVIEW_LINELENGTH		(8*5-1+FIRST_DATA_CHAR)
#define MEMVIEW_WIDTH			(MEMVIEW_LINELENGTH*6+4+18)

/* ---- globals */

WindowRef	MemViewWindow;

static uint8 (*MemViewRead)(uint16 A) = FCEUI_MemSafePeek;
static void (*MemViewWrite)(uint16 A, uint8, int highlevel) = FCEUI_MemPoke;
static int (*MemGetBank)(uint16 A) = FCEUI_MemGetBank;

/* ---- code */

static const char *print_line(char *d,
	int32 *z,
	int i)
{
	static char	b[MEMVIEW_LINELENGTH];
	char		*s;
	int		c;
#pragma unused(d)
	i<<= 4, c = MemGetBank ? MemGetBank(i):~0;
	*(int32*)&b = MK4CC(':',':',':',':');
	if (c<4096u)
		Disasm_HexStr(b,c,3);
	Disasm_HexStr(b+4,i,4);
	c = 8, s = b+(FIRST_DATA_CHAR-1), s[-1] = ' ';
	do
		*s = ' ', Disasm_HexStr(s+1,MemViewRead(i)<<8|MemViewRead(i+1),4);
	while(s+=5,i+=2,--c);
	*z = sizeof(b);
	return b;
}

static void click_line(int i,
	int x,
	int *ss,
	int *se)
{
#pragma unused(i)
	x-= FIRST_DATA_CHAR, x = x*((131071+4)/5)>>16;
	if (x>15u)
		return;
	x+= x, x+= x>>2, x+= FIRST_DATA_CHAR;
	*ss = x;
	*se = x+2;
}

#if 0
static void memview_createmenu(WindowRef win)
{
	Rect		ctlbounds;
	ControlRef	ctl;

	ctlbounds.left = 20;
	ctlbounds.top = 2;
	ctlbounds.right = 100;
	ctlbounds.right = 18;
	ctl = NewControl(win,&ctlbounds,"\pView",true,
		0,menuid_view,40,popupMenuProc,0);
}
#endif

static WindowRef memview_create(void)
{
	Rect		bounds;
	WindowRef	win;
	/* bottom-left of screen, above message log */
	bounds = (**gamegdevice).gdRect;
	bounds.top = (bounds.bottom-=9+INIT_MEMVIEW_HEIGHT)-INIT_MEMVIEW_HEIGHT;
	bounds.left = bounds.left+9;
	bounds.right = bounds.left+MEMVIEW_WIDTH;

	//getpstr(title,strlistid_common,commonstr_memoryviewtitle);
	if ((win=listwindow_create((WindowRef)-1,&bounds,PASCAL_MEMVIEW_TITLE,
			PASCAL_MEMVIEW_FONTNAME,TEXT_FONT_SIZE,
			0,MAXIMUM_LINECOUNT,0,1,MK4CC('r','a','m','v')))==NULL)
		return NULL;
	listwindow_settexthook(win,print_line,MAXIMUM_LINECOUNT);
	listwindow_setclickhook(win,click_line);
	listwindow_setindent(win,3,0);
	memview_setview(_view_mode_cpu);
	return win;
}

int memview_show(int w)
{
	int		r, wasvis;

	if (!MemViewWindow && w!=IMIN(w) && (MemViewWindow=memview_create())==NULL)
		return -1;
	wasvis = !!getvisiblewindowport(MemViewWindow);
	r = showhide(MemViewWindow,w);
	if (r!=wasvis)
	{
		movemenu(menuid_view,r);
		drawmenubar();
	}
	return r;
}

void memview_update(boolean erase)
{
#if 0
	RGBColor	gray;
	GrafPtr		oldport;

	GetPort(&oldport);
	gray.red = gray.green = gray.blue = 0x7FFF;
	PenMode(srcCopy);
	PenSize(1,1);
	RGBForeColor(&gray);
	MoveTo(0,20);
	LineTo(16000,20);
	SetPort(oldport);
#endif
	listwindow_update(MemViewWindow,erase);
}

int memview_click(EventRecord *event,
	short what)
{
	Rect		bounding;

	bounding.top = MEMVIEW_MINIMUM_HEIGHT;
	bounding.left = MEMVIEW_WIDTH+1;
	bounding.bottom = MEMVIEW_MAXIMUM_HEIGHT;
	bounding.right = MEMVIEW_WIDTH+1;

	return listwindow_click(MemViewWindow,event,what,&bounding);
}

void memview_activate(boolean active)
{
	if (listwindow_setactive(MemViewWindow,active)<0)
		return;
}

int memview_key(int k)
{
	int32		i, x, e;
	int		a;

	if (k==kEscapeCharCode)
	{
		if (MemViewWindow)
			HiliteWindow(MemViewWindow,false);
		return 0;
	}
	if (k-'0'>9u && k-'A'>15u && k-'a'>15u)
		return 0;
	if ((a=listwindow_getsel(MemViewWindow,&i,NULL,&x,&e))<0)
		return a;
	if (e-x>2u)
		return 0; /* no byte selected */
	x-= FIRST_DATA_CHAR, x-= x*((131071+4)/5)>>17;
	if (x>31u)
		return 0;
	if (k>='A')
		k+= 9;
	k&= 15;
	i = i*16u+(x>>1);
	a = MemViewRead(i);
	if (x&1)
		a = (a&0xF0)|k;
	else
		a = (a&15)|k<<4;
	MemViewWrite(i,a,0);
	/* advance and update selection */
	++x, i = (i+(x>>5))>>4, x&= 31, a = x&1;
	x+= x>>2, x+= FIRST_DATA_CHAR;
	listwindow_select(MemViewWindow,i,i,x,x+1+!a,true);
	memview_update(false);
	return 1;
}

int memview_setview(int type)
{
	switch(type)
	{
	case _view_mode_cpu:
		MemViewRead = FCEUI_MemSafePeek;
		MemViewWrite = FCEUI_MemPoke;
		MemGetBank = FCEUI_MemGetBank;
		listwindow_settexthook(MemViewWindow,print_line,MAXIMUM_LINECOUNT);
		break;
	case _view_mode_ppu:
		MemViewRead = FCEUPPU_MemPeek;
		MemViewWrite = FCEUPPU_MemPoke;
		MemGetBank = FCEUPPU_MemGetBank;
		listwindow_settexthook(MemViewWindow,print_line,0x3F20/16);
		break;
	case _view_mode_oam:
		MemViewRead = FCEUPPU_OAMPeek;
		MemViewWrite = FCEUPPU_OAMPoke;
		MemGetBank = NULL;
		listwindow_settexthook(MemViewWindow,print_line,0x100/16);
		break;
	case _view_mode_apu:
		MemViewRead = FCEUAPU_RegPeek;
		MemViewWrite = FCEUAPU_RegPoke;
		MemGetBank = NULL;
		listwindow_settexthook(MemViewWindow,print_line,3);
		break;

	default:
		return -1;
	}
	memview_update(1);
	checkmenu(_view_cpu,type+1-_view_mode_cpu,0);
	return 1;
}
