/*
 macos/menus.c

  Created: 2008.07.08
  Revised: 2017.04.23

Wednesday, August 19, 2009 1:10 AM
  increased/maxed out some option limits (1000-65535 srate, 24-429% for speed).
Friday, October 9, 2009 1:48 PM
  finishing the menubar show/hide code to work properly.
Tuseday, March 23, 2010 1:26 PM
  moved all the rest of the sound stuff declared here out, externs are now in shell.h.
Saturday, March 27, 2010 1:51 PM
  a run-in with a falsely rewritten "Set Enabled" routine prompts me to check validity.
Thursday, April 29, 2010 8:24 AM
  watch cursor before any InitSound (some carts may call some slow init re-process).
Saturday, May 4, 2013 5:19 AM
  max. volumes/scales was +/-10000, now 2000.
*/

#include <stdlib.h>

#include "shell.h"

#include "../FCEU.h"
#include "../common/dface.h"
#include "../WAVE.h"
#include "../FDS.h"

#include "shellmenus.h"
#include "messagelog.h"
#include "macdialogs.h"
#include "macvideo.h"
#include "debugger.h"
#include "config.h"
#include "tileview.h"
#include "memview.h"
#include "macinput.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- globals */

extern struct FCEUGI *GI;

#define MAX_RATE

static const uint32	sndqualities[] =
{
//	8000*16, /* was 8192 */
//	11025*16,
	15000*16,
	22050*16,
	36000*16,
	44100*16, /* added 2013.05.14 */
	66763.63*16,
	88200*16, /* added 2015.08.21 */
	96000*16,
	MAXIMUM_SRATE
};

static const uint32	emulationspeeds[] =
{
	500000000,
	629960525,
	707106781,
	793700526,
	890898718,
	1000000000,
	1259921050,
	1587401052,
	1781797436,
	2000000000
};

/* ---- code */

void redomenusgi(void)
{
	static uint8	simpled[] =
		{/*_file_close, */_file_savestate, _file_loadstate,
		_emulator_reset, _emulator_power, _emulator_step, 0};

	int		x =-1;
	uint8		i, loaded = GI!=NULL;

	while((i=simpled[++x])!=0)
		enablemenuitem(i,loaded);
	enablemenuitem(_emulator_eject,loaded && GI->type==GIT_FDS);
	if (loaded && GI->type!=GIT_VSUNI && GI->type!=GIT_FDS)
		loaded = 0;
	enablemenuitem(_emulator_switch,loaded);
	enablemenuitem(_emulator_insert,loaded);
	loaded&= GI->type==GIT_VSUNI;
	enablemenuitem(_emulator_coin1,loaded);
	enablemenuitem(_emulator_coin2,loaded);
}

void reenablemenus(boolean enable)
{
	uint8		*simpled = (uint8*)"\1\4\2\3\0";

	int		x =-1;
	uint8		i;

	while((i=simpled[++x])!=0)
		enablemenuitem(i<<6,enable);
	enablemenuitem(_apple_about,enable); /* was the whole menu; 2010.04.11 */

	drawmenubar();
}

/* in case of file overflow; 2012.02.15 */
void updatewavestatus(void)
{
	if (FCEU_WaveUpdateHeader()<0)
		setmenuitemstr(_file_logsound,getpstr(temporary,strlistid_common,
			commonstr_logsound));
}

void updatemenus(void)
{
extern uint8 PPUrendis;
	static uint8	*polo[] = {&emuset.usegenie,&emuset.autoips,&vidset.fullscreen,
		&screenshift, &screenrefreshsync,
		&fceutileview, &inputset.fscore, &inputset.mic, &sndset.rdutyrst, &sndset.ddmadis};
	static uint16	simpled[] = {_config_Genie, _config_autoips, _subvideo_fullscreen,
		_subvideo_2x_scale, _subvideo_refresh_sync,
		_subvideo_tile_view, _subinput_fscore, _subinput_mic,
		_subchannels_no_duty_reset, _subchannels_disable_4011h, 0};
	static int8	userinexptab[] = {1,64,64,64,64,2,3,4,5,6,7,8,9,10,11,12,13,14};

	int		x =-1, m;
	uint16		i;

	while((i=simpled[++x])!=0)
		checkmenuitem(i,*polo[x]&1);
	checkmenuitem(_config_snapname,eoptions&EO_SNAPNAME);
	checkmenuitem(_subinput_vsauto,!(eoptions&EO_VSMANUALINPUT));
	/* modern (and kind of dumb but useful) way: auto, or force a model */
	x = emuset.model>=GIV_default+0u ? ~0:emuset.model;
	checkmenu(_submodel_auto,x+((_submodel_auto&63)+1),0);

	checkmenu(_subinputa_none,userinputtype[0]+(_subinputa_none&63),0);
	checkmenu(_subinputb_none,userinputtype[1]+(_subinputb_none&63),0);
	checkmenu(_subinputexp_none,userinexptab[userinputtype[2]],0);

	checkmenu(_subsound_low,x+(_subsound_low&63),0);

	for(x=0,m=sndset.rate*16; x<GET_ARRAY_COUNT(sndqualities); ++x)
		if (sndqualities[x]==m)
			break;
	checkmenu(_subsound_low,x+(_subsound_low&63),0); /* if loop fails, Other will automatically be checked */

	x = soundvolume/10, x|= (unsigned)(soundvolume-x*10)<<8, ++x;
	checkmenu(_subvolume_off,MIN((unsigned)x,17),0);

	checkmenuitem(_subvideo_stretch_filter,screenshift&0x80);
	checkmenuitem(_subvideo_dup_fields,screenshift&0x40);
	enablemenuitem(_subvideo_dup_fields,screenshift&1);
//	checkmenuitem(_subvideo_clip_sides,eoptions&EO_CLIPSIDES);
	checkmenuitem(_subvideo_no_sprite_limit,(eoptions&EO_NOSPRLIM)!=0);
	checkmenuitem(_subvideo_auto_skip_frames,(eoptions&EO_AUTOFRAMESKIP)!=0);
	checkmenuitem(_subvideo_hide_pf,PPUrendis&6);
	checkmenuitem(_subvideo_hide_sf,PPUrendis&1);

	for(x=_subchannels_rect_1,m=1; x<=_subchannels_DMC; ++x,m+=m)
		checkmenuitem(x,sndactivechannels&m);
	checkmenuitem(_subexpchannels_enable,sndactivechannels&0x20);
	for(x=_subexpchannels_VRC6,m+=m; x<=_subexpchannels_SS5B; ++x,m+=m)
	{
		enablemenuitem(x,sndactivechannels&0x20);
		checkmenuitem(x,sndactivechannels&m);
	}
	checkmenuitem(_subexpchannels_n163_maxch,n163_max>255);
	checkmenuitem(_subchannels_alt_rect_duty,sndset.rduty!=RDUTYBMP_STD);

	checkmenuitem(_subspeed_adj_dmc,!(RateFixFlags&0x40));
	checkmenuitem(_subspeed_rate_only,RateFixFlags&0x80);
	for(x=0; x<GET_ARRAY_COUNT(emulationspeeds); ++x)
		if (emulationspeeds[x]==emuset.speed)
			break;
	checkmenu(_subspeed_50,x+(_subspeed_50&63),0);

	checkmenu(_subbackground_off,1+(eoptions&EO_BGRUN)+(eoptions&EO_MUTE ? 2:0),0);

	enablemenuitem(_file_close,forewindow(-1));

	checkmenuitem(_edit_memview,memview_show(IMIN(int)));

	checkmenuitem(_nsf_autoadvance,sndset.nsfauto);
	checkmenuitem(_nsf_loop,sndset.nsfauto==2);
	checkmenuitem(_nsf_loop1x,sndset.nsfauto==3);

	updatewavestatus();
}

static void setfvolume(int16 *v,
	const uint8 *t)
{
	int		tmp;

//	stopsound();
	if ((tmp=valuedialog(t,*v,-2000,2000,false))!=INT_NO_VALUE)
	{
		*v = tmp;
		update_soundsettings(1);
	}
}

boolean domenu(uint16 itemcode)
{
	double		td;
	int		tmp;

	switch(itemcode&0x7FFF)
	{ /* menubar toggle is automatic during fullscreen */
	/* apple */
	case _apple_about:
		//stopsound();
		splash();
		break;

	/* file */
	case _file_open:
		//stopsound();
		opengame(NULL);
		break;
	case _file_close:
		if (GI)
			scexit(_sc_closegame);
		else
			forewindow(0);
		break;
	case _file_savestate:
		if (!GI)
			return false;
		//stopsound();
		SaveStateOrMovie(NULL,itemcode>>15);
		break;
	case _file_loadstate:
		if (!GI)
			return false;
		//stopsound();
		LoadStateOrMovie(NULL,itemcode>>15,false);
		break;
	case _file_logsound:
		setmenuitemstr(_file_logsound,getpstr(temporary,strlistid_common,
			RecordSoundWave(NULL)>0 ? commonstr_stopwave:commonstr_logsound));
		break;
	case _file_saveconfig:
		SaveConfig();
		break;
	case _file_quit:
		scexit(_sc_quit);
		break;

	/* emulator */
	case _emulator_reset:
		FCEUI_Reset(itemcode>>15);
		goto subreset;
	case _emulator_power:
		FCEUI_Power();
	subreset:
		if (debuggerwindow_visible())
			debuggerwindow_step(1); /* start */
		break;
	case _emulator_eject:
		if (GI && GI->type==GIT_FDS)
			FCEU_FDSEject();
		goto end;
	case _emulator_switch:
		if (!GI)
			goto end;
		if (GI->type==GIT_FDS)
			FCEUI_FDSSelect();
		else if (GI->type==GIT_VSUNI)
			FCEUI_VSUniToggleDIPView();
		goto end;
	case _emulator_insert:
		if (!GI)
			goto end;
		if (GI->type==GIT_FDS)
			FCEUI_FDSInsert(0);
		else if (GI->type==GIT_VSUNI)
			FCEUI_VSUniCoin(2);
		goto end;
	case _emulator_coin1:
	case _emulator_coin2:
		if (GI && GI->type==GIT_VSUNI)
			FCEUI_VSUniCoin(itemcode-_emulator_coin1);
		goto end;
	case _emulator_debug:
		checkmenuitem(_emulator_debug,debuggerwindow_show(-1)>0);
		break;
	case _emulator_step:
		checkmenuitem(_emulator_debug,debuggerwindow_step(1)>0);
		break;

	/* model/region */
	case _submodel_auto:
	case _submodel_NTSC_A:
	case _submodel_NTSC_G:
	case _submodel_RGB:
	case _submodel_PAL:
	case _submodel_Dendy:
		itemcode-= _submodel_auto+1;
		if (emuset.model==itemcode)
			goto end;
		setcursor(4,false);
		stopsound();
		emuset.model = itemcode;
		update_emumodel();
		RefreshThrottleFPS(); //first to clear
		videorefresh(1);
		RefreshThrottleFPS(); //second to ensure timing update
		goto rsnd;

	/* config */
	case _config_Genie:
		FCEUI_SetGameGenie(emuset.usegenie^=1);
		break;
	case _config_autoips:
		FCEUI_SetAutoIPS(emuset.autoips^=1);
		break;
	case _config_snapname:
		FCEUI_SetSnapName((eoptions^=EO_SNAPNAME)&EO_SNAPNAME);
		break;
	case _config_palette:
		//stopsound();
		gamewindowtofront();
		PaletteDialog(NULL,0);
		break;
	case _config_netplay:
		//stopsound();
/*		ShowNetplayConsole();*/
		break;
	/* cheat item replaced with commandline commands (will be implemented eventually) */

	/* edit */
	case _edit_copy:
		if (logwindow_show(IMIN(int)))
			logwindow_scrap(0);
		break;
	case _edit_select_all:
		if (logwindow_scrap(IMAX(int)))
			enablemenuitem(_edit_copy,true);
		break;
	case _edit_log:
		tmp = logwindow_show(-1)>0;
		checkmenuitem(_edit_log,tmp);
		enablemenuitem(_edit_select_all,tmp);
		enablemenuitem(_edit_copy,tmp && logwindow_scrap(-1));
		break;
	case _edit_memview:
		memview_show(-1);
		break;

	case _view_cpu:
	case _view_ppu:
	case _view_oam:
	case _view_apu:
		memview_setview(itemcode-_view_cpu+_view_mode_cpu);
		break;

	case _nsf_autoadvance:
		sndset.nsfauto = !sndset.nsfauto;
		goto nsfa;
	case _nsf_loop1x:
		sndset.nsfauto = 3;
		goto nsfa;
	case _nsf_loop:
		sndset.nsfauto = 2;
	nsfa:	FCEUI_NSFSetAuto(sndset.nsfauto);
		break;
	case _nsf_silencedetect:
		if ((tmp=valuedialog("\pSilence Detection (in milliseconds; 0=off)",
				sndset.nsfasilence,0,30000,false))
				==INT_NO_VALUE)
			break;
		FCEUI_NSFSetAutoSilenceTime(sndset.nsfasilence=tmp);
		break;
	case _nsf_max_init_time:
		if (sndset.nsfmif<0)
			sndset.nsfmif =-1;
		if ((tmp=valuedialog("\pMaximum Initialization Frames (-1=off)",
				sndset.nsfmif,-1,127,false))
				==INT_NO_VALUE)
			break;
		FCEUI_NSFSetMaxInit(sndset.nsfmif=tmp);
		break;
	case _nsf_def_track_time:
		if (sndset.nsfdefttime<0)
			sndset.nsfdefttime = 0;
		if ((td=valuedialog("\pDefault Track Time (in minutes; 0=unlimited)",
				(uint32)(sndset.nsfdefttime/.6+.5)/100000.,0,35791,true))
				==INT_NO_VALUE)
			break;
		FCEUI_NSFSetDefTrackTime(sndset.nsfdefttime=td*60000+.5);
		break;

	case _subsound_latency:
		if ((td=valuedialog("\pSound Latency (in milliseconds)",
				sndset.latency,4,125,true))
				==INT_NO_VALUE)
			break;
		sndset.latency = td;
		RefreshThrottleFPS();
		break;
	/* sound quality */
//	case _subsound_telephone:
//	case _subsound_lower:
	case _subsound_low:
	case _subsound_default:
	case _subsound_moderate:
	case _subsound_low_cd:
	case _subsound_triple_22:
	case _subsound_double_cd:
	case _subsound_better:
	case _subsound_best:
	case _subsound_other:
		stopsound();

		if ((itemcode+=-_subsound_low)>=GET_ARRAY_COUNT(sndqualities))
		{
			if ((td=valuedialog("\pCustom Sample Rate",
					(int32)(sndset.rate*100)/100.,1000,(MAXIMUM_SRATE+8)/16.,true))
					==INT_NO_VALUE)
				goto end;
			/* rate warning moved to files.c; 2012.03.18 */
			tmp = (int32)(td*16+.5);
			if (tmp>MAXIMUM_SRATE)
				tmp = MAXIMUM_SRATE;
			sndset.rate = tmp/16.;
		} else
			sndset.rate = sndqualities[itemcode]/16.;
		if (GI)
			GI->soundrate = 0;
	#if HAVE_HQ
		FCEUI_SetSoundQuality(soundquality=0); /*(qualities[itemcode]>>24)&7);*/
	#endif
		goto rsnd;

	/* sound volume submenu (sorry, no slider) */
	case _subvolume_off:
	case _subvolume_10:
	case _subvolume_20:
	case _subvolume_30:
	case _subvolume_40:
	case _subvolume_50:
	case _subvolume_60:
	case _subvolume_70:
	case _subvolume_80:
	case _subvolume_90:
	case _subvolume_100:
	case _subvolume_110:
	case _subvolume_120:
	case _subvolume_130:
	case _subvolume_140:
	case _subvolume_150:
	case _subvolume_other:
		tmp = soundvolume;
		if ((itemcode-=_subvolume_off)==0)
			goto zerovolume;
		if (itemcode>=_subvolume_other-_subvolume_off)
		{
			if ((tmp=valuedialog("\pMaster Volume (percent)",
					soundvolume,-2000,2000,false))==INT_NO_VALUE)
				goto end;
			if ((soundvolume=tmp)==0)
			{
			zerovolume:
				soundvolume = 0; /* now must set ahead of InitSound */
				//stopsound();
				InitSound(NULL);
				RefreshThrottleFPS(); /* 2015.08.04 */
			}
		} else
			soundvolume = itemcode*10;
		FCEUI_SetSoundVolume(soundvolume);
		if (soundon && soundvolume && !tmp)
			goto rsnd;
		break;

	/* video */
	case _subvideo_fullscreen:
		if (!GI)
			vidset.fullscreen^= 1;
		else
			initializescreen(vidset.fullscreen^1);
		checkmenuitem(_subvideo_fullscreen,vidset.fullscreen);
		goto end;
	case _subvideo_2x_scale:
		checkmenuitem(_subvideo_2x_scale,(screenshift^=1)&1);
		if (GI)
			initializescreen(vidset.fullscreen);
		videorefresh(1);
		/* 'duplicate fields' set to scale factor by default; 2016.12.21 */
		screenshift&=~0x40;
		screenshift|= screenshift<<6&0x40;
		break;
	case _subvideo_dup_fields:
		screenshift^= 0x40;
		break;
	case _subvideo_stretch_filter:
		screenshift^= 0x80;
		screenshift|= screenshift>>7|(screenshift>>1&0x40);
		if (GI)
			initializescreen(vidset.fullscreen); /* may shut filter if not enough memory */
		screenshift|= screenshift>>7;
		videorefresh(1);
		break;
	case _subvideo_refresh_sync:
		screenrefreshsync^= 1;
		RefreshThrottleFPS(); /* 2015.07.31 */
		break;
	case _subvideo_config_filter:
		FilterDialog();
		break;
/*	case _subvideo_clip_sides:
		eoptions^= EO_CLIPSIDES;
		videorefresh(1);
		break;*/
	case _subvideo_overscan:
		if (overscan_dialog()==1)
			videorefresh(1);
		break;
	case _subvideo_no_sprite_limit:
		eoptions^= EO_NOSPRLIM;
		videorefresh(0);
		break;
	case _subvideo_auto_skip_frames:
		eoptions^= EO_AUTOFRAMESKIP;
		videorefresh(2);
		break;
	case _subvideo_hide_pf:
		tmp = PPUrendis, PPUrendis&=~6;
		if (!(tmp&6))
			PPUrendis|= 6;
		break;
	case _subvideo_hide_sf:
		FCEUI_SetRenderDisable(2,-1);
		break;
	case _subvideo_tile_view:
		ShowHideTileView(-1);
		break;

	/* sound channels */
	case _subchannels_rect_1:
	case _subchannels_rect_2:
	case _subchannels_triangle:
	case _subchannels_noise:
	case _subchannels_DMC:
		sndactivechannels^= 1<<itemcode-_subchannels_rect_1;
	sche:	FCEUI_SetSoundActiveChannels(sndactivechannels);
		break;
	case _subchannels_alt_rect_duty:
		sndset.rduty = sndset.rduty==RDUTYBMP_STD ? RDUTYBMP_SWP:RDUTYBMP_STD;
		FCEUI_SetDutyBitmap(sndset.rduty);
		break;
	case _subchannels_no_duty_reset:
		FCEUI_SetDutyReset(sndset.rdutyrst);
		sndset.rdutyrst^= 1;
		break;
	case _subchannels_disable_4011h:
		FCEUI_SetDirectDADisable(sndset.ddmadis^=1);
		break;
	case _subchannels_rect_amp:
		//stopsound();
		if ((tmp=valuedialog("\pRectangle Filter Amp (FCEU 1.0 = 256)",
				rectangle_amp,-2000,2000,false))!=INT_NO_VALUE)
			FCEUI_SetRectFilterAmp(rectangle_amp=tmp);
		goto end;
	case _subchannels_tnd_amp: /* teenage nusaince minja durdles */
		//stopsound();
		if ((tmp=valuedialog("\pTri/Noise/DMC Amp (FCEU 1.0 = 256)",
				tnd_amp,-2000,2000,false))!=INT_NO_VALUE)
			FCEUI_SetTNDFilterAmp(tnd_amp=tmp);
		goto end;
	case _subchannels_tri_scale:
		//stopsound();
		if ((tmp=valuedialog("\pTriangle Scale (3 = 1.0)",
				tri_scale,0,9,false))!=INT_NO_VALUE)
			FCEUI_SetTriangleScale(tri_scale=tmp);
		goto end;
	case _subchannels_noise_scale:
		//stopsound();
		if ((tmp=valuedialog("\pNoise Scale (2 = 1.0)",
				noise_scale,0,9,false))!=INT_NO_VALUE)
			FCEUI_SetNoiseScale(noise_scale=tmp);
		goto end;
	case _subchannels_dc_filter:
		//stopsound();
		if ((tmp=valuedialog("\pDC Filter Loss (over 94; default: 93)",
				sndset.dcfl,0,127,false))!=INT_NO_VALUE)
			FCEUI_SetDCFilterLoss(sndset.dcfl=tmp);
		goto end;
	case _subexpchannels_enable:
		sndactivechannels^= 0x20;
		goto sche;
	case _subexpchannels_VRC6:
	case _subexpchannels_VRC7:
	case _subexpchannels_FDS:
	case _subexpchannels_MM5:
	case _subexpchannels_N163:
	case _subexpchannels_SS5B:
		sndactivechannels^= 64u<<itemcode-_subexpchannels_VRC6;
		FCEUI_SetSoundActiveChannels(sndactivechannels);
		break;
	case _subexpchannels_vrc6_volume:
		setfvolume(&vrc6_vol,"\pVRCVI Volume (%)");
		goto end;
	case _subexpchannels_vrc7_volume:
		setfvolume(&vrc7_vol,"\pVRCVII Volume (%)");
		goto end;
	case _subexpchannels_fds_volume:
		setfvolume(&fds_vol,"\pFDS Volume (%)");
		goto end;
	case _subexpchannels_mmc5_volume:
		setfvolume(&mmc5_vol,"\pMMC5 Volume (%)");
		goto end;
	case _subexpchannels_n163_volume:
		setfvolume(&n163_vol,"\pNamco 163 Volume (%)");
		goto end;
	case _subexpchannels_ss5b_volume:
		setfvolume(&ss5b_vol,"\pSunsoft 5B Volume (%)");
		goto end;
	case _subexpchannels_n163_maxch:
		n163_max = n163_max==256 ? 32:256;
		update_soundsettings(1);
		break;

	/* speed */
	case _subspeed_adj_dmc:
		RateFixFlags^= 0x40;
	relut:	FCEUI_SetTriangleScale(tri_scale); /* to rebuild the tables */
		break;
	case _subspeed_rate_only:
		RateFixFlags^= 0x80;
		goto rsnd;
	case _subspeed_50:
	case _subspeed_63:
	case _subspeed_71:
	case _subspeed_79:
	case _subspeed_89:
	case _subspeed_100:
	case _subspeed_126:
	case _subspeed_159:
	case _subspeed_178:
	case _subspeed_200:
	case _subspeed_other:
		if ((itemcode+=-_subspeed_50)>=GET_ARRAY_COUNT(emulationspeeds))
		{
			//stopsound();
			/* sound buffer clipping occurs for values lower than 17.3% */
			if ((td=valuedialog("\pCustom Emulation Speed",
					emuset.speed/10000000.,
					#if HAVE_HQ
						soundquality ? 66.6:
					#endif
						17.3,429,true))==INT_NO_VALUE)
				goto end;
			FCEUI_SetSpeed(emuset.speed=td*10000000.0+.5);
		} else
			FCEUI_SetSpeed(emuset.speed=emulationspeeds[itemcode]);
	rsnd:	setcursor(4,false); /* set to watch cursor; 2010.04.29 */
		InitSound(GI);
		setcursor(-1,false);
		break;

	/* background */
	case _subbackground_off:
	case _subbackground_full:
	case _subbackground_muted:
	case _subbackground_bg_muted:
		eoptions&=~(EO_BGRUN|EO_MUTE);
		if (--itemcode&1)
			eoptions|= EO_BGRUN;
		if (itemcode&2)
			eoptions|= EO_MUTE;
		break;
	/* input */
	case _subinput_fscore:
		FCEUI_DisableFourScore(inputset.fscore);
		inputset.fscore = !inputset.fscore;
		break;
	case _subinput_mic:
		inputset.mic = !inputset.mic;
		FCEUI_EnableMic(inputset.mic);
		break;
	case _subinput_vsauto:
		eoptions^= EO_VSMANUALINPUT;
		break;
	case _subinput_gamepads:
		config_gamepads();
		goto end;
	case _subinputa_none:
	case _subinputa_none+SI_GAMEPAD:
	case _subinputa_none+SI_ZAPPER:
	case _subinputa_none+SI_POWERPADA:
	case _subinputa_none+SI_POWERPADB:
	case _subinputa_none+SI_ARKANOID:
		userinputtype[0] = itemcode-_subinputa_none;
		goto initin;
	case _subinputb_none:
	case _subinputb_none+SI_GAMEPAD:
	case _subinputb_none+SI_ZAPPER:
	case _subinputb_none+SI_POWERPADA:
	case _subinputb_none+SI_POWERPADB:
	case _subinputb_none+SI_ARKANOID:
		userinputtype[1] = itemcode-_subinputb_none;
		goto initin;
	case _subinputexp_none:
	case _subinputexp_none+SI_ARKANOID-4:
	case _subinputexp_none+SI_SHADOW-4:
	case _subinputexp_none+SI_4PLAYER-4:
	case _subinputexp_none+SI_FKB-4:
	case _subinputexp_none+SI_HYPERSHOT-4:
	case _subinputexp_none+SI_MAHJONG-4:
	case _subinputexp_none+SI_QUIZKING-4:
	case _subinputexp_none+SI_FTRAINERA-4:
	case _subinputexp_none+SI_FTRAINERB-4:
	case _subinputexp_none+SI_OEKAKIDS-4:
	case _subinputexp_none+SI_BWORLD-4:
	case _subinputexp_none+SI_TOPRIDER-4:
	case _subinputexp_none+SI_SUBORKB-4:
		if ((itemcode-=_subinputexp_none)!=0)
			itemcode+= 4;
		userinputtype[2] = itemcode;
	initin:	initializeinput();
		cidisabled = inputtype[2]==SI_FKB || inputtype[2]==SI_SUBORKB;
		break;

	default:
		if ((itemcode&0x7FC0)!=(submenuid_recent-menuid_apple<<6))
			break;
		itemcode&= 63;
		if (--itemcode==clearrecent(false))
			clearrecent(true);
		else
			takefilepath(getrecentpath(itemcode));
	}
	updatemenus();
end:	return true;
}
