/*
 macos/palettedialog.c

  Created: 2009.01.24

  MacFCEU 0.9a1

Tuesday, October 6, 2009 7:15 PM EDST
  now the palette dialog can be used for just about everything!
*/

#include <stdio.h>
#include <fp.h>

#include <Controls.h>
#include <ControlDefinitions.h>
#include <ColorPicker.h>

#include "macdialogs.h"

#include "../FCEU.h"
#include "../palette.h"
#include "../video.h"

#include "shellmenus.h"
#include "macvideo.h"

#include "../nes_ntsc/nes_ntsc.h"

#ifdef env68k
 #pragma segment shell
#endif

/* ---- globals */

extern struct FCEUGI	*GI;

const static char sliderLabels[] = "PRAFDHSB";

static boolean		updateinterface = false;

/* ---- constants */

/* sliders don't have incremental left/right buttons! */
/* Appearance would also have to be detected (Gestalt) for compatibility */

#define SLIDER_WIDTH			218
#define SLIDER_HEIGHT			16

#if 0
 #define SLIDER_PROC			kControlSliderProc
#else
 #define SLIDER_PROC			scrollBarProc
#endif

enum
{ /* palette dialog items */
	_palettesetup_reset		= 3,
	_palettesetup_open,
	_palettesetup_save,
	_palettesetup_stats,
	_palettesetup_table, /* 8x8 pixels 32x8 full */
/*	_palettesetup_hue,
	_palettesetup_tint,*/
	_palettesetup_col,
	_palettesetup_custom,
	_palettesetup_huesat_caption
};

/* frame sizes for colors counts */
#define PEL_BSIZE_64			16
#define PEL_BSIZE_256			8

/* ---- code */

static void updatepalstrings(DialogRef dialog,
	const uint8 *pel)
{ /* update whatever else that needs to be updated */
	char		buf[64];

	buf[0] = sprintf(1+buf,
		!pel ? "%d\n%g\n%d":"%u",
		!pel ? ntschue:*pel,
		(double_t)ntscsat,
		ntscbrt);
	setdialogitem_pstr(dialog,_palettesetup_stats,(uint8*)buf);
}

void (*updatestrs_proc)(DialogRef dialog,
	const uint8 *pel) = updatepalstrings;

/* ---- slider code */

static void updatestats(ControlRef control,
	SInt16 old)
{
	char		buf[256];
	SInt16		value = GetControlValue(control);

	if (old==value)
		return;
	GetControlTitle(control,(uint8*)buf);
	updateinterface = 1;
	switch(buf[1])
	{
	case 'H': ntschue = value; goto col;
	case 'S': ntscsat = value/2.; goto col;
	case 'B': ntscbrt = value;
	col:	updateinterface = ntsccol!=0;
		break;

	case 'P': vidset.shp = value; break;
	case 'R': vidset.res = value; break;
	case 'A': vidset.art = value; break;
	case 'F': vidset.frn = value; break;
	case 'D': vidset.bld = value;
	}
	updatestrs_proc(GetDialogFromWindow(GetControlOwner(control)),NULL);
}

static pascal void setting_actionproc(ControlRef control,
	ControlPartCode part)
{
	unsigned long	ticks = TickCount();
	unsigned long	reference = GetControlReference(control);
	SInt16		old;

	if (ticks<reference)
		return;
	old = GetControlValue(control);
	switch(part)
	{
	case kControlUpButtonPart:
		if (old>GetControlMinimum(control))
			SetControlValue(control,old-1);
		break;
	case kControlDownButtonPart:
		if (old<GetControlMaximum(control))
			SetControlValue(control,old+1);
		break;
	case kControlPageUpPart:
		SetControlValue(control,old-8);
		break;
	case kControlPageDownPart:
		SetControlValue(control,old+8);
		break;
	case kControlIndicatorPart:
		break;
	}

	updatestats(control,old);

	if (reference==0)
		SetControlReference(control,ticks+30); /* 30/60 sec. delay */
}

static ControlRef makeslider(WindowRef window,
	int v,
	int top,
	const UInt8 *title)
{
	Rect		bounds;
	int		min = memchr(sliderLabels,title[1],5) ? -100:0;

	SetRect(&bounds,108,top,108+SLIDER_WIDTH,top+SLIDER_HEIGHT);
	return NewControl(window,&bounds,title,true,v,min,min ? 100:title[1]=='S' ? 256:128,
		SLIDER_PROC,0);
}

static int anycontrolclick(ControlRef control,
	Point where,
	short part,
	ControlActionUPP actionupp)
{
	char		buf[256];

	GetControlTitle(control,(uint8*)buf);
	if (buf[0]!=1 || !memchr(sliderLabels,buf[1],8))
		return 0;

	SetControlReference(control,0); /* initialize timing for control */

	switch(part)
	{
	case kControlUpButtonPart:
	case kControlDownButtonPart:
	case kControlPageUpPart:
	case kControlPageDownPart:
		TrackControl(control,where,actionupp);
		return 1;
	case kControlIndicatorPart:
		TrackControl(control,where,NULL);
		updatestats(control,-1);
		return 1;
	}

	return -1;
}

/* ---- palette code */

static void copy4to3(uint8 *d,
	const pal *s,
	int c)
{
	if (!d || c<=0)
		return;
	do
		*d++ = s->r,
		*d++ = s->g,
		*d++ = s->b;
	while(++s,--c);
}

/* ---- dialogs */

#include "../utility/interpreter.h"

static char	*pal_gfsdir; /* current directory storage; 2010.04.08 */

static int load_palette(void)
{
	FILE		*fp;
	char		*path;
	uint16		nameoffs;
	int		result;

	if ((path=getfilepath(gfp_getfile,
			getpstr(temporary,strlistid_common,commonstr_loadpalette),
			NULL,pal_gfsdir,NULL,
			"All usable files(*.pal)\0*.pal\0All files (*.*)\0*.*\0",
			&nameoffs,
			NULL,
			gfpf_hidereadonly|gfpf_filemustexist))==NULL)
		goto end;

	tracknewpath(&pal_gfsdir,path,NULL,nameoffs,2);

	if ((fp=fopen(path,"rb"))==NULL)
		FCEUD_PrintError("Error %d opening palette file.");
	else
	{
		if ((result=fread(cpalette,1,sizeof(cpalette),fp))>0)
		{
			if (!memcmp(cpalette,"JASC-PAL\15\0120100\15\012256\15\012",21))
			{ /* Jasc */
				char		buf[256];
				const char	*s;
				long int	r, g, b;
				int		o;

				fseek(fp,21,SEEK_SET);
				o = 0;
				do
				{
					r = g = b = 0;
					fgets(buf,255,fp);
					if ((s=readanyinteger(buf,strlen(buf),&r))!=NULL)
						s = parse_skipwsp(s,s+strlen(s));
					if ((s=readanyinteger(s,strlen(s),&g))!=NULL)
						s = parse_skipwsp(s,s+strlen(s));
					s = readanyinteger(s,strlen(s),&b);

				//	FCEU_printf("%u %u %u\n",
				//		r,g,b);

					cpalette[o] = r>255 ? ~r>>31:r;
					cpalette[o+1] = g>255 ? ~g>>31:g;
					cpalette[o+2] = b>255 ? ~b>>31:b;
				} while((o+=3)<64*3);
			} else if (result<sizeof(cpalette)) //clear end if short
				memset(cpalette+result,0,sizeof(cpalette)-result);
			FCEUI_SetPaletteArray(cpalette);
		} else
			FCEUD_PrintError(result ?
				"Error reading the palette file.":"Palette file is empty.");
		fclose(fp);
		eoptions|= EO_CPALETTE;

		return result;
	}
end:	return 0;
}

static boolean save_palette(const uint8 *palette)
{
	static char	currentname[256];

	FILE		*fp;
	char		*path;
	unsigned	length = 0;
	uint16		nameoffs = 0;

	if (!currentname[0])
		getcstr(currentname,strlistid_common,commonstr_defaultpalettename);

	if ((path=getfilepath(gfp_putfile,
			getpstr(temporary,strlistid_common,commonstr_savepalette),
			NULL,pal_gfsdir,currentname,
			"PAL file(*.pal)\0*.pal\0",
			&nameoffs,
			NULL,
			gfpf_hidereadonly|gfpf_filemustexist))==NULL)
		goto end;

	tracknewpath(&pal_gfsdir,path,NULL,nameoffs,8);

	if (nameoffs>0 && (unsigned)(length=1+strlen(path)-nameoffs)<=256)
		memcpy(currentname,nameoffs+path,length);
	/* dual try - creation, workaround for B2 ExtFS creation error */
	if ((fp=fopen(path,"wb"))==NULL && (fp=fopen(path,"wb"))==NULL)
		FCEUD_PrintError("Error opening file to write palette.");
	else
	{
		fwrite(palette,1,sizeof(cpalette),fp);
		fclose(fp);
		return true;
	}
end:	return false;
}

static void hilitepel(const Rect *rect)
{
	Rect		outer = *rect;

	PenSize(2,2); /* thicker outward to make it more visible; 2009.12.29 */
	PenMode(srcXor);
	++outer.bottom, ++outer.right;
	FrameRect(&outer);
	PenNormal();
}

static void paintrectgray(const Rect *rect)
{
	Pattern		graypattern;

	PenPat(GetQDGlobalsGray(&graypattern));
	PaintRect(rect);
	PenNormal();
}

static uint8 colortable_msg(DialogRef dialog,
	boolean custom,
	int first,
	Point hitpoint,
	Rect *selrect,
	uint8 *pickpel,
	int c)
{
extern ColorTable myctable;
	Rect		bounds, pelrect;
	RGBColor	color;

	Handle		handle;
	GrafPtr		oldport;
	uint8		*p;
	int		i, z, iz = 1;
	int		rectsize = PEL_BSIZE_64;
	short		type;
	uint8		sel = 0;

	GetPort(&oldport);
	SetPortDialogPort(dialog);

	GetDialogItem(dialog,_palettesetup_table,&type,&handle,&bounds);

	if (!selrect)
	{
		InvalWindowRect(GetDialogWindow(dialog),&bounds);
		return 0;
	}

	pelrect = bounds;
	/* draw bevel */
	RGBForeColor(&myctable.ctTable[PAL_BGRAY].rgb);
	PenNormal();
	MoveTo(pelrect.left-1,pelrect.bottom);
	LineTo(pelrect.left-1,pelrect.top-1);
	LineTo(pelrect.right,pelrect.top-1);
	RGBForeColor(&myctable.ctTable[0].rgb);
	MoveTo(pelrect.right+1,pelrect.top);
	LineTo(pelrect.right+1,pelrect.bottom+1);
	LineTo(pelrect.left,pelrect.bottom+1);
	PenMode(srcCopy);

	if (c>0)
	{
		custom = 1; //`custom` has to be non-zero for the picker
		p = (uint8*)&myctable.ctTable->rgb;
		z =-1;
		if (c<=64)
			goto gamepal;
	} else
		c = 64;
	if (c>64)
	{
		rectsize = PEL_BSIZE_256;
		p = (uint8*)&myctable.ctTable->rgb;
		c = 256;
		z = 8;
	gamepal:
		iz = 2;
	} else if (custom)
		p = cpalette, z = 3;
	else
	{
		if (GI && GI->type==GIT_NSF) /* handle this case; 2009.12.29 10:55 EST */
		{
			++bounds.top, ++bounds.left, --bounds.bottom, --bounds.right;
			paintrectgray(&bounds);
			goto end;
		}
		p = (uint8*)palo, z = sizeof(pal);
	}

	pelrect.bottom = rectsize+pelrect.top++;
	pelrect.right = (pelrect.left=bounds.left+1)+rectsize-1;
	for(i=0; i<c && pelrect.bottom<bounds.bottom; p+=z>0 ? z:0,++i)
	{
		if (i>=first)
		{
			if (z<0)
				p = (uint8*)&myctable.ctTable[epal_inv[i]].rgb;
			color.red = p[0]*257;
			color.green = p[iz]*257;
			color.blue = p[iz*2]*257;
			if (first>=0)
			{
				if (pickpel)
					++pelrect.bottom, ++pelrect.right;
				RGBForeColor(&myctable.ctTable[0].rgb);
				FrameRect(&pelrect);

				if (pickpel && *pickpel==i)
				{ /* hilite, take rect */
					*selrect = pelrect;
					hilitepel(&pelrect);
				}
				RGBForeColor(&color);
				/* inset for inset color */
				++pelrect.top, ++pelrect.left;
				--pelrect.bottom, --pelrect.right;

				/* gray pattern for last (black is #7) */
				if (i==255)
					paintrectgray(&pelrect);
				else
					PaintRect(&pelrect);

				--pelrect.top, --pelrect.left;
				if (!pickpel)
					++pelrect.bottom, ++pelrect.right;
			} else if (custom && PtInRect(hitpoint,&pelrect)) /* click */
			{
				*(int32*)&hitpoint =-1;

				/* hilite; 2009.10.06 18:37 EDT */
				hilitepel(selrect);
				hilitepel(&pelrect);
				*selrect = pelrect;

				if (pickpel)
				{
					*pickpel = i;
					break;
				}
				getpstr(temporary,strlistid_common,commonstr_pickcolor);
				if (GetColor(hitpoint,(uint8*)temporary,&color,&color))
				{
					p[0] = color.red>>8;
					p[1] = color.green>>8;
					p[2] = color.blue>>8;
					sel = 1;
				} else
					sel =-1; /* hit but no change; 2010.01.16 */
				break;
			}
		}

		pelrect.left+= rectsize;
		if ((pelrect.right+=rectsize)>=bounds.right)
		{
			pelrect.bottom = (pelrect.top+=rectsize)+rectsize-1;
			pelrect.right = (pelrect.left=bounds.left+1)+rectsize-1;
		}
	}
end: /* reset black */
	RGBForeColor(&myctable.ctTable[255].rgb);

	SetPort(oldport);

	return sel;
}

#define CUSTOM_USED			(!ntsccol && eoptions&EO_CPALETTE)

int PaletteDialog(uint8 *pickpel,
	int colc)
{
	Rect		selrect;

	ControlActionUPP	actionupp;

	DialogRef	dialog = NULL;
	WindowRef	dialogwindow;
	ControlRef	huectl = NULL, satctl = NULL, brtctl = NULL;
	GrafPtr		oldport;
	float		save_sat;
	int		save_col, save_hue, save_brt, save_eoptions;
	int		hit =-1;

	GetPort(&oldport);

	ntscsat = FCEUI_GetNTSC_HSB(&ntschue,&ntscbrt)/2.; /* update from internal values */

	save_col = ntsccol;
	save_sat = ntscsat;
	save_hue = ntschue;
	save_brt = ntscbrt;
	save_eoptions = eoptions;

	((int32*)&selrect)[0] = ((int32*)&selrect)[1] =-1; /* no hilite */

	/* set up */
	if ((actionupp=NewControlActionUPP(setting_actionproc))==NULL)
		return -1;
	if ((dialog=getnewdialog(dialogid_palettesetup,std_ok,(WindowRef)-1))==NULL)
		goto end;
	dialogwindow = GetDialogWindow(dialog);
	SetPortDialogPort(dialog);

	updatestrs_proc = updatepalstrings;
	updatepalstrings(dialog,pickpel);

	/* buttons */
	if (pickpel)
	{
		SetWTitle(dialogwindow,"\pPEL Picker");

		HideDialogItem(dialog,_palettesetup_reset);
		HideDialogItem(dialog,_palettesetup_open);
		HideDialogItem(dialog,_palettesetup_save);

		/* show, but disable */
		setcontrolitem(dialog,_palettesetup_col,ntsccol,0,1,-1,NULL);
		setcontrolitem(dialog,_palettesetup_custom,eoptions&EO_CPALETTE ? 1:0,
			0,1,-1,NULL);
		setdialogitem_pstr(dialog,_palettesetup_huesat_caption,"\pIndex");
	} else
	{
		/* sliders */
		if ((huectl=makeslider(dialogwindow,ntschue,78,"\pH"))==NULL
				|| (satctl=makeslider(dialogwindow,ntscsat*2,98,"\pS"))==NULL
				|| (brtctl=makeslider(dialogwindow,ntscbrt,118,"\pB"))==NULL)
			goto end;

		setbinaryitem(dialog,_palettesetup_col,ntsccol);
		setbinaryitem(dialog,_palettesetup_custom,eoptions&EO_CPALETTE ? 1:0);

		/* force color update */
		mycolorsinvalidate|= 2;
		videocolorsinit(); /* just in case this is the first time */
		FCEUD_BlitScreen(NULL);
	}

	SetDialogCancelItem(dialog,std_cancel);
	SetDialogDefaultItem(dialog,std_ok);

	ShowWindow(dialogwindow);

	/* event loop */
	reenablemenus(false);

	do
	{
		Point		hitpoint;
		EventRecord	event;

		hit =-1;
		while(WaitNextEvent(everyEvent,&event,1,NULL))
		{
			hitpoint = event.where;
			localwindowpoint(dialogwindow,&hitpoint);

			if (event.what==mouseDown)
			{
				ControlRef	control;
				short		part;


				if ((part=FindControl(hitpoint,dialogwindow,&control))
							!=kControlNoPart && control)
					anycontrolclick(control,hitpoint,part,actionupp);
			}

			hit =-1;
			if (!pickpel && event.what==keyDown && event.modifiers&cmdKey)
			{ /* cmd-O for Open, cmd-S for Save */
				char		charcode = event.message&charCodeMask;

				if (charcode=='o')
					hit = _palettesetup_open;
				else if (charcode=='s')
					hit = _palettesetup_save;
				else if (charcode=='r')
					hit = _palettesetup_reset;
				else if (charcode=='c')
					hit = _palettesetup_custom;
				else if (charcode=='n')
					hit = _palettesetup_col;
			}

			if (hit>0)
				animatebuttonitem(dialog,hit);
			else if ((hit=handleevent(&event,dialog))==0)
			{
				InitCursor();
				hit = standardmodal(dialog,&event);
			} else if (hit>=0 || hit==-2)
				break;
			if (event.what==updateEvt && (WindowRef)event.message==dialogwindow)
				if (colortable_msg(dialog,CUSTOM_USED,0,
						hitpoint,&selrect,pickpel,colc) && !pickpel)
					updateinterface = true;
		}

		if ((unsigned)(hit-2)<=_palettesetup_reset-2) /* -2 to include cancel */
		{
			if (pickpel)
				break;
			/* copy internal calculated palette if holding down shift key */
			if (event.modifiers&shiftKey)
				copy4to3(cpalette,palo,64);
		reset:	setbinaryitem(dialog,_palettesetup_col,ntsccol=save_col);
			ntscsat = save_sat, ntschue = save_hue, ntscbrt = save_brt;
			if (((eoptions=save_eoptions)&EO_CPALETTE)==0)
				FCEUI_SetPaletteArray(NULL);
			setbinaryitem(dialog,_palettesetup_custom,
				(eoptions&EO_CPALETTE)!=0);
			if (huectl)
				SetControlValue(huectl,ntschue);
			if (satctl)
				SetControlValue(satctl,ntscsat*2);
			if (brtctl)
				SetControlValue(brtctl,ntscbrt);
			updatepalstrings(dialog,pickpel);
			goto update;
		} else if (hit==_palettesetup_custom)
		{
			if (setbinaryitem(dialog,_palettesetup_custom,INT_NO_VALUE)>0)
			{
				if (getbinaryitem(dialog,_palettesetup_custom))
				{
					eoptions|= EO_CPALETTE;
					FCEUI_SetPaletteArray(cpalette);
				} else
				{
					eoptions&=~EO_CPALETTE;
					FCEUI_SetPaletteArray(NULL);
				}
				if (!ntsccol) /* unnecessary (overridden) */
					goto inval;
			}
		} else if (hit==_palettesetup_col)
		{
			if (setbinaryitem(dialog,_palettesetup_col,INT_NO_VALUE)>0)
			{
				ntsccol = getbinaryitem(dialog,_palettesetup_col);
				goto update;
			}
		} else if (hit==_palettesetup_open)
		{
			if (load_palette()>0)
			{
				setbinaryitem(dialog,_palettesetup_custom,1);
				goto update;
			}
		} else if (hit==_palettesetup_save)
			save_palette(event.modifiers&shiftKey ? (const uint8*)palo:cpalette);
		else if (hit==_palettesetup_table)
		{
			boolean		hit = colortable_msg(dialog,CUSTOM_USED,-1,
						hitpoint,&selrect,pickpel,colc);

			updatepalstrings(dialog,pickpel);
			if (hit)
				goto inval;
		}

		if (updateinterface)
		{
		update:	FCEUI_SetNTSC_HSB(/*!ntsccol ? 0:hit!=-2 && (unsigned)(hit-1)>std_cancel-1 ? 1:4*/
				 ntsccol,ntschue,ntscsat*2,ntscbrt);
			if (event.what!=updateEvt)
			{
			inval:	colortable_msg(dialog,CUSTOM_USED,-1,
					hitpoint,NULL,NULL,colc);
			}
			if (GI && GI->vidsys==GIV_PAL)
				continue;
			mycolorsinvalidate|= 2;
			if (GI)
				videorefresh(1);
			else
				FCEUD_BlitScreen(NULL);
			updateinterface = false;
		}
	} while(hit!=-2 && (unsigned)(hit-1)>std_cancel-1);

end:	SetPort(oldport);

	reenablemenus(true);

	if (dialog)
	{
		/* DisposeDialog expects its controls for deletion (don't KillControls) */
		if (huectl)
			DisposeControl(huectl);
		if (satctl)
			DisposeControl(satctl);
		if (brtctl)
			DisposeControl(brtctl);
		DisposeDialog(dialog);
		InitCursor();
	}

	DisposeControlActionUPP(actionupp);

	return hit;
}

/* filter dialog */

static void updatefilterstrings(DialogRef dialog,
	const uint8 *pel)
{
	char		buf[256];
(void)pel;
	buf[0] = sprintf(1+buf,
		"%g\n%g\n%g\n%g\n%g",
		(double_t)vidset.shp/100.,
		(double_t)vidset.res/100.,
		(double_t)vidset.art/100.,
		(double_t)vidset.frn/100.,
		(double_t)vidset.bld/100.);
	setdialogitem_pstr(dialog,_palettesetup_stats,(uint8*)buf);
}

void FilterDialog(void)
{
	video_settings	savevid;
	Rect		bounds;

	ControlActionUPP	actionupp;

	Handle		handle;
	DialogRef	dialog = NULL;
	WindowRef	dialogwindow;
	ControlRef	shpctl = NULL, resctl = NULL, artctl = NULL, frnctl = NULL, bldctl = NULL;
	GrafPtr		oldport;
	int		hit =-1;
	short		type;

	GetPort(&oldport);

	savevid = vidset;

	/* set up */
	if ((actionupp=NewControlActionUPP(setting_actionproc))==NULL)
		return;
	if ((dialog=getnewdialog(dialogid_palettesetup,std_ok,(WindowRef)-1))==NULL)
		goto end;
	dialogwindow = GetDialogWindow(dialog);

	SetPortDialogPort(dialog);
	SetWTitle(dialogwindow,"\pFilter Parameters");

	updatestrs_proc = updatefilterstrings;
	updatefilterstrings(dialog,NULL);

	HideDialogItem(dialog,_palettesetup_open);
	HideDialogItem(dialog,_palettesetup_save);
	HideDialogItem(dialog,_palettesetup_custom);
	HideDialogItem(dialog,_palettesetup_table);

	/* turn HSB into PRAFD */
	GetDialogItem(dialog,_palettesetup_huesat_caption,&type,&handle,&bounds);
	bounds.top-= 40;
	SetDialogItem(dialog,_palettesetup_huesat_caption,type,handle,&bounds);
	SetDialogItemText(handle,"\pSharp\nRes.\nArt.\nFring.\nBleed");

	GetDialogItem(dialog,_palettesetup_stats,&type,&handle,&bounds);
	bounds.top-= 40;
	SetDialogItem(dialog,_palettesetup_stats,type,handle,&bounds);

	GetDialogItem(dialog,_palettesetup_col,&type,&handle,&bounds);
	SetControlTitle((ControlRef)handle,"\pMerge Fields");
	bounds.right+= 35;
	bounds.bottom = bounds.bottom-bounds.top+10;
	bounds.top = 10;
	SetControlBounds((ControlRef)handle,&bounds);
	SetControlValue((ControlRef)handle,vidset.mrg);
	SetDialogItem(dialog,_palettesetup_col,type,handle,&bounds);

	/* sliders */
	if ((shpctl=makeslider(dialogwindow,vidset.shp,33,"\pP"))==NULL
			|| (resctl=makeslider(dialogwindow,vidset.res,53,"\pR"))==NULL
			|| (artctl=makeslider(dialogwindow,vidset.art,73,"\pA"))==NULL
			|| (frnctl=makeslider(dialogwindow,vidset.frn,93,"\pF"))==NULL
			|| (bldctl=makeslider(dialogwindow,vidset.bld,113,"\pD"))==NULL)
		goto end;

	/* force color update */
	if (GI)
	{
		mycolorsinvalidate|= 2;
		videocolorsinit(); /* just in case this is the first time */
		FCEUD_BlitScreen(XBuf);
	}

	SetDialogCancelItem(dialog,std_cancel);
	SetDialogDefaultItem(dialog,std_ok);

	ShowWindow(dialogwindow);

	/* event loop */
	reenablemenus(false);

	do
	{
		Point		hitpoint;
		EventRecord	event;

		hit =-1;
		while(WaitNextEvent(everyEvent,&event,1,NULL))
		{
			hitpoint = event.where;
			localwindowpoint(GetDialogWindow(dialog),&hitpoint);

			if (event.what==mouseDown)
			{
				ControlRef	control;
				short		part;


				if ((part=FindControl(hitpoint,dialogwindow,&control))
							!=kControlNoPart && control)
					anycontrolclick(control,hitpoint,part,actionupp);
			}

			hit =-1;
			if (event.what==keyDown && event.modifiers&cmdKey)
			{ /* cmd-O for Open, cmd-S for Save */
				char		charcode = event.message&charCodeMask;

				if (charcode=='r')
					hit = _palettesetup_reset;
				else if (charcode=='m')
					hit = _palettesetup_col;
			}

			if (hit>0)
				animatebuttonitem(dialog,hit);
			else if ((hit=handleevent(&event,dialog))==0)
			{
				InitCursor();
				hit = standardmodal(dialog,&event);
			} else if (hit>=0 || hit==-2)
				break;
		}

		if ((unsigned)(hit-2)<=_palettesetup_reset-2) /* -2 to include cancel */
		{
		reset:	vidset = savevid;
			if (shpctl)
				SetControlValue(shpctl,vidset.shp);
			if (resctl)
				SetControlValue(resctl,vidset.res);
			if (artctl)
				SetControlValue(artctl,vidset.art);
			if (frnctl)
				SetControlValue(frnctl,vidset.frn);
			if (bldctl)
				SetControlValue(bldctl,vidset.bld);
			setbinaryitem(dialog,_palettesetup_col,vidset.mrg);
			updatefilterstrings(dialog,NULL);
			goto update;
		} else if (hit==_palettesetup_col)
			if (setbinaryitem(dialog,_palettesetup_col,INT_NO_VALUE)>0)
			{
				vidset.mrg = getbinaryitem(dialog,_palettesetup_col);
				goto update;
			}

		if (updateinterface)
		{
		update:	if (GI && GI->vidsys==GIV_PAL)
				continue;
			if (GI)
				videorefresh(2);
			updateinterface = false;
		}
	} while(hit!=-2 && (unsigned)(hit-1)>std_cancel-1);

end:	SetPort(oldport);

	reenablemenus(true);

	if (dialog)
	{
		/* DisposeDialog expects its controls for deletion (don't KillControls) */
		if (shpctl)
			DisposeControl(shpctl);
		if (resctl)
			DisposeControl(resctl);
		if (artctl)
			DisposeControl(artctl);
		if (frnctl)
			DisposeControl(frnctl);
		if (bldctl)
			DisposeControl(bldctl);
		DisposeDialog(dialog);
		InitCursor();
	}
	DisposeControlActionUPP(actionupp);
}
