/*
 mappers/Sunsoft5B.c
  Sunsoft FME-7 and 5A/5B layer

 * MacFCEU 0.9a2
 *
 * Copyright notice for this file:
 *  69.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mapinc.h"

#define sunselect	mapbyte1[0]
#define sungah		mapbyte1[1]
#define sunindex	mapbyte1[2]
#define regs		MapperExRAM

/* ---- code */

/* bit 7: WRAM read active */
/* bit 6: WRAM write active */

static void SunFixPRGR(int v)
{
(void)v;
	if (sungah>=0xC0)
	{
		//if (sungah&0x80) // Select WRAM
		setprg8r(1,0x6000,sungah&3); //63);
	} else
		setprg8(0x6000,sungah&63);
}

static DECLFW(SUN5BWRAM)
{
	if (sungah>=0xC0)
		CartBW(A,V);
}

static DECLFR(SUN5AWRAM)
{
	if ((sungah&0xC0)==0x40)
		return A>>8; //XFrame.DB
	return CartBR(A);
}

static DECLFW(M69_W8)
{
#pragma unused(A)
	sunselect = V;
}

static DECLFW(M69_WA)
{
#pragma unused(A)
	switch(sunselect&15)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:	VROM_BANK1(sunselect<<10,V);
		break;
	case 8:	sungah = V;
		SunFixPRGR(0);
		break;
	case 9: ROM_BANK8(0x8000,V); break;
	case 10: ROM_BANK8(0xA000,V); break;
	case 11: ROM_BANK8(0xC000,V); break;
	case 12: if (V&2)
			onemir(V&1);
		else
			MIRROR_SET(V&1);
		break;
		/* acknowledge fix - clear bit 0 of 13 only; 2015.09.15 */
	case 13: IRQa = V;
		if (!(V&1))
			X65xx_IRQEnd(FCEU_IQEXT);
		break;
	case 14: IRQCount&= 0xFF00, IRQCount|= V; /*X6502_IRQEnd(FCEU_IQEXT);*/ break;
	case 15: IRQCount&= 0x00FF, IRQCount|= V<<8; /*X6502_IRQEnd(FCEU_IQEXT);*/ break;
	}
}

static void FP_FASTAPASS(1) SunIRQHook(int a)
{
	if (IRQa)
		if ((IRQCount-=a)<=0)
		{
			X65xx_IRQBegin(FCEU_IQEXT);
			IRQa = 0, IRQCount = 0xFFFF;
		}
}

/* ---- Yamaha YM2149F sound */

/* new implementation for more complete non-linear sound
 partly based on code in mame.zip/src/emu/sound/ay8910.c

three things set the YM2149 apart from an AY-3-8910:
 pin 26, which if low halves the clock rate (halved in Sun5B)
 2VDC offset, regardless of volume level 0
  (neutralized here for sake of DC filtering/overflow)
 internal resistance of 353 instead of 356

 both are affected by the output load resistance
 the 2149 and 3439 chips produce the same results
*/

#define AY_AFINE	0
#define AY_ACOARSE	1
#define AY_BFINE	2
#define AY_BCOARSE	3
#define AY_CFINE	4
#define AY_CCOARSE	5
#define AY_NOISEPER	6
#define AY_ENABLE	7
#define AY_AVOL		8
#define AY_BVOL		9
#define AY_CVOL		10
#define AY_EFINE	11
#define AY_ECOARSE	12
#define AY_ESHAPE	13
/* port writes (14/15) not handled */

#define WSCALE		4
#define ESCALE		4
#define ERESOLUTION	3

#define NOISEON(_chan)	(regs[AY_ENABLE]>>(3+_chan)&1)
#define TONEON(_chan)	(regs[AY_ENABLE]>>(_chan)&1)

static int	CAYBC[3];
static int32	wlen[3];
static int32	vcount[3], vnoise, dcount[3], dnoise;
static int32	rng = 1, env_period, noise_period, count_env;
static uint8	env_step, attack, hold, holding, alternate;

static int32	voltab[32];
static int16	svolume = 100;

static void BuildVolTable(void)
{
	static const uint16 lintab[] =
	{ /* curve based on resistor table, resistance load of 390 ohms */
	  /* as the channels are mixed electronically with parallel resistance,
	    the 3-ix table (32768 entries) would produce more genuine sound
	    but this newer implementation is already slow enough...
	  */
		0,60,144,261,330,412,485,568,
		684,819,946,1090,1308,1564,1811,2093,
		2523,3033,3541,4127,5000,6044,7090,8302,
		10234,12578,15060,17969,22957,29080,36416,45093,
	};

	int		x;

	x = 0;
	do
		voltab[x] = lintab[x]*svolume>>9;
	while(++x<32);
}

void FCEUI_SetS5BVolume(int16 v)
{
#if HAVE_HQ
	if (v<0 && FSettings.soundq)
		v =-v;
#endif
	svolume = v;
	BuildVolTable();
}

static void DoYM2149F(void)
{
	int32		amp[3], v;
	int32		tk[3], nk[3];
	int32		x, start = CAYBC[0], end = (SOUNDTS<<16)/soundtsinc;

	if (end<=(x=start) /*|| regs[7]&1u<<x*/)
		return;
	CAYBC[0] = end;
	v = env_step^attack;
	amp[0] = voltab[regs[AY_AVOL]&16 ? v:(regs[AY_AVOL]&15)*2u+1];
	amp[1] = voltab[regs[AY_BVOL]&16 ? v:(regs[AY_BVOL]&15)*2u+1];
	amp[2] = voltab[regs[AY_CVOL]&16 ? v:(regs[AY_CVOL]&15)*2u+1];
	tk[0] =-TONEON(0), nk[0] =-NOISEON(0);
	tk[1] =-TONEON(1), nk[1] =-NOISEON(1);
	tk[2] =-TONEON(2), nk[2] =-NOISEON(2);
	do
	{
		v = amp[0]&(dcount[0]|tk[0])&(dnoise|nk[0]);
		v+= amp[1]&(dcount[1]|tk[1])&(dnoise|nk[1]);
		v+= amp[2]&(dcount[2]|tk[2])&(dnoise|nk[2]);
		FCEUWave[x>>4]+= v;
		/* increment */
		vcount[0]-= nesincsize;
		while(vcount[0]<=0)
			dcount[0]^=~0, vcount[0]+= wlen[0];
		vcount[1]-= nesincsize;
		while(vcount[1]<=0)
			dcount[1]^=~0, vcount[1]+= wlen[1];
		vcount[2]-= nesincsize;
		while(vcount[2]<=0)
			dcount[2]^=~0, vcount[2]+= wlen[2];
		/* noise */
		vnoise-= nesincsize;
		while(vnoise<=0)
		{
			vnoise+= noise_period;
			if ((rng+1)&2) /* bit0^bit1 (pre) */
				dnoise^=~0;
			if (rng&1)
				rng^= 0x24000; /* Galois configuration */
			rng>>= 1;
		}
		/* update envelope */
		if (!holding && (count_env-=nesincsize)<0)
		{ /* check current envelope position */
			/* adjusted to act more like Nestopia and NSFPlay */
			/* "drum machine" still doesn't work as expected */
			do
				--env_step;
			while((count_env+=env_period)<0);
			if (env_step>31)
				if (hold)
				{
					attack^= alternate;
					holding = 1, env_step = 0;
				} else
				{
					if (alternate && env_step&32)
						attack^= 31;
					env_step&= 31;
				}
			v = voltab[env_step^attack];
			if (regs[AY_AVOL]&16)
				amp[0] = v;
			if (regs[AY_BVOL]&16)
				amp[1] = v;
			if (regs[AY_CVOL]&16)
				amp[2] = v;
		}
	} while(++x<end);
}

#if HAVE_HQ
static void DoAYSQHQ(int x) 
{ /* needs update (if ever going to be used) */
	int32	freq, amp, V;

	if ((amp=regs[8+x]&15)!=0)
	{
		amp = amp*volume>>8;
		freq = ((regs[x<<1]|(regs[x+x+1]&15)<<8)+1)<<4;
		for(V=CAYBC[x]; V<SOUNDTS; ++V)
		{
			if (dcount[x])
				FCEUWave[V]+= amp;
			if (--vcount[x]<=0)
				dcount[x]^=~0, vcount[x] = freq;
		}
	}
	CAYBC[x] = SOUNDTS;
}
#endif

static void AYSound(int32 Count)
{
	DoYM2149F();
	CAYBC[0] = CAYBC[1] = CAYBC[2] = Count;
}

#if HAVE_HQ
static void AYSoundHQ(void)
{
	DoAYSQHQ(0);
	DoAYSQHQ(1);
	DoAYSQHQ(2);
}
#endif

static DECLFW(M69_SWL)
{
(void)A;
	sunindex = V&15;
}

static DECLFW(M69_SWH)
{
(void)A;
	GameExpSound.Fill[SCSS5B] = AYSound;
#if HAVE_HQ
	GameExpSound.HiFill[SCSS5B] = AYSoundHQ;
#endif
	if (FSettings.SndRate)
		DoYM2149F();
	regs[sunindex] = V;
	switch(sunindex)
	{
	case AY_AFINE:
	case AY_ACOARSE:
		if ((wlen[0]=(regs[AY_AFINE]|(regs[AY_ACOARSE]&15)<<8)<<WSCALE+17)==0)
			wlen[0] = 1<<WSCALE+17;
		break;
	case AY_BFINE:
	case AY_BCOARSE:
		if ((wlen[1]=(regs[AY_BFINE]|(regs[AY_BCOARSE]&15)<<8)<<WSCALE+17)==0)
			wlen[1] = 1<<WSCALE+17;
		break;
	case AY_CFINE:
	case AY_CCOARSE:
		if ((wlen[2]=(regs[AY_CFINE]|(regs[AY_CCOARSE]&15)<<8)<<WSCALE+17)==0)
			wlen[2] = 1<<WSCALE+17;
		break;
	case AY_ESHAPE:
		holding = 0;
		attack = V&4 ? 31:0;
		if (V&8)
			hold = V&1, alternate = V&2 ? 31:0;
		else
			hold = 1, alternate = attack;
		env_step = 31;
		//count_env = env_period;
		break;
	case AY_EFINE:
	case AY_ECOARSE:
		if ((env_period=(regs[AY_EFINE]|regs[AY_ECOARSE]<<8)<<ESCALE+17)==0)
			env_period = 1<<ERESOLUTION+17;
		break;
	case AY_NOISEPER:
		if ((noise_period=(regs[AY_NOISEPER]&31)<<WSCALE+17)==0)
			noise_period = 1<<WSCALE+17;
	}
}

static void AYRestore(int v)
{
(void)v;
	if ((wlen[0]=(regs[AY_AFINE]|(regs[AY_ACOARSE]&15)<<8)<<WSCALE+17)==0)
		wlen[0] = 1<<WSCALE+17;
	if ((wlen[1]=(regs[AY_BFINE]|(regs[AY_BCOARSE]&15)<<8)<<WSCALE+17)==0)
		wlen[1] = 1<<WSCALE+17;
	if ((wlen[2]=(regs[AY_CFINE]|(regs[AY_CCOARSE]&15)<<8)<<WSCALE+17)==0)
		wlen[2] = 1<<WSCALE+17;
	if ((env_period=(regs[AY_EFINE]|regs[AY_ECOARSE]<<8)<<ESCALE+17)==0)
		env_period = 1<<ERESOLUTION+17;
	if ((noise_period=(regs[AY_NOISEPER]&31)<<WSCALE+17)==0)
		noise_period = 1<<WSCALE+17;
}

static void AYHiSync(int32 ts)
{
	CAYBC[0] = CAYBC[1] = CAYBC[2] = ts;
}

void Mapper69_ESI(void)
{
	GameExpSound.RChange[SCSS5B] = Mapper69_ESI;
	GameExpSound.HiSync[SCSS5B] = AYHiSync;
	memset(dcount,0,sizeof(dcount));
	memset(vcount,0,sizeof(vcount));
	memset(CAYBC,0,sizeof(CAYBC));
	AYRestore(0);
}

void NSFAY_Init(void)
{ /* end changed from FFFF to EFFF to resolve conflict with N163 board */
	sunindex = env_step = attack = hold = holding = alternate = 0;
	count_env = 0;
	SetWriteHandler(0xC000,0xDFFF,M69_SWL);
	SetWriteHandler(0xE000,0xEFFF,M69_SWH); /* Nestopia does only E000h */
	Mapper69_ESI();
	MapStateRestore = AYRestore;
	BuildVolTable();
}

void Mapper69_init(void)
{
	NSFAY_Init();

	SetupCartPRGMapping(0x10,WRAM,8192,1);

	SetWriteHandler(0x8000,0x8FFF,M69_W8);
	SetWriteHandler(0xA000,0xA000,M69_WA);
	SetWriteHandler(0x6000,0x7FFF,SUN5BWRAM);
	SetReadHandler(0x6000,0x7FFF,SUN5AWRAM);
	MapIRQHook = SunIRQHook;
	MapStateRestore = SunFixPRGR;
}
