/*
 x65xx.h
  6502-relative structures and functions

  Revised: 2016.12.22

 * Copyright notice for this file:
 *  x6502.h, x6502struct.h Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _X65xxH

#include "6502struct.h"

/* ---- constants */

#define FCEU_IQEXT		0x1
#define FCEU_IQFDS		0x2	/* prev. IQEXT2 */
/* ... */
#define FCEU_IQRESET		0x20
#define kIQNMI2			6
#define FCEU_IQNMI2		(1<<kIQNMI2) /* Delayed NMI, gets converted to *_IQNMI */
#define FCEU_IQNMI		0x80
#define FCEU_IQDMC		0x100
#define FCEU_IQFCOUNT		0x200
#define FCEU_IQTEMP		0x800
/* ... */
#define FCEU_JAMMED		0x8000

/* number of subcycles per CPU cycle */
#define CPU_SUBCYC		48u
/* interrupt edge - 1.5 cc */
#define CPU_INTEDGE		(CPU_SUBCYC*3/2)
/* cycle delay before NMI executes */
#define CPU_NMI_DELAY_CYC	3
/* total time of interrupt execution - *including prefetch op* */
#define CPU_INT_CYC		7

enum
{
	CPU_2A03_A, /* A-D: Famicom - HVC */
	CPU_2A03_B,
	CPU_2A03_C,
	CPU_2A03_D,
	CPU_2A03_E, /* Famicom - Nintendo */
	CPU_2A03_F, /* F-H have more N.C. pins */
	CPU_2A03_G, /* ref: front-loading NES */
	CPU_2A03_H, /* final generation NTSC Famicom */
	CPU_2A07, /* PAL */
	CPU_6527P /* Dendy */
};

/* ---- globals */

extern uint32 timestamp;
/* [noirq.1] [fedge.1] [rbdec.1] [radec.1] [preven.4] [dmrcol.1] [res.5] [dmad.2] [irqc.16] */
extern int32 XFStatL;
#define XFStat			((uint8*)&XFStatL)[BIG_ENDIAN ? 0:3] /* top bits */
#define XFDMA			((uint8*)&XFStatL)[BIG_ENDIAN ? 1:2]
#define XFIRQ			((uint16*)&XFStatL)[!BIG_ENDIAN] /* low bits */

extern X6502U X65xxF;
#define XFrame X65xxF.low

extern void FP_FASTAPASS(1) (*MapIRQHook)(int cpucyc);
/* count used by the debugger - pass zero */
extern void FP_FASTAPASS(1) (*DebugStepHook)(int inscount);

extern void FP_FASTAPASS(1) (*X65xx_Run)(int32 subcycles);

/* ---- prototypes */

/* positive to OR, eitherwise AND (~x) */
void X65xx_DebugLevel(int bits);

void X65xx_Init(void);
void X65xx_Reset(void);
void X65xx_Power(void);

uint8 FASTAPASS(2) X65xx_DMR(uint32 A,
	int8 c);
void FASTAPASS(2) X65xx_DMW(uint32 A,
	uint8 V);
void FASTAPASS(1) X65xx_NMI(int w);
void FASTAPASS(1) X65xx_IRQBegin(int w);
void FASTAPASS(1) X65xx_IRQEnd(int w);

void X65xx_SetModel(int model);

#define _X65xxH
#endif
